package at.ac.tuwien.vitalab.genesis2.config

import org.apache.log4j.Logger
import org.codehaus.groovy.runtime.InvokerHelper

/**
 * Configuration reader utility.
 * 
 * @author Lukasz Juszczyk
 *
 */
public class Genesis2Configuration {

    protected static Logger logger=Logger.getLogger(Genesis2Configuration.class.getSimpleName())

    private static Properties props=new Properties()
    private static checkedProps=[:]

    private static boolean initialized=false

    protected static defaultValues=[:]
    protected static valueCheckers=[:]

    public static final String LOCALHOST_HOST="net.local.host"
    public static final String LOCALHOST_PORT="net.local.port"


    // initialize the maps
    private synchronized static init() {

	if (initialized) {
	    return;
	}

	// declare mandatory entries

	declareConfigEntry(LOCALHOST_HOST, "localhost") { val ->
	    InetAddress.getByName(val)
	}

	declareConfigEntry(LOCALHOST_PORT, 8181) { val ->
	    (val instanceof Number) && val>0 && val<65536
	}

	// read config from file

	String confPath="conf/genesis2.conf";

	try {
	    def reader=new FileReader(confPath);
	    props.load(reader);
	    reader.close();

	} catch (e) {
	    logger.warn("Cannot read from config file: "+e.message);
	}

	props.putAll(System.getProperties())

	initialized=true;
    }


    // get a value
    def static Object get(def name) {

	init()

	// see whether this property has been seen and checke before
	def result=checkedProps.get(name);

	if (result) {
	    return result
	}

	// otherwise take from config file
	result=props.get(name)

	// see whether it is a
	def defVal=defaultValues.get(name)

	// transfrom from string to another type?
	if (result && defVal && defVal.class!=String) {
	    try {
		def tmp=InvokerHelper.invokeConstructorOf(defVal.class, result)
		result = tmp
	    } catch (e) {

		def className=result.class.name

		logger.warn("Config entry '$name' has invalid type : "+
			"$result ($className). Taking default value: $defVal.");

		checkedProps.put(name,defVal)
		return defVal
	    }
	}


	// check whether the value is valid
	def checker=valueCheckers.get(name)

	if (result && !checkValue(result,checker)) {

	    logger.warn("Config entry '$name' has invalid "+
		    "value : $result. Taking default value: $defVal.");

	    result=defVal
	}

	// if no value has been read, take the default one

	if (!result) {
	    result=defVal
	}

	// save to the map containin the checked values
	if (result) {
	    checkedProps.put(name,result)
	}

	return result;
    }


    public static String getLocalHost() {
	return get(LOCALHOST_HOST)
    }



    public static Integer getLocalPort() {
	return get(LOCALHOST_PORT)
    }


    private static checkValue(def value, checker) {

	boolean ok=false

	if (checker) {
	    try {
		if (checker instanceof IConfigEntryChecker) {
		    if (checker.check(value)) {
			ok=true
		    }
		} else if (checker instanceof Closure) {
		    if (checker.call(value)) {
			ok=true
		    }
		} else assert false;
	    } catch (e) {
	    }

	    return ok
	}

	return true
    }

    // for declaring new config entries, together with a default value
    // and a closure for checking the validity
    def static declareConfigEntry(String name, defValue, Closure checker) {

	declare(name,defValue,checker)
    }

    // for declaring new config entries, together with a default value
    // and a checker class for verifying the validity
    def static declareConfigEntry(String name, defValue, IConfigEntryChecker checker) {

	declare(name,defValue,checker)
    }

    def private static declare(name, defValue, checker) {
	if (!checkValue(defValue,checker)) {
	    throw new IllegalArgumentException("Checker closure rejected default value $defValue");
	}

	defaultValues.put(name,defValue)
	valueCheckers.put(name,checker)
    }
}
