package at.ac.tuwien.vitalab.genesis2.utils

import groovy.xml.QName


/**
 * Utility class for reading XSD files.
 * 
 * @author Lukasz Juszczyk
 *
 */
public class XsdReader {

    private final static XSD_NS="http://www.w3.org/2001/XMLSchema"

    /**
     * Returns a specific type/element definition from an XSD.
     * 
     * @param type <code>true</code> if type definition is required, <code>false</code> if element definition.
     * @param xsdFile XSD.
     * @param name Type/element name.
     * @return Definition as String.
     */
    def private static String getDefinition(boolean type, File xsdFile, String name) {

	def doc=new XmlParser().parse(xsdFile)

	def typeStr=type ? "complexType": "element"

	def s=null

	doc.children().grep {

	    it.name() == new QName(XSD_NS, typeStr) &&
		    it.attributes().get("name") == name
	}.each { el ->
	    def sw = new StringWriter()
	    def pw = new PrintWriter(sw)

	    XmlNodePrinter xnp=new XmlNodePrinter(pw)

	    xnp.print(el)

	    pw.close()

	    s = sw.toString()
	}

	if (s!=null) {
	    return s
	}

	throw new IllegalArgumentException("Cannot find "+typeStr+" '"+name+"' in XSD file '"+xsdFile+"'")
    }

    /**
     * Returns a specific element definition from an XSD.
     * 
     * @param xsdFile XSD.
     * @param name Element name.
     * @return Definition as String.
     */
    def static String getElementDefinition(File xsdFile, String name) {
	return getDefinition(false, xsdFile, name)
    }

    /**
     * Returns a specific type definition from an XSD.
     * 
     * @param xsdFile XSD.
     * @param name Type name.
     * @return Definition as String.
     */
    def static String getTypeDefinition(File xsdFile, String name) {
	return getDefinition(true, xsdFile, name)
    }

    /**
     * Returns the target namespace of the XSD.
     * 
     * @param xsdFile XSD.
     * @return Namespace.
     */
    def static String getTargetNamespace(File xsdFile) {

	def doc=new XmlParser().parse(xsdFile)
	return doc.attributes().targetNamespace;
    }
}
