package at.ac.tuwien.vitalab.genesis2;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.concurrent.Semaphore;

import org.apache.log4j.Logger;

import at.ac.tuwien.vitalab.genesis2.base.Genesis2Base;
import at.ac.tuwien.vitalab.genesis2.base.PluginRepositoryMonitor;
import at.ac.tuwien.vitalab.genesis2.control.Genesis2Frontend;
import at.ac.tuwien.vitalab.genesis2.core.model.AbstractModelElement;
import at.ac.tuwien.vitalab.genesis2.plugin.AbstractGenesis2Plugin;

/**
 * Main executable class for Genesis2.
 * 
 * @author Lukasz Juszczyk
 * 
 */
public final class Genesis2 {

    private static final Logger LOGGER = Logger.getLogger(Genesis2.class);

    private static Genesis2Base base = null;
    private static Genesis2Frontend frontend;

    // flags which tell, what should be done
    private static boolean startFrontend = false;
    private static boolean execScript = false;
    private static String scriptPath;

    private static Semaphore bootstrapSem = new Semaphore(0);

    /**
     * Constructor.
     */
    private Genesis2() {
	super();
    }

    /**
     * Main executable.
     * 
     * @param args Arguments.
     */
    public static void main(String[] args) {

	for (int i = 0; i < args.length; i++) {
	    if (args[i].equals("-frontend") || args[i].equals("-fe")) {
		startFrontend = true;
	    }
	    if (args[i].equals("-script") || args[i].equals("-s")) {
		try {
		    scriptPath = args[++i];
		    startFrontend = true;
		    execScript = true;
		} catch (Exception e) {
		    LOGGER.fatal("No script file specified.");
		    System.exit(1);
		}
	    }
	}

	start();

	int exitCode = 0;

	if (startFrontend) {

	    try {
		if (execScript) {
		    LOGGER.info("Genesis2 is running script '" + scriptPath + "'.");

		    frontend.execScript(new File(scriptPath));
		} else {

		    LOGGER.info("Genesis2 is starting console GUI.");

		    frontend.run();
		}
	    } catch (Exception e) {
		LOGGER.error(e);
	    }

	} else {

	    System.out.println("Genesis2 is running. Enter 'exit' or 'x' to quit."); // NOPMD

	    // listen on stdin for shutdown commands
	    BufferedReader reader = new BufferedReader(new InputStreamReader(System.in));

	    while (true) {

		try {
		    String inputLine = reader.readLine();

		    if (inputLine == null || inputLine.equalsIgnoreCase("exit") || inputLine.equalsIgnoreCase("x")) {
			break;
		    }
		} catch (IOException e) {
		    System.err.println("Cannot read from stdin: " + e.getMessage()); // NOPMD
		    exitCode = 2;
		    break;
		}
	    }
	}

	stop();

	System.exit(exitCode);
    }

    /**
     * Returns the frontend handler.
     * 
     * @return Frontend.
     */
    public static Genesis2Frontend getFrontend() {
	return frontend;
    }

    /**
     * Start procedure, basically bootstraps all plugins.
     */
    public static void start() {

	BufferedReader reader = null;

	try {
	    base = new Genesis2Base();
	    base.start();

	    File pluginConf = new File("conf/plugins.conf");

	    FileInputStream fis = new FileInputStream(pluginConf);

	    reader = new BufferedReader(new InputStreamReader(fis));

	    String pluginClassName = null;

	    while ((pluginClassName = reader.readLine()) != null) {
		try {
		    AbstractGenesis2Plugin.startPlugin(pluginClassName);
		} catch (Exception e) {
		    LOGGER.error("Error while loading plugin '" + pluginClassName + "': " + e.getMessage());
		}
	    }

	} catch (Exception e1) {
	    LOGGER.fatal("Cannot start Genesis2 Base: ", e1);
	    stop();
	    System.exit(1);
	} finally {
	    try {
		if (reader != null) {
		    reader.close();
		}
	    } catch (IOException e) {
		LOGGER.warn(e);
	    }
	}

	frontend = new Genesis2Frontend();

	PluginRepositoryMonitor.start();

	bootstrapSem.release();
    }

    /**
     * Blocks until G2 has been started.
     * 
     * @throws InterruptedException On interrupts.
     */
    public static void waitUntilStarted() throws InterruptedException {
	bootstrapSem.acquire();
    }

    /**
     * Shutdown.
     */
    public static void stop() {

	PluginRepositoryMonitor.stop();

	for (AbstractGenesis2Plugin<?> gp : AbstractGenesis2Plugin.getRunningPlugins()) {
	    try {
		gp.stop();
	    } catch (Exception e) {
		LOGGER.error("Cannot shut down plugin " + gp.getPluginName(), e);
	    }
	}

	AbstractModelElement.cleanupAllElements();
    }

    /**
     * Is this a frontend instance?
     * 
     * @return <code>true</code> if frontend.
     */
    public static boolean isFrontend() {
	return startFrontend;
    }

    /**
     * Is this a backend instance?
     * 
     * @return <code>true</code> if backend.
     */
    public static boolean isBackend() {
	return !startFrontend;
    }

}
