package at.ac.tuwien.vitalab.genesis2.base;

import groovy.lang.Binding;
import groovy.lang.Closure;
import groovy.lang.MissingMethodException;
import groovy.lang.MissingPropertyException;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import lombok.Getter;

import org.apache.log4j.Logger;
import org.codehaus.groovy.runtime.InvokerHelper;

import at.ac.tuwien.vitalab.genesis2.utils.ClosureUtil;

/**
 * Execution environment for front-end control and back-end instances.
 * 
 * @author Lukasz Juszczyk
 * 
 */
public class ExecutionEnvironment extends Binding implements Cloneable {

    private static final Logger LOGGER = Logger.getLogger(ExecutionEnvironment.class);

    protected static ExecutionEnvironment globalExecEnv = null;

    @Getter
    protected final List<Object> varBindingInterceptors = new ArrayList<Object>();

    private static final ClosureUtil CLOSURE_UTIL = ClosureUtil.getInstance();

    /**
     * Factory.
     * 
     * @return Instance.
     */
    public synchronized static ExecutionEnvironment getGlobalInstance() {

	if (globalExecEnv != null) {
	    return globalExecEnv;
	}

	LOGGER.debug("Creating global execution environment.");

	return globalExecEnv = new ExecutionEnvironment();
    }

    /**
     * Clean up all declared variables.
     */
    public synchronized static void clearGlobalInstance() {
	getGlobalInstance().getVariables().clear();
    }

    /**
     * Constructor.
     */
    public ExecutionEnvironment() {

	super();

	LOGGER.debug("Creating execution environment.");

    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Object#clone()
     */
    @Override
    public ExecutionEnvironment clone() throws CloneNotSupportedException { // NOPMD
	LOGGER.debug("Cloning execution environment.");

	ExecutionEnvironment env = new ExecutionEnvironment();

	env.varBindingInterceptors.addAll(this.varBindingInterceptors);

	return env;
    }

    /*
     * (non-Javadoc)
     * 
     * @see groovy.lang.Binding#setProperty(java.lang.String, java.lang.Object)
     */
    @Override
    public void setProperty(String property, Object newValue) {

	LOGGER.debug("Property '" + property + "' got set to '" + newValue + "'.");

	super.setProperty(property, newValue);
    }

    /*
     * (non-Javadoc)
     * 
     * @see groovy.lang.Binding#setVariable(java.lang.String, java.lang.Object)
     */
    @Override
    public void setVariable(String name, Object value) {

	LOGGER.debug("Setting variable '" + name + "' got set to '" + value + "'.");

	super.setVariable(name, value);
    }

    /*
     * (non-Javadoc)
     * 
     * @see groovy.lang.GroovyObjectSupport#invokeMethod(java.lang.String,
     * java.lang.Object)
     */
    @Override
    public Object invokeMethod(String name, Object args) {

	LOGGER.debug("Calling method " + name + "(" + InvokerHelper.asList(args) + ").");

	try {
	    Closure closure = (Closure) getVariable(name);
	    closure = (Closure) closure.clone();
	    CLOSURE_UTIL.setPrimaryDispatcher(closure, this);

	    return closure.call(InvokerHelper.asArray(args));
	} catch (Exception e) {
	    // ignore and go on with fallback
	}

	for (Object go : varBindingInterceptors) {
	    try {
		return InvokerHelper.invokeMethod(go, name, args);
	    } catch (MissingMethodException mme) {
		// try next
	    }
	}

	// global macro?
	if (this != globalExecEnv) {
	    try {
		return globalExecEnv.invokeMethod(name, args);
	    } catch (MissingMethodException mme) {
		// try next fallback
	    }
	}

	return super.invokeMethod(name, args);
    }

    /*
     * (non-Javadoc)
     * 
     * @see groovy.lang.Binding#getProperty(java.lang.String)
     */
    @Override
    public Object getProperty(String property) {

	for (Object go : varBindingInterceptors) {
	    try {
		Object p = InvokerHelper.getProperty(go, property);
		if (p != null) {
		    return p;
		}
	    } catch (MissingPropertyException mpe) {
		// ignore and go on with fallback
	    }
	}

	Object prop = null;

	try {
	    prop = super.getProperty(property);
	} catch (MissingPropertyException mpe) {
	    // ignore and go on with fallback
	}

	if (this != globalExecEnv && prop == null) {
	    prop = globalExecEnv.getProperty(property);
	}

	return prop;
    }

    /*
     * (non-Javadoc)
     * 
     * @see groovy.lang.Binding#getVariable(java.lang.String)
     */
    @Override
    public Object getVariable(String name) {

	for (Object go : varBindingInterceptors) {
	    try {
		Object p = InvokerHelper.getProperty(go, name);
		if (p != null) {
		    return p;
		}
	    } catch (MissingPropertyException mpe) {
		// ignore and go on
	    }
	}

	Object prop = null;

	try {
	    prop = super.getVariable(name);
	} catch (MissingPropertyException mpe) {
	    // ignore and go on
	}

	if (this != globalExecEnv && prop == null) {
	    prop = globalExecEnv.getVariable(name);
	}

	return prop;
    }

    /*
     * (non-Javadoc)
     * 
     * @see groovy.lang.Binding#getVariables()
     */
    @SuppressWarnings("unchecked")
    @Override
    public Map<?, ?> getVariables() {
	Map<Object, Object> vars = new HashMap<Object, Object>();

	if (this != globalExecEnv) {
	    vars.putAll(globalExecEnv.getVariables());
	}

	vars.putAll(super.getVariables());

	return vars;
    }
}
