package at.ac.tuwien.vitalab.genesis2.base;

import groovy.util.Expando;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.concurrent.locks.ReentrantLock;
import java.util.jar.Attributes;
import java.util.jar.JarFile;
import java.util.jar.Manifest;

import org.apache.log4j.Logger;

import at.ac.tuwien.vitalab.genesis2.errors.G2Exception;
import at.ac.tuwien.vitalab.genesis2.plugin.AbstractGenesis2Plugin;
import at.ac.tuwien.vitalab.genesis2.utils.Base64;
import at.ac.tuwien.vitalab.genesis2.utils.G2UrlClassLoader;

/**
 * Installs automatically Genesis 2 plugins from their directory.
 * 
 * @author Lukasz Juszczyk
 * 
 */
public class PluginRepositoryMonitor extends Expando {

    private static final Logger LOGGER = Logger.getLogger(PluginRepositoryMonitor.class);

    // main lock
    private static final ReentrantLock LOCK = new ReentrantLock();

    private static Boolean monitorStarted = false;

    private static MonitorThread thread = null;

    // the files of the plugins
    private static Map<String, File> files = Collections.synchronizedMap(new HashMap<String, File>());

    private static G2UrlClassLoader loader = new G2UrlClassLoader();

    /**
     * Private constructor.
     */
    private PluginRepositoryMonitor() {
	super();
    }

    /**
     * Returns the file of a plugin.
     * 
     * @param name Plugin name.
     * @return Plugin file.
     */
    public static File getFileOfPlugin(String name) {

	File file = files.get(name);

	if (file.exists() && file.canRead()) {
	    return file;
	}

	throw new IllegalArgumentException("Unknown plugin '" + name + "' or " + "its Jar file has been deleted.");
    }

    /**
     * Returns the Base64-encoded version of the plugin file.
     * 
     * @param name Plugin file.
     * @return Base64-encoded String.
     * 
     * @throws IOException If IO errors happen during packaging.
     */
    public static String getBase64EncodedPlugin(String name) throws IOException {

	File pluginFile = getFileOfPlugin(name);

	return Base64.encodeFromFile(pluginFile.getAbsolutePath());
    }

    /**
     * Installs a new plugin.
     * 
     * @param fileName Name of the plugin file.
     * @param jarContentBase64 Base64-encoded plugin file.
     * 
     * @throws IOException If the plugin archive cannot be read or files cannot
     *             be created.
     */
    public static void installPlugin(String fileName, String jarContentBase64) throws IOException {

	LOCK.lock();

	try {

	    File pluginDir = new File("plugins");

	    if (!pluginDir.canWrite() || !pluginDir.isDirectory()) {
		throw new IOException("Cannot access plugin directory.");
	    }

	    File jarFile = new File(pluginDir, fileName);
	    FileOutputStream fos = new FileOutputStream(jarFile);

	    fos.write(Base64.decode(jarContentBase64));
	    fos.close();

	    if (thread != null) {
		// install it immediately
		thread.startPlugin(jarFile);
	    }

	} finally {
	    LOCK.unlock();
	}
    }

    /**
     * Start the monitor which automatically deploys all plugins in the repo.
     */
    public static void start() {

	if (monitorStarted) {
	    return;
	}

	MonitorThread.running = true;

	synchronized (monitorStarted) {

	    if (!monitorStarted) {

		(thread = new MonitorThread()).start();

		LOGGER.info("Started plugin repository monitor.");

		monitorStarted = true;
	    }
	}
    }

    /**
     * Stop the monitor.
     */
    public static void stop() {

	if (!monitorStarted) {
	    return;
	}

	synchronized (monitorStarted) {

	    if (monitorStarted) {

		MonitorThread.running = false;

		if (thread != null && thread.isAlive()) {
		    try {
			thread.join();
		    } catch (InterruptedException e) {
			// ignore
		    }
		}

	    }
	}
    }

    // plugin-monitoring thread
    private static class MonitorThread extends Thread {

	// if set to false, thread will terminate itself
	protected static boolean running = true;

	// last seen modification timestamps of plugin files
	// used to determine whether plugin file got updated
	private final Map<File, Long> timestamps = new HashMap<File, Long>();

	// plugins waiting for dependencies to be deployed
	private final Map<File, Collection<String>> waiting = new HashMap<File, Collection<String>>();

	/**
	 * Constructor.
	 */
	public MonitorThread() {
	    super();
	    setDaemon(true);
	}

	public void run() {

	    while (running) {

		LOCK.lock();

		try {
		    check();
		} catch (Exception e) {
		    LOGGER.warn("Error while monitoring plugin directory.", e);
		} finally {

		    LOCK.unlock();

		    try {
			// check every 2 seconds
			Thread.sleep(2000);
		    } catch (InterruptedException e) {
			// ignore
		    }
		}

	    }

	    PluginRepositoryMonitor.monitorStarted = false;

	    LOGGER.info("Stopped plugin repository monitor.");
	}

	// check for new plugins
	private void check() {

	    File pluginDir = new File("plugins");

	    if (pluginDir.canRead() && pluginDir.isDirectory()) {

		// iterate through all jar files
		for (File pluginFile : pluginDir.listFiles()) {

		    if (pluginFile.getName().endsWith(".jar") && pluginFile.canRead()) {

			pluginFile = pluginFile.getAbsoluteFile();
			long timestamp = pluginFile.lastModified();

			if (timestamps.containsKey(pluginFile) && timestamp == timestamps.get(pluginFile)) {
			    continue;
			}

			timestamps.put(pluginFile, timestamp);

			try {
			    startPlugin(pluginFile);
			} catch (Exception e) {
			    LOGGER.warn("Cannot install plugin " + pluginFile.getName(), e);
			    continue;
			}
		    }
		}
	    }
	}

	/**
	 * Starts a plugin from an archive.
	 * 
	 * @param pluginFile
	 * @throws G2Exception
	 * @throws IOException
	 */
	public void startPlugin(File pluginFile) throws G2Exception, IOException {

	    JarFile jarFile = new JarFile(pluginFile);
	    Manifest manifest = jarFile.getManifest();

	    if (manifest == null || manifest.getMainAttributes() == null) {
		throw new IllegalArgumentException("No manifest entry.");
	    }

	    Attributes atts = manifest.getMainAttributes();

	    String name = (String) atts.getValue("Plugin-Name");
	    String className = (String) atts.getValue("Plugin-Class");
	    String dependencyStr = (String) atts.getValue("Plugin-Depends");

	    if (name == null || className == null) {
		throw new IllegalArgumentException(
			"Manifest does not contain mandatory 'Plugin-Name' and 'Plugin-Class' entries.");
	    }

	    if (AbstractGenesis2Plugin.hasPluginInstalled(name)) {
		// already installed? skip the rest
		return;
	    }

	    List<String> dependencies = new ArrayList<String>();

	    if (dependencyStr != null) {
		StringTokenizer deps = new StringTokenizer(dependencyStr, ", ");

		while (deps.hasMoreTokens()) {

		    String depName = deps.nextToken();

		    if (!AbstractGenesis2Plugin.hasPluginInstalled(depName)) {
			dependencies.add(depName);
		    }
		}

		if (!dependencies.isEmpty()) {
		    waiting.put(pluginFile, dependencies);
		    throw new G2Exception("Cannot deploy now due to missing dependencies " + dependencies
			    + ". Waiting...");
		}
	    }

	    loader.addURL(pluginFile.toURI().toURL());

	    AbstractGenesis2Plugin<?> plugin;

	    try {
		Class<? extends AbstractGenesis2Plugin<?>> clazz = (Class<? extends AbstractGenesis2Plugin<?>>) loader
			.loadClass(className);
		Constructor<? extends AbstractGenesis2Plugin<?>> constructor = clazz.getConstructor(new Class<?>[] {});
		plugin = constructor.newInstance();
	    } catch (Exception e) {
		throw new G2Exception("Cannot instantiate plugin.", e);
	    }

	    if (!name.equals(plugin.getPluginName())) {
		LOGGER.warn("Plugin " + pluginFile.getName() + " has been declared with different name ('" + name
			+ "') in Manifest file.");
	    }

	    files.put(name, pluginFile);
	    plugin.start();

	    LOGGER.info("Plugin " + pluginFile.getName() + " has been installed.");

	    waiting.remove(pluginFile);

	    for (File pf : waiting.keySet()) {
		Collection<String> pDeps = waiting.get(pf);

		if (pDeps.contains(name)) {
		    startPlugin(pf);
		}
	    }
	}
    }
}
