package at.ac.tuwien.vitalab.genesis2.control;

import groovy.lang.Binding;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;
import java.util.concurrent.Semaphore;

import org.apache.log4j.Logger;
import org.codehaus.groovy.control.CompilationFailedException;

import at.ac.tuwien.vitalab.genesis2.base.ExecutionEnvironment;
import at.ac.tuwien.vitalab.genesis2.base.Genesis2Base;
import at.ac.tuwien.vitalab.genesis2.base.PluginRepositoryMonitor;
import at.ac.tuwien.vitalab.genesis2.base.ws.IGenesis2BaseWebService;
import at.ac.tuwien.vitalab.genesis2.core.model.AbstractModelElement;
import at.ac.tuwien.vitalab.genesis2.core.model.ElementMap;
import at.ac.tuwien.vitalab.genesis2.core.model.IModelObserver;
import at.ac.tuwien.vitalab.genesis2.core.model.types.Host;
import at.ac.tuwien.vitalab.genesis2.plugin.AbstractGenesis2Plugin;
import at.ac.tuwien.vitalab.genesis2.plugin.IPluginObserver;

/**
 * Front-end control center.
 * 
 * @author Lukasz Juszczyk
 * 
 */
public class Genesis2Frontend {

    protected static final Logger LOGGER = Logger.getLogger(Genesis2Frontend.class);

    protected Map<Host, Collection<String>> pluginsOnHosts = new HashMap<Host, Collection<String>>();

    protected List<String> localPlugins = new ArrayList<String>();

    /**
     * Constructor.
     * 
     */
    public Genesis2Frontend() {

	ElementMap<Host> hostMap = AbstractModelElement.getElementMap(Host.class);

	// start monitoring in order to deploy plugins at backend hosts
	hostMap.observe(new HostObserver());
	AbstractGenesis2Plugin.observe(new PluginObserver());
    }

    /**
     * Main executable for GUI.
     * 
     */
    public void run() {

	Semaphore sem = new Semaphore(1);

	try {
	    sem.acquire();
	} catch (InterruptedException e) {
	    // ignore
	}

	Binding binding = ExecutionEnvironment.getGlobalInstance();

	G2Console console = new G2Console(binding, sem);

	console.setG2Shell(InterceptingGroovyShell.newInstance(binding));

	console.run();

	try {
	    sem.acquire();
	} catch (InterruptedException e) {
	    // ignore
	}
    }

    /**
     * Returns the ID-String for this script file, which is needed by the Groovy
     * console.
     * 
     * @param file Script file.
     * @return ID.
     */
    private static String getScriptIdFromFile(File file) {
	int lastDotIndex = file.getName().lastIndexOf('.');
	return file.getName().substring(0, lastDotIndex);
    }

    /**
     * Main executable for script input.
     * 
     * @param scriptFile Script file.
     * @throws CompilationFailedException On syntax errors.
     * @throws IOException On IO errors.
     * 
     * @return Script result.
     */
    public Object execScript(File scriptFile) throws CompilationFailedException, IOException {

	Binding binding = ExecutionEnvironment.getGlobalInstance();

	InterceptingGroovyShell shell = InterceptingGroovyShell.newInstance(binding);

	byte[] buffer = new byte[(int) scriptFile.length()];

	BufferedInputStream bis = null;

	try {
	    bis = new BufferedInputStream(new FileInputStream(scriptFile));
	    bis.read(buffer);
	    String script = new String(buffer);

	    shell.setId(getScriptIdFromFile(scriptFile));

	    return shell.run(script, scriptFile.getAbsolutePath(), new String[] {});

	} finally {
	    if (bis != null) {
		bis.close();
	    }
	}
    }

    /**
     * Generic plugin propagator to the passed host.
     * 
     * @param host Host.
     */
    protected void installMissingPlugins(Host host) {

	if (!host.isLocalhost()) {

	    // get client directly
	    Genesis2Base basePlugin = (Genesis2Base) AbstractGenesis2Plugin.getPlugin(Genesis2Base.class);

	    if (basePlugin == null) {
		LOGGER.fatal("Cannot get instance of Genesis2 Base Plugin.");
		return;
	    }

	    IGenesis2BaseWebService client = null;

	    try {
		client = (IGenesis2BaseWebService) basePlugin.getPluginWsClient(host);
	    } catch (Exception e) {
		LOGGER.warn("Cannot get instance of Genesis2 Base Plugin Web service client.", e);
		return;
	    }

	    Collection<String> remotePlugins = client.listPlugins();
	    Collection<String> plugins = new Vector<String>(localPlugins);
	    plugins.removeAll(remotePlugins);

	    // install all of them
	    for (String pluginName : plugins) {
		try {
		    String jarName = PluginRepositoryMonitor.getFileOfPlugin(pluginName).getName();
		    String b64 = PluginRepositoryMonitor.getBase64EncodedPlugin(pluginName);
		    client.installPlugin(jarName, b64);
		    remotePlugins.add(pluginName);
		} catch (Exception e) {
		    LOGGER.warn("Cannot install plugin '" + pluginName + "' on remote host. Removing host.", e);
		    host.deregister();
		}
	    }

	    pluginsOnHosts.put(host, remotePlugins);
	}
    }

    /**
     * Checks whether new back end hosts have been referenced.
     * 
     * @author Lukasz Juszczyk
     * 
     */
    protected class HostObserver implements IModelObserver<Host> {

	@Override
	public void onChangedModelElement(Host t, int how, Object... info) {

	    synchronized (pluginsOnHosts) {
		switch (how) {
		case IModelObserver.CREATED:
		    installMissingPlugins(t);
		    break;
		case IModelObserver.REMOVED:
		    pluginsOnHosts.remove(t);
		    break;
		default:
		}
	    }
	}

    }

    /**
     * Checks whether new plugins have been deployed.
     * 
     * @author Lukasz Juszczyk
     * 
     */
    protected class PluginObserver implements IPluginObserver {

	@Override
	public void onPluginStarted(AbstractGenesis2Plugin<?> plugin) {
	    synchronized (pluginsOnHosts) {
		localPlugins.add(plugin.getPluginName());
		for (Host h : pluginsOnHosts.keySet()) {
		    installMissingPlugins(h);
		}
	    }
	}

	@Override
	public void onPluginStopped(AbstractGenesis2Plugin<?> plugin) {
	    synchronized (pluginsOnHosts) {
		localPlugins.remove(plugin.getPluginName());
	    }
	}
    }
}
