package at.ac.tuwien.vitalab.genesis2.core.model;

import groovy.lang.Closure;
import groovy.util.Expando;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;

import org.apache.log4j.Logger;

import at.ac.tuwien.vitalab.genesis2.core.model.types.Host;
import at.ac.tuwien.vitalab.genesis2.core.model.ws.IModelElementWebService;
import at.ac.tuwien.vitalab.genesis2.core.model.ws.wrappers.KeyValueWrapper;
import at.ac.tuwien.vitalab.genesis2.core.model.ws.wrappers.ModelElementWrapper;
import at.ac.tuwien.vitalab.genesis2.errors.G2Exception;
import at.ac.tuwien.vitalab.genesis2.errors.G2RemoteException;
import at.ac.tuwien.vitalab.genesis2.plugin.AbstractGenesis2Plugin;
import at.ac.tuwien.vitalab.genesis2.utils.ClosureUtil;
import at.ac.tuwien.vitalab.genesis2.utils.ModelElementUtil;

/**
 * Abstract element of the model tree.
 * 
 * @author Lukasz Juszczyk
 * 
 */
public abstract class AbstractModelElement extends Expando implements ICleanable, Cloneable {

    /**
     * Key = Type class Value = Plugin that introduces this class.
     */
    private static final Map<Class<? extends AbstractModelElement>, AbstractGenesis2Plugin<? extends AbstractModelElement>> PLUGIN_MAP = Collections
	    .synchronizedMap(new HashMap<Class<? extends AbstractModelElement>, AbstractGenesis2Plugin<? extends AbstractModelElement>>());

    /**
     * Key = Type name. Value = Type class.
     */
    private static final Map<String, Class<? extends AbstractModelElement>> TYPE_MAP = new HashMap<String, Class<? extends AbstractModelElement>>();

    /**
     * Key = Type class. Value = ElementMap for this type class.
     */
    private static final Map<Class<? extends AbstractModelElement>, ElementMap<? extends AbstractModelElement>> ELEMENTMAP_MAP = new HashMap<Class<? extends AbstractModelElement>, ElementMap<? extends AbstractModelElement>>();

    /**
     * Element map for this type of element.
     */
    @SuppressWarnings("unchecked")
    protected ElementMap<AbstractModelElement> myMap = (ElementMap<AbstractModelElement>) getElementMap(getClass());

    /**
     * Children.
     */
    protected static Map<AbstractModelElement, Set<AbstractModelElement>> kids = new HashMap<AbstractModelElement, Set<AbstractModelElement>>();

    /**
     * Parents.
     */
    protected static Map<AbstractModelElement, Set<AbstractModelElement>> parents = new HashMap<AbstractModelElement, Set<AbstractModelElement>>();

    /**
     * Whether a property lookup should be followed up to the parent(s).
     */
    protected boolean followUpProps = true;

    /**
     * Logger for static methods.
     */
    protected static final Logger LOGGER = Logger.getLogger(AbstractModelElement.class);

    /**
     * Whether this element is locked for remote updates.
     */
    protected boolean locked = false;

    /**
     * At which hosts has this element been created.
     */
    protected final Set<Host> createdAt = new HashSet<Host>();

    /**
     * At which hosts has this element been deployed.
     */
    protected final Set<Host> deployedAt = new HashSet<Host>();

    /**
     * To which hosts has this element been bound.
     */
    protected final Set<Host> boundTo = new HashSet<Host>();

    /**
     * Element hash codes of last updates per host. Indicators whether element
     * is "dirty".
     */
    protected Map<Host, Integer> remotePropHash = new HashMap<Host, Integer>();

    /**
     * Utility class.
     */
    protected final static ModelElementUtil UTIL = ModelElementUtil.getInstance();

    /**
     * Constructor.
     */
    public AbstractModelElement() {

	super();

	setPropertyDirectly("id", UUID.randomUUID().toString());
	setPropertyDirectly("type", this.getClass().getSimpleName());
	setPropertyDirectly("tags", new HashSet<String>());

	getMap();
    }

    /**
     * Constructor.
     * 
     * @param id Element's id.
     */
    public AbstractModelElement(String id) {

	this();

	setPropertyDirectly("id", id);
    }

    /**
     * Returns this elements UUID.
     * 
     * @return UUID string.
     */
    public final String getId() {
	return (String) getProperty("id");
    }

    /**
     * Overrides this elements UUID.
     * 
     * @param id ID String.
     */
    public synchronized final void overrideId(String id) { // NOPMD

	String oldId = getId();

	if (oldId != null && oldId.equals(id)) {
	    // same? nothing to do
	    return;
	}

	if (oldId != null) {
	    deregister();
	}

	setPropertyDirectly("id", id);
	register();
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Object#clone()
     */
    @Override
    public AbstractModelElement clone() throws CloneNotSupportedException {

	AbstractModelElement result;
	try {
	    result = this.getClass().newInstance();
	} catch (Exception e) {
	    throw new CloneNotSupportedException("Cannot create new instance of " + getClass().getName() + ".");
	}

	Map<?, ?> props = getProperties();

	for (Object key : props.keySet()) {
	    if (key.toString().equals("id")) {
		continue;
	    }
	    Object value = props.get(key);
	    result.setPropertyDirectly(key.toString(), value);
	}

	result.register();

	return result;
    }

    /**
     * Create this model element at the passed hosts.
     * 
     * @param hosts Hosts.
     */
    public void createAt(Host... hosts) {

	for (Host h : hosts) {
	    if (!h.isRegistered()) {
		continue;
	    }

	    boundTo.add(h);
	}

	update();
    }

    /**
     * Create this model element at the passed hosts.
     * 
     * @param hosts Hosts.
     */
    public void createAt(Collection<Host> hosts) {
	createAt(hosts.toArray(new Host[] {}));
    }

    /**
     * Attaches this model element to the passed hosts.
     * 
     * @param hosts Hosts.
     */
    public void bindTo(Host... hosts) {

	for (Host h : hosts) {

	    if (boundTo.contains(h) || !h.isRegistered()) {
		continue;
	    }

	    h.attach(this);
	    boundTo.add(h);
	}

	update();
    }

    /**
     * Attaches this model element to the passed hosts.
     * 
     * @param hosts Hosts.
     */
    public void bindTo(Collection<Host> hosts) {
	bindTo(hosts.toArray(new Host[] {}));
    }

    /**
     * Detaches this model element from the passed hosts.
     * 
     * @param hosts Hosts.
     */
    public void releaseFrom(Host... hosts) {
	for (Host h : hosts) {

	    if (boundTo.contains(h)) {
		h.detach(this);
		boundTo.remove(h);
	    }
	}
    }

    /**
     * Detaches this model element from the passed hosts.
     * 
     * @param hosts Hosts.
     */
    public void releaseFrom(Collection<Host> hosts) {
	releaseFrom(hosts.toArray(new Host[] {}));
    }

    /**
     * Registers an element at the map.
     */
    protected void register() {

	myMap.registerIfNew(this);
    }

    /**
     * Unregisters this element from the map.
     */
    public synchronized void deregister() {

	// for all elements which are not hosts, remove them from remote hosts
	if (!(this instanceof Host)) {

	    for (Host host : createdAt) {

		if (!host.isLocalhost()) {
		    getWsClient(host).remove(getId());
		}
	    }
	}

	myMap.deregister(this); // is safe
    }

    /**
     * Checks whether the element has been registered at the map.
     * 
     * @return <code>true</code> if registered, otherwise <code>false</code>.
     */
    protected boolean isRegistered() {
	return myMap.has(getId());
    }

    /**
     * Returns the declaring plugin.
     * 
     * @return Plugin instance.
     */
    public AbstractGenesis2Plugin<?> getPlugin() {
	return myMap.getPlugin();
    }

    /**
     * Returns the map containing all elements of this type.
     * 
     * @return Map.
     */
    public final ElementMap<? extends AbstractModelElement> getMap() {
	return myMap;
    }

    /**
     * Clean up everything.
     */
    public static void cleanupAllElements() {

	for (ElementMap<? extends AbstractModelElement> map : ELEMENTMAP_MAP.values()) {
	    for (AbstractModelElement element : map.getAll()) {
		element.cleanup();
	    }
	}
    }

    /*
     * (non-Javadoc)
     * 
     * @see groovy.util.Expando#setProperty(java.lang.String, java.lang.Object)
     */
    @Override
    public synchronized void setProperty(String property, Object newValue) { // NOPMD

	setProperty(property, newValue, true);
    }

    /**
     * Sets a property only on the local element, skips remote updates.
     * 
     * @param property Property name.
     * @param newValue Value.
     */
    protected synchronized void setPropertyLocally(String property, Object newValue) { // NOPMD
	setProperty(property, newValue, false);
    }

    /**
     * Sets a property value.
     * 
     * @param property Property name.
     * @param newValue Value.
     * @param remote Whether to update also remote instances.
     */
    @SuppressWarnings({ "unchecked", "PMD" })
    // im too lazy to refactor this complex method, sorry
    private synchronized void setProperty(String property, Object newValue, boolean remote) {

	if ("children".equals(property) || "id".equals(property)) {
	    throw new IllegalArgumentException("Cannot change property '" + property + "'.");
	}

	// process aliases
	try {
	    Map<String, String> aliases = getChildAliases();
	    if (aliases.get(property.toLowerCase()) != null) {
		property = aliases.get(property.toLowerCase());
	    } else if (aliases.get(property) != null) {
		property = aliases.get(property);
	    }
	} catch (NullPointerException npe) {
	    // ingore
	}

	if ("tags".equals(property)) {

	    if (newValue == null) {
		newValue = new HashSet<String>();
	    } else if (!(newValue instanceof Collection)) {
		throw new IllegalArgumentException(
			"Impossible to overwrite tags property with a non-Collection object.");
	    }

	    // convert to hashset
	    newValue = new HashSet<String>((Collection<String>) newValue);
	}

	// skip unnecessary setting if the value has not changed

	if (!(newValue != null && newValue.equals(getProperty(property)))
		&& !(newValue == null && getProperty(property) == null)) {

	    setPropertyDirectly(property, newValue);

	    if (remote) {
		updateRemoteProperty(property, newValue);
	    }
	}

    }

    /**
     * Assigns a property directly, bypassing the preprocessing.
     * 
     * @param property Name.
     * @param newValue Value.
     */
    protected final synchronized void setPropertyDirectly(String property, Object newValue) {

	// if it is null, delete it
	if (newValue == null) {
	    getProperties().remove(property);
	    return;
	}

	super.setProperty(property, newValue);

	// keep track about parents and kids...

	Set<AbstractModelElement> refs = UTIL.listReferences(this.getProperties());
	Set<AbstractModelElement> myKids = kids.get(this);

	if (myKids == null) {
	    myKids = new HashSet<AbstractModelElement>();
	    kids.put(this, myKids);
	}

	myKids.removeAll(refs);

	for (AbstractModelElement old : new HashSet<AbstractModelElement>(myKids)) {

	    myKids.remove(old);

	    Set<AbstractModelElement> pars = parents.get(old);

	    if (pars != null) {
		pars.remove(this);
	    }
	}

	for (AbstractModelElement r : refs) {

	    myKids.add(r);

	    Set<AbstractModelElement> pars = parents.get(r);

	    if (pars == null) {
		pars = new HashSet<AbstractModelElement>(); // NOPMD
		parents.put(r, pars);
	    }

	    pars.add(this);
	}

	this.myMap.fireNotification(this, IModelObserver.UPDATED, property);

    }

    /**
     * Update a property value on a remote host.
     * 
     * @param property Property name.
     * @param value Value.
     */
    private void updateRemoteProperty(final String property, final Object value) {

	// skip unnecessary remote calls, e.g. if the value has not changed
	// if ((value != null && value.equals(getProperty(property))) || (value
	// == null && getProperty(property) == null)) {
	// return;
	// }

	// locked means that nothing should be done
	if (!isLocked()) {

	    List<Thread> threads = new ArrayList<Thread>();
	    final List<Exception> errors = new ArrayList<Exception>();

	    final AbstractModelElement thisElement = this;

	    // change in parallel on all hosts
	    for (final Host host : createdAt) {

		if (host.isLocalhost()) {
		    continue;
		}

		// update each host in a separate thread
		Thread t = new Thread() {

		    public void run() {
			try {

			    IModelElementWebService client = getWsClient(host);

			    Set<AbstractModelElement> refs = UTIL.listReferences(value, true);

			    // wrap everything that is not there yet
			    Collection<ModelElementWrapper> newElements = UTIL.wrapFreshElements(host, refs);

			    // update the referenced objects
			    client.update(new ArrayList<ModelElementWrapper>(newElements));

			    // finally, set the actual property
			    client.setParameter(thisElement.getId(), new KeyValueWrapper(property, value));

			    markAsCleanOn(host, false);
			} catch (Exception e) {
			    errors.add(e);
			}
		    }
		};

		threads.add(t);
		t.start();
	    }

	    // block until threads have finished
	    for (Thread t : threads) {
		try {
		    t.join();
		} catch (InterruptedException e) {
		}
	    }

	    if (!errors.isEmpty()) {
		throw new G2RemoteException("Cannot update property '" + property + "' on " + errors.size()
			+ " hosts. First error is:", errors.get(0));
	    }
	}
    }

    /**
     * Returns all parents of this element.
     * 
     * @param recursive Whether to be recursive.
     * @return List of parents.
     */
    public List<AbstractModelElement> getParents(boolean recursive) {

	Set<AbstractModelElement> pars = parents.get(this);

	if (pars == null) {
	    return new ArrayList<AbstractModelElement>();
	}

	ArrayList<AbstractModelElement> result = new ArrayList<AbstractModelElement>();

	for (AbstractModelElement p : pars) {
	    Set<AbstractModelElement> ks = kids.get(p);

	    if (ks != null && ks.contains(this)) {
		result.add(p);
		if (recursive) {
		    result.addAll(p.getParents(true));
		}
	    }
	}

	return result;
    }

    /**
     * Returns the property's value, however, performs a lot of alias
     * resolution.
     * 
     * @param property Property name.
     * @return Property value.
     */
    @SuppressWarnings("PMD.OnlyOneReturn")
    private Object getPropertyValue(String property) { // NOPMD to lazy to
	// refactor...

	if ("children".equals(property)) {
	    return getChildren();
	} else if ("properties".equals(property)) {
	    return getProperties();
	} else {

	    property = resolvePropertyNameAlias(property);

	    Object result = getProperties().get(property);

	    // first try own property
	    if (result != null) {
		return result;
	    }

	    // if allowed, try parents' properties
	    if (followUpProps) {

		for (AbstractModelElement p : getParents(false)) {
		    result = p.getProperty(property);

		    if (result != null) {
			return result;
		    }
		}
	    }

	    // or, if property name equals an element type, get all children of
	    // this type
	    Class<? extends AbstractModelElement> type = getClassByShortName(property);

	    if (type != null) {
		Set<AbstractModelElement> kids = getChildren();
		HashSet<AbstractModelElement> resultMap = new HashSet<AbstractModelElement>();

		for (AbstractModelElement k : kids) {
		    if (k.getClass().equals(type)) {
			resultMap.add(k);
		    }
		}
		return resultMap;
	    }
	}

	return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see groovy.util.Expando#getProperty(java.lang.String)
     */
    @Override
    public Object getProperty(String property) {
	Object result = getPropertyValue(property);

	if (result != null) {
	    result = processProperty(property, result);
	}

	return result;
    }

    /**
     * Resolve a property name, if an alias has been defined.
     * 
     * @param property Property name-
     * @return Same or aliased property name.
     */
    protected String resolvePropertyNameAlias(String property) {

	String newProperty = property;

	Map<String, String> aliases = getChildAliases();

	if (aliases != null) {
	    if (aliases.containsKey(property.toLowerCase())) {
		newProperty = aliases.get(property.toLowerCase());
	    } else if (aliases.containsKey(property)) {
		newProperty = aliases.get(property);
	    }
	}

	return newProperty;
    }

    /**
     * Process property values, e.g. by replacing them.
     * 
     * @param propName Property name.
     * @param value Property value.
     * @return Processed valued.
     */
    protected Object processProperty(String propName, Object value) {

	Object result = null;

	try {
	    result = UTIL.cloneAndEnhanceClosures(value, this);
	} catch (Exception e) {
	}

	return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see groovy.util.Expando#equals(java.lang.Object)
     */
    @Override
    public boolean equals(Object obj) {

	if (obj != null && (obj instanceof AbstractModelElement)) {
	    return ((AbstractModelElement) obj).getId().equals(getId());
	}

	return super.equals(obj);
    }

    /*
     * (non-Javadoc)
     * 
     * @see groovy.util.Expando#hashCode()
     */
    @Override
    public final int hashCode() {
	return getId().hashCode();
    }

    /*
     * (non-Javadoc)
     * 
     * @see groovy.util.Expando#toString()
     */
    @Override
    public String toString() {
	return this.getClass().getSimpleName() + getProperties();
    }

    /**
     * Declare a new element.
     * 
     * @param c Element's class.
     * @param plugin Declaring plugin.
     */
    public static void declareElementType(Class<? extends AbstractModelElement> c,
	    AbstractGenesis2Plugin<? extends AbstractModelElement> plugin) {

	synchronized (TYPE_MAP) {

	    if (TYPE_MAP.containsValue(c)) {
		return;
	    }

	    PLUGIN_MAP.put(c, plugin);

	    TYPE_MAP.put(c.getSimpleName(), c);
	    TYPE_MAP.put(c.getSimpleName().toLowerCase(), c);

	    LOGGER.info("Declared new model element type: " + c.getSimpleName());
	}
    }

    /**
     * Undeclare an element again.
     * 
     * @param c Element's class.
     */
    public static void undeclareElementType(Class<? extends AbstractModelElement> c) {

	synchronized (TYPE_MAP) {

	    if (!TYPE_MAP.containsValue(c)) {
		return;
	    }

	    PLUGIN_MAP.remove(c);

	    TYPE_MAP.remove(c.getSimpleName());
	    TYPE_MAP.remove(c.getSimpleName().toLowerCase());

	    ElementMap<?> map = ELEMENTMAP_MAP.remove(c);

	    if (map != null) {
		for (AbstractModelElement element : map.getAll()) {
		    element.cleanup();
		}
	    }

	    LOGGER.info("Undeclared model element type: " + c.getSimpleName());
	}
    }

    /**
     * Returns the class declares with the passed short name.
     * 
     * @param name Short name.
     * 
     * @return Class or <code>null</code> if nothing was found.
     */
    public static Class<? extends AbstractModelElement> getClassByShortName(String name) {
	return TYPE_MAP.get(name.toLowerCase());
    }

    /**
     * Returns the names of declared element types.
     * 
     * @return List of names.
     */
    public static List<String> getDeclaredElementNames() {
	return new ArrayList<String>(TYPE_MAP.keySet());
    }

    /**
     * Check if an element has been declared.
     * 
     * @param c Element's class.
     * 
     * @return <code>null</code> if declared, otherwise <code>false</code>.
     */
    public static boolean isModelTypeDeclared(Class<? extends AbstractModelElement> c) {
	return TYPE_MAP.containsValue(c);
    }

    /**
     * Returns the map containing all elements of the passed type.
     * 
     * @param <S> Type of element.
     * 
     * @param c Model Element type.
     * @return Map.
     * 
     */
    @SuppressWarnings("unchecked")
    public static <S extends AbstractModelElement> ElementMap<S> getElementMap(Class<S> c) {

	if (c == null) {
	    throw new IllegalArgumentException("Class is null.");
	}

	if (!isModelTypeDeclared(c)) {
	    throw new IllegalArgumentException("Model element type '" + c.getSimpleName() + "' has not been declared.");
	}

	ElementMap<S> map = (ElementMap<S>) ELEMENTMAP_MAP.get(c); // is safe

	if (map == null) {
	    map = new ElementMap<S>(c, (AbstractGenesis2Plugin<S>) PLUGIN_MAP.get(c));
	    ELEMENTMAP_MAP.put(c, map);
	}

	return map;
    }

    /**
     * Returns the map containing all elements of the passed type.
     * 
     * @param typeName Type name.
     * @return Map.
     */
    public static ElementMap<? extends AbstractModelElement> getElementMap(String typeName) {

	Class<? extends AbstractModelElement> c = getClassByShortName(typeName);

	if (c == null) {
	    throw new IllegalArgumentException("Model element type '" + typeName + "' has not been declared.");
	}

	return getElementMap(c);
    }

    /**
     * Returns own children.
     * 
     * @return Children.
     */
    public Set<AbstractModelElement> getChildren() {

	HashSet<AbstractModelElement> result = new HashSet<AbstractModelElement>();

	// filter out all orphans
	for (AbstractModelElement kid : UTIL.listReferences(this.getProperties())) {
	    if (kid.isRegistered()) {
		result.add(kid);
	    }
	}

	return result;
    }

    /**
     * Returns all children, with grandchildren, etc..
     * 
     * @return Children.
     */
    public Set<AbstractModelElement> getAllChildren() {

	HashSet<AbstractModelElement> result = new HashSet<AbstractModelElement>();

	for (AbstractModelElement kid : getChildren()) {

	    result.add(kid);
	    result.addAll(kid.getAllChildren());
	}

	return result;
    }

    /**
     * Returns the client for the remote model manipulator web service at the
     * particular host.
     * 
     * @param host Host.
     * @return Client instance.
     */
    protected IModelElementWebService getWsClient(Host host) {

	return getPlugin().getModelWsClient(host.getHost(), host.getPort(), getClass());
    }

    /**
     * Returns the URL string to the (remote) model element web service.
     * 
     * @param h Destination host.
     * 
     * @return URL.
     */
    public String getModelWsURL(Host h) {
	return AbstractGenesis2Plugin.getModelWsURL(h.getHost(), h.getPort(), getClass()) + "/" + getId();
    }

    /**
     * Closure call.
     * 
     * @param c1 Closure.
     * @return Call result.
     */
    public final Object call(Closure c1) {

	Closure c = (Closure) c1.clone();

	ClosureUtil.getInstance().setPrimaryDispatcher(c, this);

	return c.call();
    }

    /**
     * Updates the model element's properties on bound hosts.
     * 
     */
    public synchronized void update() {

	List<Thread> threads = new ArrayList<Thread>();
	final List<RuntimeException> errors = new ArrayList<RuntimeException>();
	final AbstractModelElement thisElement = this;

	for (final Host host : getHostsForUpdate()) {

	    if (host.isLocalhost()) {
		continue;
	    }

	    Thread t = new Thread() { // NOPMD
		public void run() {

		    try {
			List<ModelElementWrapper> wrappedElements = new ArrayList<ModelElementWrapper>();

			for (AbstractModelElement kid : getChildrenForUpdate(host)) {
			    wrappedElements.add(new ModelElementWrapper(kid));
			}

			if (createdAt.contains(host)) {

			    // if it has been already there, just update

			    if (isDirtyOn(host)) {
				wrappedElements.add(new ModelElementWrapper(thisElement));
			    }

			    if (!wrappedElements.isEmpty()) {

				getWsClient(host).update(wrappedElements);

				markAsCreatedAt(host);
				markAsCleanOn(host, true);
			    }

			} else {

			    // otherwise create from scratch

			    getWsClient(host).create(new ModelElementWrapper(thisElement), wrappedElements);

			    markAsCreatedAt(host);
			    markAsCleanOn(host, true);
			}

		    } catch (G2Exception g2e) {
			errors.add(g2e);
		    } catch (Exception e) {
			errors.add(new G2RemoteException(e));
		    }
		}
	    };

	    threads.add(t);
	    t.start();

	}

	// block until threads have finished
	for (Thread t : threads) {
	    try {
		t.join();
	    } catch (InterruptedException e) {
	    }
	}

	if (!errors.isEmpty()) {
	    throw errors.get(0);
	}
    }

    /**
     * Returns a set of child elements, which need to be updatef on the given
     * host.
     * 
     * @param host Host.
     * @return Set of children.
     */
    private Set<AbstractModelElement> getChildrenForUpdate(Host host) {

	Set<AbstractModelElement> elements = new HashSet<AbstractModelElement>();

	for (AbstractModelElement kid : getAllChildren()) {

	    if (kid.isDirtyOn(host)) {
		elements.add(kid);
	    }
	}

	return elements;
    }

    /**
     * Returns a list of hosts which have to be update.
     * 
     * @return List of hosts.
     */
    private List<Host> getHostsForUpdate() {

	int propHash = getProperties().hashCode();

	List<Host> result = new ArrayList<Host>();

	for (Host host : boundTo) {
	    if (isDirtyOn(host, propHash)) {
		result.add(host);
	    }
	}

	return result;
    }

    /**
     * Deploys at all remote hosts where this element has been bound to.
     * 
     */
    public void deploy() {
	deployAt(createdAt);
    }

    /**
     * Deploys at the given hosts.
     * 
     * @param hosts Hosts.
     */
    public void deployAt(Host... hosts) {

	bindTo(hosts);
	update();

	for (Host h : hosts) {
	    if (createdAt.contains(h) && !deployedAt.contains(h)) {
		getWsClient(h).deployElement(getId());
		markAsDeployedAt(h);
	    }
	}
    }

    /**
     * Deploys at the given hosts.
     * 
     * @param hosts Hosts.
     */
    public void deployAt(Collection<Host> hosts) {
	deployAt(hosts.toArray(new Host[] {}));
    }

    /**
     * Redeploys at all remote hosts where this element has been deployed at.
     * 
     */
    public void redeploy() {
	redeployAt(deployedAt);
    }

    /**
     * Redeploys at the given hosts.
     * 
     * @param hosts Hosts.
     */
    public void redeployAt(Host... hosts) {

	for (Host h : hosts) {
	    if (deployedAt.contains(h)) {
		getWsClient(h).redeployElement(getId());
	    }
	}
    }

    /**
     * Redeploys at the given hosts.
     * 
     * @param hosts Hosts.
     */
    public void redeployAt(Collection<Host> hosts) {
	redeployAt(hosts.toArray(new Host[] {}));
    }

    /**
     * Undeploys from all remote hosts where this element has been deployed at.
     * 
     */
    public void undeploy() {
	undeployFrom(deployedAt);
    }

    /**
     * Undeploys from the given hosts.
     * 
     * @param hosts Hosts.
     */
    public void undeployFrom(Host... hosts) {

	for (Host h : hosts) {
	    if (deployedAt.contains(h)) {
		getWsClient(h).undeployElement(getId());
		markAsUndeployedAt(h);
	    }
	}
    }

    /**
     * Undeploys from the given hosts.
     * 
     * @param hosts Hosts.
     */
    public void undeployFrom(Collection<Host> hosts) {
	undeployFrom(hosts.toArray(new Host[] {}));
    }

    /**
     * Checked whether it is locked.
     * 
     * @return <code>true</code> if locked, otherwise <code>false</code>.
     */
    public boolean isLocked() {
	return locked;
    }

    /**
     * Declare as locked for remote updates.
     * 
     */
    public void lock() {
	locked = true;
    }

    /**
     * Declare as unlocked for remote updates.
     * 
     */
    public void unlock() {
	locked = false;
	update();
    }

    /**
     * Declare this element as created at the given host.
     * 
     * @param host Host.
     */
    protected void markAsCreatedAt(Host host) {

	createdAt.add(host);
	boundTo.add(host);

	for (AbstractModelElement kid : getChildren()) {
	    kid.markAsCreatedAt(host);
	}
    }

    /**
     * Declare this element as clean (=non-dirty, =up-to-date) at the given
     * host.
     * 
     * @param host Host.
     */
    protected void markAsCleanOn(Host host, boolean recursive) {
	remotePropHash.put(host, getProperties().hashCode());

	if (recursive) {
	    for (AbstractModelElement kid : getChildren()) {
		kid.markAsCleanOn(host, recursive);
	    }
	}
    }

    /**
     * Returns children property name aliases. Supposed to be overwritten on
     * demand.
     * 
     * @return Map of aliases.
     */
    protected Map<String, String> getChildAliases() {
	return new HashMap<String, String>();
    }

    /**
     * Declares this element as deployed on localhost.
     */
    public void markAsLocallyDeployed() {
	markAsDeployedAt(Host.getLocalhost());
    }

    /**
     * Declares this element as deployed on the given host.
     * 
     * @param host Host.
     */
    public void markAsDeployedAt(Host host) {
	createdAt.add(host);
	boundTo.add(host);
	if (deployedAt.add(host)) {
	    myMap.fireNotification(this, IModelObserver.DEPLOYED, host);
	}
    }

    /**
     * Declares this element as undeployed from the given host.
     */
    protected void markAsUndeployedAt(Host host) {
	if (deployedAt.remove(host)) {
	    myMap.fireNotification(this, IModelObserver.UNDEPLOYED, host);
	}
    }

    /**
     * Declares this element as undeployed from localhost.
     */
    public void markAsLocallyUndeployed() {
	markAsUndeployedAt(Host.getLocalhost());
    }

    /**
     * Checks whether it is deployed on localhost.
     * 
     * @return <code>true</code> if deployed, otherwise <code>false</code>.
     */
    public boolean isLocallyDeployed() {
	return deployedAt.contains(Host.getLocalhost());
    }

    /**
     * Checks whether it is deployed on the given host.
     * 
     * @param host Host.
     * @return <code>true</code> if deployed, otherwise <code>false</code>.
     */
    public boolean isDeployedOn(Host host) {
	return deployedAt.contains(host);
    }

    /**
     * Checks whether this element has been changed since the last update on
     * this host.
     * 
     * @param host Host.
     * 
     * @return <code>true</code> if it has changed, otherwise <code>false</code>
     *         .
     */
    protected boolean isDirtyOn(Host host) {
	return isDirtyOn(host, getProperties().hashCode());
    }

    /**
     * Checks whether this element has been changed since the last update on
     * this host.
     * 
     * @param host Host.
     * @param currentHash Hashhocode to compare with.
     * 
     * @return <code>true</code> if it has changed, otherwise <code>false</code>
     *         .
     */
    protected boolean isDirtyOn(Host host, int currentHash) {

	Integer hash = remotePropHash.get(host);

	return (hash == null || hash.intValue() != currentHash);

    }

    /**
     * Returns a set of hosts where this element has been created at.
     * 
     * @return Set of hosts.
     */
    public Set<Host> getCreatedAt() {
	return createdAt;
    }

    /**
     * Returns a set of hosts where this element has been deployed at.
     * 
     * @return Set of hosts.
     */
    public Set<Host> getDeployedAt() {
	return deployedAt;
    }

    /**
     * Returns a set of hosts where this element has been bound to.
     * 
     * @return Set of hosts.
     */
    public Set<Host> getBoundTo() {
	return boundTo;
    }

}
