package at.ac.tuwien.vitalab.genesis2.core.model;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import lombok.Data;

import org.codehaus.groovy.runtime.InvokerHelper;

import at.ac.tuwien.vitalab.genesis2.base.ExecutionEnvironment;

/**
 * Intermediary component for dispatching variable access in closures.
 * 
 * @author Lukasz Juszczyk
 * 
 */
@Data
public class ClosureDelegateDispatcher extends ExecutionEnvironment {

    protected static final String REFLECTION_KW = "my";

    protected Map<String, AbstractModelElement> reflectionMap = new ReflectionMap(this);

    protected Map<String, Object> constants = new HashMap<String, Object>();

    protected Object primaryDispatcher = null;

    /**
     * Constructor.
     */
    public ClosureDelegateDispatcher() {
	super();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.base.ExecutionEnvironment#getProperty(java
     * .lang.String)
     */
    @Override
    public Object getProperty(String property) {

	Object result = interceptGet(property);

	if (result == null) {
	    result = super.getProperty(property);
	}

	return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.base.ExecutionEnvironment#getVariable(java
     * .lang.String)
     */
    @Override
    public Object getVariable(String name) {

	Object result = interceptGet(name);

	if (result == null) {
	    result = super.getVariable(name);
	}

	return result;
    }

    /**
     * Intercept getter.
     * 
     * @param name Property name.
     * @return Value.
     */
    private Object interceptGet(String name) {

	if (constants.containsKey(name)) {
	    return constants.get(name);
	}

	if (primaryDispatcher != null) {
	    Object result = InvokerHelper.getProperty(primaryDispatcher, name);

	    if (result != null) {
		return result;
	    }
	}

	if (name.equals(REFLECTION_KW)) {
	    return reflectionMap;
	}

	return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.base.ExecutionEnvironment#setProperty(java
     * .lang.String, java.lang.Object)
     */
    @Override
    public void setProperty(String property, Object newValue) {

	if (interceptSet(property, newValue)) {
	    return;
	}

	super.setProperty(property, newValue);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.base.ExecutionEnvironment#setVariable(java
     * .lang.String, java.lang.Object)
     */
    @Override
    public void setVariable(String name, Object value) {

	if (interceptSet(name, value)) {
	    return;
	}

	super.setVariable(name, value);
    }

    /**
     * Intercept setter.
     * 
     * @param name Property name.
     * @param value Value.
     * 
     * @return Whether the setting has been intercepted.
     */
    private boolean interceptSet(String name, Object value) {

	boolean intercepted = false;

	if (primaryDispatcher != null) {
	    InvokerHelper.setProperty(primaryDispatcher, name, value);
	    intercepted = true;
	}

	return intercepted;
    }

    private static class ReflectionMap extends HashMap<String, AbstractModelElement> {

	/**
	 * Serial ID.
	 */
	private static final long serialVersionUID = -1878698344853210601L;

	private final ExecutionEnvironment env;

	/**
	 * Constructor.
	 * 
	 * @param env Basic execution environment.
	 */
	public ReflectionMap(ExecutionEnvironment env) {
	    super();
	    this.env = env;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.util.HashMap#get(java.lang.Object)
	 */
	@Override
	public AbstractModelElement get(Object key) {

	    // first try directy registered
	    AbstractModelElement result = super.get(key);

	    if (result != null) {
		return result;
	    }

	    // now try direct interceptors
	    result = searchListForElement(env.getVarBindingInterceptors(), key);

	    if (result != null) {
		return result;
	    }

	    // eventuallty, try parents of direct interceptors
	    List<Object> parents = new ArrayList<Object>();

	    for (Object interceptor : env.getVarBindingInterceptors()) {

		try {
		    AbstractModelElement ame = (AbstractModelElement) interceptor;

		    parents.addAll(ame.getParents(true));
		} catch (ClassCastException cce) {
		    // ignore
		}
	    }

	    return searchListForElement(parents, key);
	}

	/**
	 * Searches a list for elements of a given type.
	 * 
	 * @param list List.
	 * @param name Type name.
	 * @return First found element.
	 */
	private AbstractModelElement searchListForElement(Collection<Object> list, Object name) {

	    for (Object interceptor : list) {

		if (interceptor instanceof AbstractModelElement) {
		    AbstractModelElement ame = (AbstractModelElement) interceptor;

		    if (ame.getClass().getSimpleName().equalsIgnoreCase(name.toString())) {
			return ame;
		    }
		}
	    }

	    return null;
	}
    }

}
