package at.ac.tuwien.vitalab.genesis2.core.model;

import groovy.lang.Closure;
import groovy.lang.GroovyObjectSupport;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import lombok.Getter;

import org.apache.log4j.Logger;
import org.codehaus.groovy.runtime.InvokerHelper;

import at.ac.tuwien.vitalab.genesis2.core.model.builder.AbstractModelBuilder;
import at.ac.tuwien.vitalab.genesis2.core.model.types.Host;
import at.ac.tuwien.vitalab.genesis2.core.model.ws.IModelElementWebService;
import at.ac.tuwien.vitalab.genesis2.core.model.ws.wrappers.ClosureWrapper;
import at.ac.tuwien.vitalab.genesis2.core.model.ws.wrappers.MapWrapper;
import at.ac.tuwien.vitalab.genesis2.core.model.ws.wrappers.ModelElementReferenceWrapper.ModelElementReferenceReplacer;
import at.ac.tuwien.vitalab.genesis2.core.model.ws.wrappers.ModelElementWrapper;
import at.ac.tuwien.vitalab.genesis2.plugin.AbstractGenesis2Plugin;
import at.ac.tuwien.vitalab.genesis2.utils.ClosureCollectionIterator;
import at.ac.tuwien.vitalab.genesis2.utils.ClosureUtil;
import at.ac.tuwien.vitalab.genesis2.utils.ModelElementUtil;

/**
 * Map of model elements.
 * 
 * @author Lukasz Juszczyk
 * 
 * @param <T> Type of contained elements.
 * 
 */
public class ElementMap<T extends AbstractModelElement> extends GroovyObjectSupport {

    private static final Logger LOGGER = Logger.getLogger(ElementMap.class);

    /**
     * This map's element type.
     */
    @Getter
    private Class<T> elementClass = null;

    /**
     * The plugin provinding this element type.
     */
    @Getter
    private AbstractGenesis2Plugin<T> plugin = null;

    /**
     * Content of this type.
     */
    private final Map<String, T> elements = new HashMap<String, T>();

    /**
     * All registered elements.
     */
    private static Map<String, AbstractModelElement> allElements = new HashMap<String, AbstractModelElement>();

    /**
     * Observers.
     */
    private final Set<IModelObserver<T>> observers = new HashSet<IModelObserver<T>>();

    /**
     * Util for model elements.
     */
    private final static ModelElementUtil MODEL_UTIL = ModelElementUtil.getInstance();

    /**
     * Util for preparing closure objects.
     */
    private final static ClosureUtil CLOSURE_UTIL = ClosureUtil.getInstance();

    /**
     * Constructor.
     * 
     * @param elClass Class for items of this map.
     * @param plugin Plugin which declares this element type.
     */
    protected ElementMap(Class<T> elClass, AbstractGenesis2Plugin<T> plugin) {

	super();

	if (elClass == null || plugin == null) {
	    throw new IllegalArgumentException("Element class and plugin cannot be null.");
	}

	this.elementClass = elClass;
	this.plugin = plugin;
    }

    /**
     * Create a new instance of this map's element type.
     * 
     * @return Instance.
     */
    public synchronized T create() {
	return create(new Object[] {});
    }

    /**
     * Create a new instance of this map's element type.
     * 
     * @param args Constructur arguments.
     * @return Instance.
     */
    @SuppressWarnings("unchecked")
    public synchronized T create(Object... args) {

	T newEl = null;

	try {
	    newEl = (T) InvokerHelper.invokeStaticMethod(elementClass, "create", args);
	} catch (Exception e) {
	    newEl = (T) InvokerHelper.invokeConstructorOf(elementClass, args);
	}

	return registerIfNew(newEl);
    }

    /**
     * Create a new instance of this map's element type.
     * 
     * @param props Properties.
     * @return Instance.
     * 
     */
    public synchronized T create(Map<?, ?> props) {
	return create(props, new Object[] {});
    }

    /**
     * Create a new instance of this map's element type.
     * 
     * @param props Properties.
     * @param args Constructur arguments.
     * @return Instance.
     */
    @SuppressWarnings("unchecked")
    public synchronized T create(Map<?, ?> props, Object... args) {

	T newEl = null;

	try {
	    newEl = (T) InvokerHelper.invokeStaticMethod(elementClass, "create", args);
	} catch (Exception e) {
	    newEl = (T) InvokerHelper.invokeConstructorOf(elementClass, args);
	}

	// set all given properties
	for (Object key : props.keySet()) {

	    if (key.equals("id")) {
		newEl.setPropertyDirectly(key.toString(), props.get(key));
	    } else {
		newEl.setProperty(key.toString(), props.get(key));
	    }
	}

	return registerIfNew(newEl);
    }

    /**
     * Register in map if no element has been registered before.
     * 
     * @param newEl Element.
     * 
     * @return Current element in map. Old one if that existed before, otherwise
     *         the passed element.
     */
    protected synchronized T registerIfNew(T newEl) {

	T currentEl = elements.get(newEl.getId());

	if (!newEl.equals(currentEl)) {
	    elements.put(newEl.getId(), newEl);
	    allElements.put(newEl.getId(), newEl);

	    LOGGER.debug("Registered new model element in map: " + newEl);

	    for (IModelObserver<T> mo : observers) {
		mo.onChangedModelElement(newEl, IModelObserver.CREATED);
	    }

	    currentEl = newEl;
	}

	return currentEl;
    }

    /**
     * Deregister from map.
     * 
     * @param ame Element.
     */
    protected synchronized void deregister(T ame) {

	if (ame.equals(elements.remove(ame.getId()))) {
	    LOGGER.debug("Unregistered model element from map: " + ame);

	    allElements.remove(ame.getId());
	    ame.cleanup();

	    fireNotification(ame, IModelObserver.REMOVED);
	}
    }

    /**
     * Register an observer.
     * 
     * @param mo Observer instance.
     */
    public synchronized void observe(IModelObserver<T> mo) {

	observers.add(mo);

	for (T t : getAll()) {
	    mo.onChangedModelElement(t, IModelObserver.CREATED);
	}
    }

    /**
     * Deregister an observer.
     * 
     * @param mo Observer instance.
     */
    public synchronized void stopObserving(IModelObserver<T> mo) {

	observers.remove(mo);
    }

    /**
     * Fires notifications to observing object.
     * 
     * @param t Element.
     * @param how What happened?
     * @param info Additional data.
     */
    public void fireNotification(T t, int how, Object... info) {
	for (IModelObserver<T> mo : observers) {
	    mo.onChangedModelElement(t, how, info);
	}
    }

    /**
     * Get items via a closure predicate.
     * 
     * @param c Closure.
     * @return Item collection.
     */
    @SuppressWarnings("unchecked")
    public Collection<T> get(Closure c) {

	return ClosureCollectionIterator.grep(elements.values(), c); // safe
    }

    /**
     * Checks whether the map contains an element with this key.
     * 
     * @param id ID key.
     * 
     * @return <code>true<code> if found, otherwise <code>false</code>.
     */
    public boolean has(String id) {
	return elements.containsKey(id);
    }

    /**
     * Returns an item by its ID key.
     * 
     * @param id ID key.
     * @return Element or <code>null</code> if nothing was found.
     */
    public T get(String id) {
	return elements.get(id);
    }

    /**
     * Returns the wanted element.
     * 
     * @param id Element id.
     * @return Element instance or <code>null</code> if nothing found.
     */
    public static AbstractModelElement getById(String id) {
	return allElements.get(id);
    }

    /**
     * Get all items.
     * 
     * @return Item collection.
     */
    @SuppressWarnings("unchecked")
    public ArrayList<T> getAll() {
	return new ArrayList<T>(elements.values());
    }

    /**
     * Applies a closure to all items.
     * 
     * @param c Closure.
     * 
     * @return Collection of results.
     */
    public Object forAll(Closure c) {
	return ClosureCollectionIterator.each(getAll(), c);
    }

    /**
     * Call Web service of remote element(s).
     * 
     * @param vars Variables which should be sent to the remote hosts.
     * @param filterClosure Filter closure.
     * @param commandClosure Command closure.
     * 
     * @return Result of local execution.
     */
    public final Object call(final Map<String, Object> vars, final Closure filterClosure, final Closure commandClosure) {

	Collection<T> relevantElements = get(filterClosure);

	if (relevantElements == null || relevantElements.isEmpty()) {
	    return null;
	}

	Closure clonedCommand = (Closure) commandClosure.clone();

	CLOSURE_UTIL.registerConstants(clonedCommand, vars);

	// local result, will be returned
	Object result = ClosureCollectionIterator.each(relevantElements, clonedCommand);

	HashSet<Host> destHosts = new HashSet<Host>();

	for (T t : relevantElements) {
	    destHosts.addAll(t.getCreatedAt());
	}

	Vector<Thread> threads = new Vector<Thread>();

	final AbstractGenesis2Plugin<?> fPlugin = plugin;
	final Class<? extends AbstractModelElement> fType = elementClass;

	final Set<AbstractModelElement> refs = MODEL_UTIL.listReferences(vars);

	for (final Host destHost : destHosts) {
	    Thread t = new Thread() {

		/*
		 * (non-Javadoc)
		 * 
		 * @see java.lang.Thread#run()
		 */
		@Override
		public void run() {

		    try {

			IModelElementWebService client = fPlugin.getModelWsClient(destHost.getHost(),
				destHost.getPort(), fType);

			Collection<ModelElementWrapper> newElements = MODEL_UTIL.wrapFreshElements(destHost, refs);

			client.update(new Vector<ModelElementWrapper>(newElements));

			client.change(new MapWrapper(vars, new ModelElementReferenceReplacer()), new ClosureWrapper(
				filterClosure), new ClosureWrapper(commandClosure));

		    } catch (Exception e) {
			LOGGER.error("Cannot modify model element on host " + destHost.getLocation() + ": "
				+ e.getMessage());
		    }
		}
	    };
	    threads.add(t);
	    t.start();
	}

	// block until threads have finished
	for (Thread t : threads) {
	    try {
		t.join();
	    } catch (InterruptedException e) {
	    }
	}

	return result;
    }

    /**
     * Call Web service of remote element(s).
     * 
     * @param filterClosure Filter closure.
     * @param commandClosure Command closure.
     * 
     * @return Result of local execution.
     */
    public final Object call(Closure filterClosure, Closure commandClosure) {
	return call(new HashMap<String, Object>(), filterClosure, commandClosure);
    }

    public static class VariableMap extends GroovyObjectSupport {

	HashMap<Object, Object> vars = new HashMap<Object, Object>();

	/**
	 * @param vars
	 */
	public VariableMap(Map<?, ?> vars) {
	    super();
	    if (vars != null) {
		this.vars.putAll(vars);
	    }
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see groovy.lang.GroovyObjectSupport#getProperty(java.lang.String)
	 */
	@Override
	public Object getProperty(String property) {

	    return vars.get(property);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see groovy.lang.GroovyObjectSupport#setProperty(java.lang.String,
	 * java.lang.Object)
	 */
	@Override
	public void setProperty(String property, Object newValue) {
	    this.vars.put(property, newValue);
	}

    }

    /**
     * Calls the builder.
     * 
     * @param c Builder closure.
     * @return List if built elements.
     */
    public ArrayList<T> build(Closure c) {

	try {
	    return (ArrayList<T>) InvokerHelper.invokeStaticMethod(elementClass, "build", new Object[] { c });
	} catch (Exception e) {
	    // try default
	}

	AbstractModelBuilder<T> builder;
	try {
	    builder = AbstractModelBuilder.getBuilder(elementClass);
	} catch (Exception e) {
	    throw new IllegalArgumentException("Cannot create builder for type '" + elementClass.getName() + "': "
		    + e.getMessage());
	}

	c.setDelegate(builder);
	c.call();

	return builder.getResultingModelElements();
    }

    /*
     * (non-Javadoc)
     * 
     * @see groovy.lang.GroovyObject#invokeMethod(java.lang.String,
     * java.lang.Object)
     */
    public Object invokeMethod(String name, Object args) {

	return InvokerHelper.invokeStaticMethod(elementClass, name, args);
    }
}
