package at.ac.tuwien.vitalab.genesis2.core.model.builder;

import groovy.lang.Closure;
import groovy.lang.MissingMethodException;
import groovy.util.BuilderSupport;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.codehaus.groovy.runtime.InvokerHelper;

import at.ac.tuwien.vitalab.genesis2.base.ExecutionEnvironment;
import at.ac.tuwien.vitalab.genesis2.core.model.AbstractModelElement;
import at.ac.tuwien.vitalab.genesis2.core.model.ElementMap;
import at.ac.tuwien.vitalab.genesis2.utils.ClosureUtil;

/**
 * Abtract class for model builder utility.
 * 
 * @author Lukasz Juszczyk
 * 
 * @param <T> Resulting type which should be built.
 * 
 */
public abstract class AbstractModelBuilder<T extends AbstractModelElement> extends BuilderSupport {

    // registry maps
    private final static Map<Class<? extends AbstractModelElement>, Class<? extends AbstractModelBuilder<? extends AbstractModelElement>>> BUILDERS = new HashMap<Class<? extends AbstractModelElement>, Class<? extends AbstractModelBuilder<? extends AbstractModelElement>>>();

    private final static Map<Class<? extends AbstractModelBuilder<? extends AbstractModelElement>>, Class<? extends AbstractModelElement>> BUILDERTYPES = new HashMap<Class<? extends AbstractModelBuilder<? extends AbstractModelElement>>, Class<? extends AbstractModelElement>>();

    // next type in builder recursion
    protected Class<? extends AbstractModelElement> nextType = null;

    // currently built element
    protected T currentElement = null;

    // child elements
    protected List<AbstractModelElement> kids = new ArrayList<AbstractModelElement>();

    // built elements
    protected ArrayList<T> result = new ArrayList<T>();

    // util for manipulating closures
    protected static final ClosureUtil CLOSURE_UTIL = ClosureUtil.getInstance();

    /**
     * Register a builder.
     * 
     * @param typeClass Type class.
     * @param builderClass Builder class.
     */
    public static void register(Class<? extends AbstractModelElement> typeClass,
	    Class<? extends AbstractModelBuilder<? extends AbstractModelElement>> builderClass) {

	if (builderClass == null) {
	    deregister(typeClass);
	}

	BUILDERS.put(typeClass, builderClass);
	BUILDERTYPES.put(builderClass, typeClass);
    }

    /**
     * Unregister a builder.
     * 
     * @param type Type of builder element.
     */
    public static void deregister(Class<? extends AbstractModelElement> type) {
	BUILDERTYPES.remove(BUILDERS.remove(type));
    }

    /**
     * Creates an instance of the builder for the passed element type.
     * 
     * @param <S> Element type.
     * 
     * @param type Element type class.
     * 
     * @return Builder instance.
     * 
     * @throws InstantiationException If builder cannot be created.
     * @throws IllegalAccessException If builder class cannot be accessed.
     */
    @SuppressWarnings("unchecked")
    public static <S extends AbstractModelElement> AbstractModelBuilder<S> getBuilder(Class<S> type)
	    throws InstantiationException, IllegalAccessException {

	Class<? extends AbstractModelBuilder<S>> builderClass = (Class<? extends AbstractModelBuilder<S>>) BUILDERS
		.get(type); // safe

	if (builderClass == null) {
	    throw new IllegalArgumentException("No builder registered for type '" + type.getName() + "'.");
	}

	return builderClass.newInstance();
    }

    /**
     * Returns element type.
     * 
     * @return Element type class.
     */
    @SuppressWarnings("unchecked")
    protected Class<T> getType() {
	return (Class<T>) BUILDERTYPES.get(getClass());
    }

    /**
     * Attaches an orphaned child. Orphaned means, that it has not been assigned
     * to a variable/property.
     * 
     * @param parent Parent instance.
     * @param child Child instance.
     */
    protected abstract void attachOrphanedChild(T parent, AbstractModelElement child);

    /**
     * Determines the type of the child, based on the name and attributes of the
     * called method.
     * 
     * @param name Operation name.
     * @param value Argument value.
     * @param attributes Argument map.
     */
    protected abstract Class<? extends AbstractModelElement> nextChildType(Object name, Object value,
	    Map<?, ?> attributes);

    // element map for this type
    private final ElementMap<T> getMap() {
	return AbstractModelElement.getElementMap(getType());
    }

    /*
     * (non-Javadoc)
     * 
     * @see groovy.util.BuilderSupport#createNode(java.lang.Object)
     */
    @Override
    protected T createNode(Object name) {

	T t = getMap().create(name);
	nextType = nextChildType(name, null, null);

	currentElement = t;

	return t;
    }

    /*
     * (non-Javadoc)
     * 
     * @see groovy.util.BuilderSupport#createNode(java.lang.Object,
     * java.lang.Object)
     */
    @Override
    protected T createNode(Object name, Object value) {

	T t = getMap().create(name, value);
	nextType = nextChildType(name, value, null);

	currentElement = t;

	return t;
    }

    /*
     * (non-Javadoc)
     * 
     * @see groovy.util.BuilderSupport#createNode(java.lang.Object,
     * java.util.Map)
     */
    @Override
    @SuppressWarnings("rawtypes")
    protected T createNode(Object name, Map attributes) {

	T t = getMap().create(attributes, name);
	nextType = nextChildType(name, null, attributes);

	currentElement = t;

	return t;
    }

    /*
     * (non-Javadoc)
     * 
     * @see groovy.util.BuilderSupport#createNode(java.lang.Object,
     * java.util.Map, java.lang.Object)
     */
    @Override
    @SuppressWarnings("rawtypes")
    protected T createNode(Object name, Map attributes, Object value) {

	T t = getMap().create(attributes, name, value);
	nextType = nextChildType(name, value, attributes);

	currentElement = t;

	return t;
    }

    /*
     * (non-Javadoc)
     * 
     * @see groovy.util.BuilderSupport#invokeMethod(java.lang.String,
     * java.lang.Object)
     */
    public Object invokeMethod(String methodName, Object args) {

	try {
	    return ExecutionEnvironment.getGlobalInstance().invokeMethod(methodName, args);

	} catch (MissingMethodException mme) {
	    // ignore and go on
	}

	if (nextType != null) {

	    // delegate to child builder
	    try {
		AbstractModelElement res = (AbstractModelElement) getBuilder(nextType).invokeMethod(methodName, args);

		kids.add(res);

		return res;

	    } catch (Exception e) {
		throw new IllegalArgumentException(e.getMessage());
	    }
	}

	// or handle here
	Object name = getName(methodName);

	// (pre)process arguments
	List<?> procArgs = processArgs(InvokerHelper.asList(args));

	return doInvokeMethod(methodName, name, procArgs);
    }

    /**
     * Argument preprocessing.
     * 
     * @param args Input.
     * @return Output.
     */
    protected List<?> processArgs(List<?> args) {
	// by default, don't change
	return args;
    }

    /*
     * (non-Javadoc)
     * 
     * @see groovy.util.BuilderSupport#postNodeCompletion(java.lang.Object,
     * java.lang.Object)
     */
    @SuppressWarnings("unchecked")
    @Override
    protected Object postNodeCompletion(Object parent, Object node) {

	result.add((T) node); // should be safe

	kids.removeAll(currentElement.getAllChildren());

	// attach kids which have not been bound to variables/props
	for (AbstractModelElement kid : kids) {
	    attachOrphanedChild(currentElement, kid);
	}

	nextType = null;
	kids.clear();

	return super.postNodeCompletion(parent, node);
    }

    /**
     * Get result.
     * 
     * @return Collection of built elements.
     */
    public ArrayList<T> getResultingModelElements() {
	return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see groovy.util.BuilderSupport#setClosureDelegate(groovy.lang.Closure,
     * java.lang.Object)
     */
    @Override
    protected void setClosureDelegate(Closure closure, Object node) {

	try {
	    CLOSURE_UTIL.setPrimaryDispatcher(closure, (AbstractModelElement) node);
	} catch (ClassCastException cce) {
	    // should not happen
	}

	CLOSURE_UTIL.attachDelegates(closure, ExecutionEnvironment.getGlobalInstance());
    }

    /*
     * (non-Javadoc)
     * 
     * @see groovy.util.BuilderSupport#setParent(java.lang.Object,
     * java.lang.Object)
     */
    @Override
    protected void setParent(Object parent, Object child) { // NOPMD
    }

    /*
     * (non-Javadoc)
     * 
     * @see groovy.lang.GroovyObjectSupport#getProperty(java.lang.String)
     */
    @Override
    public Object getProperty(String property) {
	return ExecutionEnvironment.getGlobalInstance().getVariable(property);
    }

}
