package at.ac.tuwien.vitalab.genesis2.core.model.types;

import groovy.lang.Closure;

import java.io.StringWriter;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.List;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.Marshaller;
import javax.xml.bind.annotation.XmlType;
import javax.xml.namespace.QName;

import lombok.Setter;

import org.codehaus.groovy.runtime.InvokerHelper;

import at.ac.tuwien.vitalab.genesis2.errors.G2Exception;
import at.ac.tuwien.vitalab.genesis2.utils.ClosureUtil;

/**
 * Abstract class for data type instances. Used for copy assignments.
 * 
 * @author Lukasz Juszczyk
 * 
 */
@XmlType(name = "DataTypeInstance")
public abstract class ADataTypeInstance {

    @Setter
    private String xsdNamespace = null;

    private static final ClosureUtil CLOSURE_UTIL = ClosureUtil.getInstance();

    /**
     * Creates a new element/field in an XML datatype.
     * 
     * @param name Field name.
     * @return Field instance.
     */
    public Object newField(String name) {
	return newField(this, name);
    }

    /**
     * Creates a new element/field in an XML datatype.
     * 
     * @param o XML object.
     * @param name Field name.
     * @return Field instance.
     */
    public static Object newField(Object o, String name) {

	Method[] methods = o.getClass().getDeclaredMethods();

	for (Method m : methods) {
	    Class<?>[] parTypes = m.getParameterTypes();

	    if (parTypes.length == 1 && m.getName().toLowerCase().equals("set" + name.toLowerCase())) {
		Class<?> parType = parTypes[0];

		Object result;
		try {
		    result = parType.newInstance();
		    m.invoke(o, result);
		} catch (Exception e) {
		    throw new G2Exception("Cannot instantiate or invoke setter on " + o.getClass().getSimpleName()
			    + ".");
		}
		return result;
	    }
	}

	throw new IllegalArgumentException("Unknown property '" + name + "'.");
    }

    /**
     * Creates an list item.
     * 
     * @param name
     * @return List instance.
     */
    public Object newList(String name) {

	Field field;
	try {
	    field = getClass().getDeclaredField(name);

	    Class<?> itemType = null;

	    for (Class<?> c : getClass().getDeclaredClasses()) {
		if (c.getName().toLowerCase()
			.equals(this.getClass().getName().toLowerCase() + "$" + field.getName().toLowerCase())) {
		    itemType = c;
		}
	    }
	    if (itemType != null) {
		try {
		    return itemType.newInstance();
		} catch (Exception e) {
		    throw new G2Exception("Cannot instantiate " + itemType.getClass().getSimpleName() + ".");
		}
	    }

	} catch (NoSuchFieldException e) {
	    // go on with the exception
	}

	throw new IllegalArgumentException("Unknown list property '" + name + "'.");
    }

    /**
     * Copies all fields from the passed object.
     * 
     * @param from Source.
     * @return This object.
     */
    public Object assign(Object from) {
	assignFromTo(from, this);
	return this;
    }

    /**
     * Copies all fields.
     * 
     * @param from Source.
     * @param to Destination.
     */
    public static void assignFromTo(Object from, Object to) {

	// try to fill every field, by taking the value from the corresponding
	// field of the source
	for (Field field : to.getClass().getDeclaredFields()) {

	    try {
		Class<?> fieldClass = field.getType();

		Object srcValue = null;
		try {
		    srcValue = InvokerHelper.getProperty(from, field.getName());
		} catch (Exception e) {
		    // ignore and go on with next iteration
		    continue;
		}

		// for lists, we need to find out the type of the items,
		// create new ones and assign their content from the
		// corresponding source objects
		if (List.class.isAssignableFrom(fieldClass)) {

		    if (srcValue == null) {
			field.set(to, null);
			continue;
		    }

		    try {

			List<?> fromList = (List<?>) srcValue;
			List<Object> toList = null;

			try {

			    toList = (List<Object>) InvokerHelper.getProperty(to, field.getName());
			} catch (Exception e) {
			    // ignore and go on with next iteration
			    continue;
			}

			Class<?> itemType = null;

			for (Class<?> c : to.getClass().getDeclaredClasses()) {
			    if (c.getName()
				    .toLowerCase()
				    .equals(to.getClass().getName().toLowerCase() + "$" + field.getName().toLowerCase())) {
				itemType = c;
			    }
			}

			// clear in any case
			toList.clear();

			// and try to fill up with new values
			if (itemType != null) {
			    for (Object o : fromList) {
				Object p = itemType.newInstance();

				assignFromTo(o, p);
				toList.add(p);
			    }
			}

		    } catch (Exception e) {
			// ignore
		    }
		} else {
		    try {
			InvokerHelper.setProperty(to, field.getName(), srcValue);
		    } catch (Exception e) {

			try {
			    Object o = newField(to, field.getName());

			    if (o instanceof ADataTypeInstance) {
				((ADataTypeInstance) o).assign(srcValue);
			    }

			    InvokerHelper.setProperty(to, field.getName(), o);

			} catch (Exception e2) {
			    // ignore
			}
		    }
		}
	    } catch (Exception e) {
		// go on with next field
	    }
	}
    }

    /**
     * Closure calling method for collecting manipulation requests in a single
     * closure.
     * 
     * @param c Closure.
     * @return Result of closure call.
     */
    public Object call(Closure c) {

	Closure clone = (Closure) c.clone();

	CLOSURE_UTIL.setPrimaryDispatcher(clone, this);

	return clone.call();
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Object#toString()
     */
    public String toString() {
	try {
	    // print out XML
	    JAXBContext jContext = JAXBContext.newInstance(getClass());
	    Marshaller marshaller = jContext.createMarshaller();
	    QName qn = null;
	    if (xsdNamespace != null) {
		qn = new QName(xsdNamespace, getClass().getSimpleName());
	    } else {
		qn = new QName(getClass().getSimpleName());
	    }
	    JAXBElement<?> el = new JAXBElement(qn, getClass(), this);
	    StringWriter writer = new StringWriter();
	    marshaller.marshal(el, writer);
	    return writer.toString();
	} catch (Exception e) {
	    return super.toString();
	}
    }
}
