package at.ac.tuwien.vitalab.genesis2.core.model.types;

import java.io.File;
import java.lang.reflect.Method;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Random;

import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.Duration;
import javax.xml.datatype.XMLGregorianCalendar;
import javax.xml.namespace.QName;

import org.codehaus.groovy.runtime.InvokerHelper;

import at.ac.tuwien.vitalab.genesis2.core.model.AbstractModelElement;
import at.ac.tuwien.vitalab.genesis2.core.model.ICleanable;
import at.ac.tuwien.vitalab.genesis2.errors.G2Exception;
import at.ac.tuwien.vitalab.genesis2.errors.G2LocalException;
import at.ac.tuwien.vitalab.genesis2.utils.G2UrlClassLoader;
import at.ac.tuwien.vitalab.genesis2.utils.Repository;

/**
 * Internal data type representation.
 * 
 * @author Lukasz Juszczyk
 * 
 */
public class DataType extends AbstractModelElement implements ICleanable {

    protected static final List<String> MODES = new ArrayList<String>();

    private static final Hashtable<String, Class<?>> SIMPLETYPE_MAP = new Hashtable<String, Class<?>>();

    private static final Map<Class<?>, DataType> SIMPLETYPES = new HashMap<Class<?>, DataType>();
    private static final Map<DataType, Map<String, DataType>> HEADERS = new HashMap<DataType, Map<String, DataType>>();
    private static final Map<DataType, DataType> ARRAYS = new HashMap<DataType, DataType>();

    // compiled Java class
    private Class<?> javaClass = null;

    private boolean array = false;

    // local repository directory
    private final File localRepoDir = null;

    private static File classRepoDir = null;

    /**
     * Classloader for all datatypes.
     */
    public final static G2UrlClassLoader LOADER = new G2UrlClassLoader();

    private static final Repository REPOSITORY = Repository.getInstance();

    static {

	SIMPLETYPE_MAP.put("string", String.class);
	SIMPLETYPE_MAP.put("integer", BigInteger.class);
	SIMPLETYPE_MAP.put("int", Integer.class);
	SIMPLETYPE_MAP.put("long", Long.class);
	SIMPLETYPE_MAP.put("short", Short.class);
	SIMPLETYPE_MAP.put("decimal", BigDecimal.class);
	SIMPLETYPE_MAP.put("float", Float.class);
	SIMPLETYPE_MAP.put("double", Double.class);
	SIMPLETYPE_MAP.put("boolean", Boolean.class);
	SIMPLETYPE_MAP.put("byte", Byte.class);
	SIMPLETYPE_MAP.put("QName", QName.class);
	SIMPLETYPE_MAP.put("datetime", XMLGregorianCalendar.class);
	SIMPLETYPE_MAP.put("base64binary", byte[].class);
	SIMPLETYPE_MAP.put("hexbinary", byte[].class);

	SIMPLETYPE_MAP.put("unsignedInt", Long.class);
	SIMPLETYPE_MAP.put("unsignedShort", Integer.class);
	SIMPLETYPE_MAP.put("unsignedByte", Short.class);

	SIMPLETYPE_MAP.put("time", XMLGregorianCalendar.class);
	SIMPLETYPE_MAP.put("date", XMLGregorianCalendar.class);

	SIMPLETYPE_MAP.put("anysimpletype", String.class);

	SIMPLETYPE_MAP.put("duration", Duration.class);
	SIMPLETYPE_MAP.put("NOTATION", QName.class);

	SIMPLETYPE_MAP.put("void", void.class);

	// simpleTypes.put("fault", Fault.class);

	for (Class<?> cl : new ArrayList<Class<?>>(SIMPLETYPE_MAP.values())) {
	    SIMPLETYPE_MAP.put(cl.getName(), cl);
	    SIMPLETYPE_MAP.put(cl.getName().toLowerCase(), cl);
	    SIMPLETYPE_MAP.put(cl.getSimpleName(), cl);
	    SIMPLETYPE_MAP.put(cl.getSimpleName().toLowerCase(), cl);
	}

	MODES.add("IN");
	MODES.add("OUT");
	MODES.add("INOUT");
    }

    /**
     * Constructor.
     */
    public DataType() {
	super();
    }

    /**
     * Constructor.
     * 
     * @param name Name of complexType of element inside this XSD.
     * @param xsdFileNames XSD file names.
     */
    public DataType(String name, String... xsdFileNames) {

	super();

	File[] files = new File[xsdFileNames.length];
	int i = 0;
	for (String fn : xsdFileNames) {
	    files[i++] = new File(fn);
	}

	init(name, files);
    }

    /**
     * Constructor.
     * 
     * @param name Name of complexType of element inside this XSD.
     * @param xsdFiles XSD files.
     */
    public DataType(String name, File... xsdFiles) {

	super();

	init(name, xsdFiles);
    }

    private void init(String name, File... xsdFiles) {
	if (xsdFiles == null || xsdFiles.length == 0 || name == null) {
	    throw new IllegalArgumentException("Arguments not valid: null.");
	}

	try {
	    XSD xsd = new XSD(xsdFiles);

	    setProperty("xsd", xsd);
	    setProperty("name", name);

	    // just for checking
	    xsd.getJavaClass(name);

	} catch (Exception e) {
	    throw new G2Exception(e);
	}
    }

    /**
     * Constructor.
     * 
     * @param javaClass Java Class.
     */
    public DataType(Class<?> javaClass) {

	super();

	if (javaClass == null) {
	    throw new IllegalArgumentException("Argument not valid: null.");
	}

	checkJavaClass(javaClass);

	this.javaClass = javaClass;
	this.array = javaClass.isArray();

	setProperty("name", javaClass.getSimpleName());

	if (javaClass.isArray()) {
	    setProperty("javaClassName", javaClass.getComponentType().getName());
	    setProperty("isArray", true);
	} else {
	    setProperty("javaClassName", javaClass.getName());
	}

    }

    /**
     * Factory.
     * 
     * @param args Arguments.
     * 
     * @return Datatype instance.
     */
    public static DataType create(Object... args) {

	if (args.length == 1) {

	    Object o = args[0];

	    if (o instanceof Class<?>) {

		if (SIMPLETYPES.containsKey(o)) {
		    return SIMPLETYPES.get(o);
		}

		Class<?> cl = (Class<?>) o;

		DataType result = new DataType(cl);
		SIMPLETYPES.put(cl, result);
		return result;

	    } else if (o instanceof DataType) {
		return (DataType) o;
	    } else if (o != null) {
		return create(o.getClass());
	    }
	}

	return (DataType) InvokerHelper.invokeConstructorOf(DataType.class, args);
    }

    /**
     * Factory.
     * 
     * @param dt Datatype.
     * 
     * @return Datatype instance.
     */
    public static DataType create(DataType dt) {
	return dt;
    }

    public static DataType asArray(DataType dt) {

	if (dt.array) {
	    return dt;
	}

	synchronized (ARRAYS) {

	    DataType at = ARRAYS.get(dt);

	    if (at != null) {
		return at;
	    }

	    at = new DataType();

	    Map<?, ?> props = dt.getProperties();

	    for (Object k : props.keySet()) {
		String s = k.toString();
		if (s.equals("id")) {
		    continue;
		}
		at.setProperty(s, props.get(s));
	    }

	    at.setProperty("isArray", true);
	    at.array = true;

	    ARRAYS.put(dt, at);

	    return at;
	}
    }

    /**
     * Transforms a normal data type into a input header type.
     * 
     * @param dt Data type.
     * @return Header type.
     */
    public static DataType asHeader(DataType dt) {
	return asHeader(dt, "IN");
    }

    /**
     * Transforms a normal data type into a header type.
     * 
     * @param dt Data type.
     * @param mode String "IN", "OUT", or "INOUT".
     * @return Header type.
     */
    public static DataType asHeader(DataType dt, String mode) {

	String uCmode = mode.toUpperCase();

	if (!MODES.contains(uCmode)) {
	    throw new IllegalArgumentException("Unknown mode. Valid modes are 'IN', 'OUT', and 'INOUT'.");
	}

	if (dt.isHeader() && dt.getHeaderMode().equals(uCmode)) {
	    return dt;
	}

	synchronized (HEADERS) {
	    DataType ht = null;

	    try {
		ht = HEADERS.get(dt).get(uCmode);
	    } catch (NullPointerException npe) {
		// ignore
	    }

	    if (ht != null) {
		return ht;
	    }

	    ht = new DataType();

	    Map<?, ?> props = dt.getProperties();

	    for (Object k : props.keySet()) {
		String s = k.toString();
		if (s.equals("id")) {
		    continue;
		}
		ht.setProperty(s, props.get(s));
	    }

	    ht.setProperty("isHeader", true);
	    ht.setProperty("headerMode", uCmode);

	    Map<String, DataType> modeMap = HEADERS.get(uCmode);

	    if (modeMap == null) {
		modeMap = new HashMap<String, DataType>();
		HEADERS.put(dt, modeMap);
	    }

	    modeMap.put(uCmode, ht);

	    return ht;
	}
    }

    /**
     * Returns the namespace.
     * 
     * @return Namespace String or <code>null</code> if none has been declared.
     */
    public String getNamespace() {
	return (String) getProperty("namespace");
    }

    /**
     * Checks whether this is a simple type.
     * 
     * @return <code>true</code> if simple type, otherwise <code>false</code>.
     * 
     */
    public boolean isSimple() {
	return SIMPLETYPE_MAP.containsValue(getJavaClass());
    }

    /**
     * Checks whether this is an array of a simple type.
     * 
     * @return <code>true</code> if an array type, otherwise <code>false</code>.
     */
    public boolean isArray() {
	return Boolean.valueOf("" + getProperty("isArray"));
    }

    /**
     * Checks whether this is a complex type.
     * 
     * @return <code>true</code> if complex type, otherwise <code>false</code>.
     */
    public boolean isComplex() {
	return !isSimple();
    }

    /**
     * Checks whether this is a header type.
     * 
     * @return <code>true</code> if header type, otherwise <code>false</code>.
     */
    public boolean isHeader() {
	return Boolean.valueOf("" + getProperty("isHeader"));
    }

    /**
     * Returns the header mode String.
     * 
     * @return Header mode String (IN, OUT, or INOUT) or null if the datatype is
     *         not a header type at all.
     */
    public String getHeaderMode() {

	if (!isHeader()) {
	    return null;
	}

	String modeStr = (String) getProperty("headerMode");

	if (modeStr != null && MODES.contains(modeStr.toUpperCase())) {
	    return modeStr.toUpperCase();
	}

	// by default
	return "IN";
    }

    /**
     * Whether this is an incoming header field.
     * 
     * @return <code>true</code> if yes.
     */
    public boolean isInHeader() {
	return "IN".equals(getHeaderMode());
    }

    /**
     * Whether this is a bidirectional header field.
     * 
     * @return <code>true</code> if yes.
     */
    public boolean isInOutHeader() {
	return "INOUT".equals(getHeaderMode());
    }

    /**
     * Whether this is an outgoing header field.
     * 
     * @return <code>true</code> if yes.
     */
    public boolean isOutHeader() {
	return "OUT".equals(getHeaderMode());
    }

    /**
     * Returns the java class of this type.
     * 
     * @return Java class.
     */
    public Class<?> getJavaClass() {

	try {
	    if (javaClass == null) {

		String javaClassName = (String) super.getProperty("javaClassName");

		if (javaClassName != null) {

		    if (SIMPLETYPE_MAP.containsKey(javaClassName)) {
			javaClass = SIMPLETYPE_MAP.get(javaClassName);
		    } else {
			javaClass = Class.forName(javaClassName);
		    }

		    array = Boolean.valueOf("" + super.getProperty("isArray"));

		    if (array && javaClass != null) {
			javaClass = Class.forName("[L" + javaClass.getName() + ";");
		    }
		}
	    }

	    if (javaClass == null) {
		compileJavaClass();
	    }
	} catch (G2Exception g2e) {
	    throw g2e;
	} catch (Exception e) {
	    throw new G2LocalException("Cannot instantiate Java class.", e);
	}

	return javaClass;
    }

    /**
     * Extracts the source representation of this class name. With brackets for
     * arrays and List<> for lists.
     * 
     * @return Source snippet.
     */
    public String extractClassName() {

	Class<?> javaClass = getJavaClass();

	if (isArray()) {
	    if (javaClass.isArray()) {
		return "List<" + javaClass.getComponentType().getName() + ">";
	    } else {
		if (javaClass.isPrimitive()) {
		    return javaClass.getName() + "[]";

		} else {
		    return "List<" + javaClass.getName() + ">";
		}
	    }
	}

	return javaClass.getName();
    }

    /**
     * Compile to java class, if neededDataType.
     * 
     * @throws ClassNotFoundException If the class cannot be found.
     */
    private void compileJavaClass() throws ClassNotFoundException {

	XSD xsd = (XSD) getProperty("xsd");

	Class<?> clazz = xsd.getJavaClass((String) getProperty("name"));

	if (isArray()) {
	    javaClass = Class.forName("[L" + clazz.getName() + ";", false, LOADER);
	} else {
	    javaClass = clazz;
	}
    }

    /**
     * Checks whether the Java class is supported.
     * 
     * @param javaClass Class.
     */
    private void checkJavaClass(Class<?> javaClass) {

	if (!SIMPLETYPE_MAP.containsValue(javaClass) && !javaClass.isPrimitive()) {
	    throw new IllegalArgumentException("This Java class is (currently) not supported.");
	}
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.core.internal.Cleanable#cleanup()
     */
    @Override
    public void cleanup() {

	REPOSITORY.deleteRepoDir(localRepoDir);

	if (classRepoDir != null) {
	    REPOSITORY.deleteRepoDir(classRepoDir);
	}
    }

    /**
     * Utility function
     * 
     * @param args Arguments forwarded to the constructor.
     * @return Instance.
     */
    public Object newInstance(Object... args) {

	Object o = InvokerHelper.invokeConstructorOf(getJavaClass(), args);

	//
	if (o instanceof ADataTypeInstance) {
	    ((ADataTypeInstance) o).setXsdNamespace(getNamespace());
	}

	return o;
    }

    /**
     * Create an instance of this datatype, filled with dummy data.
     * 
     * @return Dummy instance.
     */
    public Object createDummy() {
	return createDummy(this.getJavaClass());
    }

    /**
     * Create an instance of the passed class, filled with dummy data.
     * 
     * @param c Class.
     * @return Dummy instance.
     */
    public static Object createDummy(Class<?> c) {

	try {
	    return randomValue(c);
	} catch (IllegalArgumentException iae) {
	    // go on
	}

	Object o;
	try {
	    o = c.newInstance();
	} catch (Exception ie) {
	    LOGGER.warn("Cannot create instance of class " + c.getName() + ".");
	    return null;
	}
	Method[] methods = c.getDeclaredMethods();

	for (Method m : methods) {
	    Class<?>[] parTypes = m.getParameterTypes();

	    if (parTypes.length == 0 && m.getName().startsWith("get")) {
		Class<?> retType = m.getReturnType();

		if (!retType.equals(List.class)) {
		    continue;
		}

		try {
		    String nestedName = m.getName().substring(3);
		    nestedName = (c.getName() + "$" + nestedName).toLowerCase();

		    Class<?> nested = null;

		    for (Class<?> nc : c.getDeclaredClasses()) {
			if (nc.getName().toLowerCase().equals(nestedName)) {
			    nested = nc;
			}
		    }

		    if (nested != null) {
			List<Object> list = (List<Object>) m.invoke(o);

			for (int i = 0; i < 1 + new Random().nextInt(3); i++) {
			    list.add(createDummy(nested));
			}
		    }

		} catch (Exception e) {
		    // don't insert anything
		}
	    }

	    if (parTypes.length == 1 && m.getName().startsWith("set")) {
		Class<?> parType = parTypes[0];

		Object val = null;

		try {
		    val = randomValue(parType);
		} catch (IllegalArgumentException iae) {
		    val = createDummy(parType);
		}

		try {
		    m.invoke(o, val);
		} catch (Exception ie) {
		    // ignore
		}
	    }
	}
	return o;
    }

    @SuppressWarnings("all")
    private static Object randomValue(Class<?> parType) {

	if (parType.isAssignableFrom(int.class)) {
	    return new Random().nextInt();
	} else if (parType.isAssignableFrom(long.class)) {
	    return new Random().nextInt();
	} else if (parType.isAssignableFrom(short.class)) {
	    return (short) new Random().nextInt();
	} else if (parType.isAssignableFrom(byte.class)) {
	    return (byte) new Random().nextInt();
	} else if (parType.isAssignableFrom(char.class)) {
	    return (char) new Random().nextInt();
	} else if (parType.isAssignableFrom(float.class)) {
	    return new Random().nextFloat();
	} else if (parType.isAssignableFrom(double.class)) {
	    return new Random().nextDouble();
	} else if (parType.isAssignableFrom(boolean.class)) {
	    return new Random().nextBoolean();
	} else if (parType.isAssignableFrom(String.class)) {
	    return "str" + Math.abs(new Random().nextInt(1000));
	} else if (parType.isAssignableFrom(byte[].class)) {
	    return ("str" + Math.abs(new Random().nextInt(1000))).getBytes();
	} else if (parType.isAssignableFrom(QName.class)) {
	    return new QName("name" + Math.abs(new Random().nextInt()));
	} else if (parType.isAssignableFrom(BigInteger.class)) {
	    return new BigInteger(8, new Random());
	} else if (parType.isAssignableFrom(BigDecimal.class)) {
	    return BigDecimal.valueOf(new Random().nextInt());
	} else {
	    try {
		if (parType.isAssignableFrom(Duration.class)) {
		    return DatatypeFactory.newInstance().newDuration(new Random().nextLong());
		} else if (parType.isAssignableFrom(XMLGregorianCalendar.class)) {
		    Random r = new Random();
		    return DatatypeFactory.newInstance().newXMLGregorianCalendar(r.nextInt(3000), 1 + r.nextInt(12),
			    1 + r.nextInt(7), r.nextInt(24), r.nextInt(60), r.nextInt(60), r.nextInt(1000),
			    r.nextInt(24));
		}
	    } catch (DatatypeConfigurationException e) {
		// do nothing
	    }
	}

	throw new IllegalArgumentException("Cannot handle type.");
    }
}
