package at.ac.tuwien.vitalab.genesis2.core.model.types;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;

import at.ac.tuwien.vitalab.genesis2.base.Genesis2Base;
import at.ac.tuwien.vitalab.genesis2.base.ws.IGenesis2BaseWebService;
import at.ac.tuwien.vitalab.genesis2.config.Genesis2Configuration;
import at.ac.tuwien.vitalab.genesis2.core.model.AbstractModelElement;
import at.ac.tuwien.vitalab.genesis2.core.model.ws.wrappers.ModelElementWrapper;
import at.ac.tuwien.vitalab.genesis2.errors.G2Exception;
import at.ac.tuwien.vitalab.genesis2.plugin.AbstractGenesis2Plugin;

/**
 * Internal backend host representation.
 * 
 * @author Lukasz Juszczyk
 * 
 */
public class Host extends AbstractModelElement {

    private static final Logger LOGGER = Logger.getLogger(Host.class);

    private final Set<AbstractModelElement> elements = new HashSet<AbstractModelElement>();

    private static Map<String, Host> created = new HashMap<String, Host>();

    private static Host localhost = null;

    // stupid workaround for some plugins, which need to deploy their artifacts
    // on a different port than the one the backend is listening on
    private static Host deploymenthost = null;

    static {
	getLocalhost();

	try {
	    int port = Integer.parseInt(Genesis2Configuration.get("net.deployment.port").toString());
	    FakeHost fh = new FakeHost(getLocalhost().getHost(), port);
	    Host.setFakeLocalhostForDeployment(fh);
	} catch (Exception e) {
	    // ignore
	}

    }

    /**
     * Constructor.
     */
    public Host() {
	super();

	setPropertyDirectly("elements", elements);
    }

    /**
     * Constructor.
     * 
     * @param host Host.
     * @param port Port.
     */
    public Host(String host, int port) {
	this(host, port, false); // NOPMD
    }

    /**
     * Constructor.
     * 
     * @param host Host.
     * @param port Port.
     * @param Whether to purge the remote host.
     */
    private Host(String host, int port, boolean purge) {

	super();

	setPropertyDirectly("location", host + ":" + port);
	setPropertyDirectly("host", host);
	setPropertyDirectly("port", port);
	setPropertyDirectly("elements", elements);

	checkHost(purge);

	installPlugins();
    }

    /**
     * Install all local plugins on that host.
     */
    protected void installPlugins() {

	if (isLocalhost()) {
	    return;
	}

	IGenesis2BaseWebService client = getClient();

	for (AbstractGenesis2Plugin<?> plugin : AbstractGenesis2Plugin.getRunningPlugins()) {
	    String className = plugin.getClass().getName();
	    try {
		client.startPlugin(className);
	    } catch (Exception e) {
		throw new G2Exception("Cannot start plugin '" + className + "' on remote host.");
	    }
	}

    }

    /**
     * Factory.
     * 
     * @param host Host location.
     * @param port Port numner.
     * 
     * @return Host instance.
     */
    public static Host create(String host, int port) {
	return create(host, port, false);
    }

    /**
     * Factory.
     * 
     * @param host Host location.
     * @param port Port numner.
     * @param purge Whether to reset the remote host.
     * 
     * @return Host instance.
     */
    public static Host create(String host, int port, boolean purge) {
	String location = host + ":" + port;

	Host h = created.get(location);

	if (h != null) {

	    if (purge) { // clean everything on it
		h.purge();
	    }

	    return h;
	}

	h = new Host(host, port, purge);
	created.put(location, h);

	return h;
    }

    /**
     * Resets the remote host.
     */
    public synchronized void purge() {
	IGenesis2BaseWebService client = getClient();

	try {
	    client.purge();
	} catch (Exception e) {
	    throw new G2Exception(e);
	}

	String host = getHost();
	int port = getPort();

	// reset properties to minimal ones
	elements.clear();
	setPropertyDirectly("location", host + ":" + port);
	setPropertyDirectly("host", host);
	setPropertyDirectly("port", port);
	setPropertyDirectly("elements", elements);
    }

    /**
     * Returns the local host.
     * 
     * @return Localhost instance.
     */
    public synchronized static Host getLocalhost() {

	if (localhost != null) {
	    return localhost;
	}

	String host = Genesis2Configuration.getLocalHost();
	Integer port = Genesis2Configuration.getLocalPort();

	Host l = null;

	try {
	    l = new Host(host, port);
	} catch (Exception e1) {
	    LOGGER.fatal("Cannot access local Genesis2 Web service: ", e1);
	    System.exit(1);
	}

	@SuppressWarnings("unchecked")
	Set<String> tags = (Set<String>) l.getProperty("tags");
	tags.add("localhost");

	(localhost = l).register();

	return localhost;
    }

    /**
     * Gets the fake host where atrifacts should be deployed.
     * 
     * @return Host.
     */
    public static Host getFakeLocalhostForDeployment() {

	if (deploymenthost != null) {
	    return deploymenthost;
	}

	return getLocalhost();
    }

    /**
     * Sets the fake host where atrifacts should be deployed.
     * 
     * @param h Host.
     */
    public static void setFakeLocalhostForDeployment(Host h) {
	deploymenthost = h;
    }

    /**
     * Clear all elements bound to the local host.
     */
    public synchronized static void purgeLocalhost() {

	for (AbstractModelElement kid : getLocalhost().getAllChildren()) {
	    kid.deregister();
	}

	getLocalhost().elements.clear();
    }

    /**
     * Checks whether this is the localhost.
     * 
     * @return <code>true</code> if localhost, otherwise <code>false</code>.
     */
    public boolean isLocalhost() {

	String confHost = Genesis2Configuration.getLocalHost();
	Integer confPort = Genesis2Configuration.getLocalPort();

	if (localhost == null) {
	    return (getHost().equals(confHost) && getPort() == confPort);
	}
	return (this == localhost || (confHost.equals(getHost()) && getPort() == confPort));
    }

    /**
     * Returns the WS client to this host's G2 web service.
     * 
     * @return WS client.
     */
    private IGenesis2BaseWebService getClient() {
	// get client directly
	Genesis2Base basePlugin = AbstractGenesis2Plugin.getPlugin(Genesis2Base.class);

	if (basePlugin == null) {
	    throw new G2Exception("Cannot get instance of Genesis2 Base Plugin.");
	}

	try {
	    return (IGenesis2BaseWebService) basePlugin.getPluginWsClient(getHost(), getPort());
	} catch (Exception e) {
	    throw new G2Exception("Cannot get instance of Genesis2 Base Plugin Web service client.", e);
	}
    }

    /**
     * Check whether host can be accessed.
     * 
     * @param purge Clean everything on it?
     */
    protected void checkHost(boolean purge) {

	IGenesis2BaseWebService client = getClient();

	if (isLocalhost()) {
	    client.ping();
	} else {

	    if (purge) {

		try {
		    client.purge();
		} catch (Exception e) {
		    throw new G2Exception(e.getMessage());
		}

	    } else {
		// import model elements from remote host

		List<ModelElementWrapper> mews = client.getWholeModel();

		Host remote = (Host) (mews.get(0).unwrap());

		this.copyNewProperties(remote);

		for (int i = 1; i < mews.size(); i++) {
		    mews.get(i).unwrap().markAsDeployedAt(this);
		}

	    }
	}

	this.markAsCreatedAt(this);
    }

    private synchronized final void copyNewProperties(AbstractModelElement src) {

	if (src.getClass() != this.getClass()) {
	    throw new IllegalArgumentException("Source element must have the same class.");
	}

	Map<?, ?> props = src.getProperties();

	for (Object propKey : props.keySet()) {
	    String propName = propKey.toString();

	    Object propValue = props.get(propKey);

	    if (getProperties().keySet().contains(propName) && !propName.equals("tags") && !propName.equals("elements")) {
		continue;
	    }

	    setProperty(propName, propValue);
	}

	overrideId(src.getId());
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.core.internal.ICleanable#cleanup()
     */
    @Override
    public void cleanup() {
	// do nothing
    }

    /**
     * Gets the host name (can be the IP too).
     * 
     * @return Host name.
     */
    public String getHost() {
	return (String) getProperty("host");
    }

    /**
     * Gets the location (host:port).
     * 
     * @return Location.
     */
    public String getLocation() {
	return (String) getProperty("location");
    }

    /**
     * Gets the port number.
     * 
     * @return Port.
     */
    public int getPort() {
	Object o = getProperty("port");
	if (o instanceof Integer) {
	    return (Integer) o;
	}
	return Integer.parseInt(o.toString());
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.core.model.AModelElement#setProperty(java
     * .lang.String, java.lang.Object)
     */
    @Override
    public synchronized void setProperty(String property, Object newValue) {

	if (super.getProperties().containsKey(property)
		&& ("host".equals(property) || "port".equals(property) || "location".equals(property))) {

	    // allow setting these only once
	    return;
	}

	if ("elements".equals(property)) {

	    elements.clear();

	    if (newValue != null) {
		@SuppressWarnings("unchecked")
		Collection<AbstractModelElement> els = (Collection<AbstractModelElement>) newValue;
		elements.addAll(els);
	    }

	    return;
	}

	super.setProperty(property, newValue);
    }

    /**
     * Attach a child.
     * 
     * @param ame Child.
     * 
     * @return Whether it has been attached before.
     */
    public boolean attach(AbstractModelElement ame) {

	boolean result = elements.add(ame);

	setPropertyDirectly("elements", elements);

	return result;
    }

    /**
     * Detach a child.
     * 
     * @param ame Child.
     * 
     * @return Whether it has existed before.
     */
    public boolean detach(AbstractModelElement ame) {

	boolean result = elements.remove(ame);

	setPropertyDirectly("elements", elements);

	return result;
    }
}
