package at.ac.tuwien.vitalab.genesis2.core.model.types;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.StringReader;
import java.util.HashMap;
import java.util.Map;

import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.Namespace;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import at.ac.tuwien.vitalab.genesis2.core.model.AbstractModelElement;
import at.ac.tuwien.vitalab.genesis2.errors.G2LocalException;
import at.ac.tuwien.vitalab.genesis2.utils.CompilationUtils;
import at.ac.tuwien.vitalab.genesis2.utils.Repository;

/**
 * Representation of XSD types.
 * 
 * @author Lukasz Juszczyk
 * 
 */
public class XSD extends AbstractModelElement {

    private boolean compiled = false;

    private File repoDir = null;

    private static final Repository REPOSITORY = Repository.getInstance();

    private static final CompilationUtils COMPILATION_UTILS = CompilationUtils.getInstance();

    /**
     * Constructor.
     */
    public XSD() {
	super();
    }

    /**
     * Constructor.
     * 
     * @param xsdFile Schema files.
     * @throws IOException On errors while processing file.
     */
    public XSD(File... xsdFile) throws IOException {

	super();

	synchronized (getClass()) {

	    HashMap<String, String> xsdMap = new HashMap<String, String>();

	    for (File f : xsdFile) {

		FileReader fr = new FileReader(f);
		BufferedReader br = new BufferedReader(fr);

		String xsd = "";

		String line;
		while ((line = br.readLine()) != null) {
		    xsd += line + "\n";
		}

		br.close();

		xsdMap.put(f.getName(), xsd);

	    }

	    setProperty("xsd", xsdMap);
	}
    }

    /**
     * Returns the compiled Java class extracted from the XSD.
     * 
     * @param name Class name.
     * @return Class instance.
     */
    public Class<?> getJavaClass(String name) {
	return getJavaClass(name, false);
    }

    /**
     * Returns the compiled Java class extracted from the XSD.
     * 
     * @param name Class name.
     * @param caseSensitive Whether the name shall be resolved in a case
     *            sensitive manner.
     * @return Class instance.
     */
    public Class<?> getJavaClass(String name, boolean caseSensitive) {

	try {

	    File baseDir = compile();

	    for (File f : new File(baseDir, "type").listFiles()) {

		boolean matches;

		if (caseSensitive) {
		    matches = f.getName().equals(name + ".class");
		} else {
		    matches = f.getName().equalsIgnoreCase(name + ".class");
		}

		if (matches) {
		    Class<?> cl = DataType.LOADER.loadClass("type." + f.getName().replaceFirst(".class$", ""));

		    return cl;
		}
	    }
	} catch (Exception e) {
	    throw new G2LocalException("Cannot compile XSD to Java class.", e);
	}

	return null;

    }

    // compile xsd file to java classes
    private synchronized File compile() throws IOException, JDOMException {

	if (compiled) {
	    return repoDir;
	}

	@SuppressWarnings("unchecked")
	Map<String, String> xsdMap = (Map<String, String>) getProperty("xsd");

	if (xsdMap == null || xsdMap.isEmpty()) {
	    throw new IllegalArgumentException("No XSD document specified.");
	}

	repoDir = REPOSITORY.newRepoDir("xsd");

	File firstFile = null;

	for (String name : xsdMap.keySet()) {
	    String xsd = xsdMap.get(name);

	    SAXBuilder builder = new SAXBuilder();
	    Element el = builder.build(new StringReader(xsd)).getRootElement();
	    el.detach();

	    File f = new File(repoDir, name);

	    if (firstFile == null) {

		firstFile = f;

		String xsdPrefix = el.getNamespacePrefix();

		Namespace jaxbNS = Namespace.getNamespace("jaxb", "http://java.sun.com/xml/ns/jaxb");
		Namespace xjcNS = Namespace.getNamespace("xjc", "http://java.sun.com/xml/ns/jaxb/xjc");

		String schemaNS = "http://www.w3.org/2001/XMLSchema";
		Element anno = new Element("annotation", xsdPrefix, schemaNS);
		Element appinfo = new Element("appinfo", xsdPrefix, schemaNS);
		Element globalBindings = new Element("globalBindings", jaxbNS);
		Element superClass = new Element("superClass", xjcNS);
		superClass.setAttribute("name", ADataTypeInstance.class.getName());

		anno.addContent(appinfo);
		appinfo.addContent(globalBindings);
		globalBindings.addContent(superClass);

		el.addNamespaceDeclaration(xjcNS);
		el.addNamespaceDeclaration(jaxbNS);

		el.setAttribute("elementFormDefault", "unqualified");

		el.setAttribute("extensionBindingPrefixes", "xjc", jaxbNS);
		el.setAttribute("version", "1.0", jaxbNS);

		el.addContent(anno);
	    }

	    XMLOutputter outputter = new XMLOutputter();
	    outputter.setFormat(Format.getPrettyFormat());

	    FileWriter fw = new FileWriter(f);

	    outputter.output(el, fw);

	    fw.close();
	}

	COMPILATION_UTILS.xjc("type", repoDir, firstFile);

	File[] javaFiles = new File(repoDir, "type").listFiles(new FilenameFilter() {

	    @Override
	    public boolean accept(File dir, String name) {
		return name.endsWith(".java");
	    }
	});

	COMPILATION_UTILS.javac((File) null, repoDir, repoDir, javaFiles);

	DataType.LOADER.addURL(repoDir.toURI().toURL());

	compiled = true;

	return repoDir;
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.core.model.ICleanable#cleanup()
     */
    @Override
    public void cleanup() {
	REPOSITORY.deleteRepoDir(repoDir);
    }

}
