package at.ac.tuwien.vitalab.genesis2.core.model.ws;

import groovy.lang.Closure;

import java.util.Collection;
import java.util.List;
import java.util.Map;

import javax.jws.WebMethod;
import javax.jws.WebService;

import at.ac.tuwien.vitalab.genesis2.core.model.AbstractModelElement;
import at.ac.tuwien.vitalab.genesis2.core.model.ElementMap;
import at.ac.tuwien.vitalab.genesis2.core.model.types.Host;
import at.ac.tuwien.vitalab.genesis2.core.model.ws.wrappers.ClosureWrapper;
import at.ac.tuwien.vitalab.genesis2.core.model.ws.wrappers.KeyValueWrapper;
import at.ac.tuwien.vitalab.genesis2.core.model.ws.wrappers.MapWrapper;
import at.ac.tuwien.vitalab.genesis2.core.model.ws.wrappers.ModelElementWrapper;
import at.ac.tuwien.vitalab.genesis2.errors.G2RemoteException;
import at.ac.tuwien.vitalab.genesis2.utils.ClosureCollectionIterator;
import at.ac.tuwien.vitalab.genesis2.utils.ClosureUtil;

/**
 * Web service for model manipulation.
 * 
 * @author Lukasz Juszczyk
 * 
 * @param <T> Model type.
 */
@WebService(endpointInterface = "at.ac.tuwien.vitalab.genesis2.core.model.ws.IModelElementWebService", serviceName = "IModelElementWebServiceService")
public class ModelElementWebService<T extends AbstractModelElement> implements IModelElementWebService {

    private ElementMap<T> mapOfModel = null;

    /**
     * Constructor.
     */
    public ModelElementWebService() {
	super();
    }

    /**
     * Constructor.
     * 
     * @param map
     */
    public ModelElementWebService(ElementMap<T> map) {
	super();
	this.mapOfModel = map;
    }

    /**
     * Returns the element instance with this ID. Throws an exception if the ID
     * is invalid, in terms of that it does not reference any element instance.
     * 
     * @param id ID.
     */
    private T getElementById(String id) {

	T element = mapOfModel.get(id);

	if (element == null) {
	    throw new IllegalArgumentException("Unknown element with ID '" + id + "'.");
	}

	return element;
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.core.model.ws.IModelElementWebService#
     * setParameter(java.lang.String,
     * at.ac.tuwien.vitalab.genesis2.core.model.ws.wrappers.KeyValueWrapper)
     */
    @WebMethod
    public void setParameter(String id, KeyValueWrapper property) {

	T element = getElementById(id);

	element.setProperty(property.getUnwrappedKey().toString(), property.getUnwrappedValue());
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.core.model.ws.IModelElementWebService#change
     * (at.ac.tuwien.vitalab.genesis2.core.model.ws.wrappers.MapWrapper,
     * at.ac.tuwien.vitalab.genesis2.core.model.ws.wrappers.ClosureWrapper,
     * at.ac.tuwien.vitalab.genesis2.core.model.ws.wrappers.ClosureWrapper)
     */
    @SuppressWarnings("unchecked")
    @Override
    public void change(MapWrapper varMap, ClosureWrapper wrFilter, ClosureWrapper wrCommand) {

	Collection<T> elements = mapOfModel.get(wrFilter.unwrap());

	if (elements == null || elements.isEmpty()) {
	    return;
	}

	Closure command = wrCommand.unwrap();
	ClosureUtil.getInstance().registerConstants(command, (Map<String, Object>) varMap.unwrap());

	ClosureCollectionIterator.each(elements, command);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.core.model.ws.IModelElementWebService#update
     * (java.util.List)
     */
    @WebMethod
    public void update(List<ModelElementWrapper> updatedElements) {

	if (updatedElements == null) {
	    return;
	}

	for (ModelElementWrapper wrapper : updatedElements) {
	    wrapper.unwrap();
	}
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.core.model.ws.IModelElementWebService#create
     * (
     * at.ac.tuwien.vitalab.genesis2.core.model.ws.wrappers.ModelElementWrapper,
     * java.util.List)
     */
    @WebMethod
    public void create(ModelElementWrapper rootElement, List<ModelElementWrapper> updatedChildren) {

	AbstractModelElement root = rootElement.unwrap();

	try {
	    root.bindTo(Host.getLocalhost());
	} catch (Exception e) {
	    throw new G2RemoteException(e);
	}

	update(updatedChildren);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.core.model.ws.IModelElementWebService#remove
     * (java.lang.String)
     */
    @WebMethod
    public void remove(String id) {

	AbstractModelElement element = getElementById(id);

	undeployElement(id);
	element.deregister();
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.core.model.ws.IModelElementWebService#
     * redeployElement(java.lang.String)
     */
    @Override
    public void redeployElement(String id) {

	T element = getElementById(id);

	mapOfModel.getPlugin().redeployElement(element);
	element.markAsLocallyDeployed();
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.core.model.ws.IModelElementWebService#
     * redeployElements
     * (at.ac.tuwien.vitalab.genesis2.core.model.ws.wrappers.ClosureWrapper)
     */
    @Override
    public void redeployElements(ClosureWrapper filter) {

	Closure closure = filter.unwrap();

	for (T element : mapOfModel.get(closure)) {
	    mapOfModel.getPlugin().redeployElement(element);
	    element.markAsLocallyDeployed();
	}
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.core.model.ws.IModelElementWebService#
     * deployElement(java.lang.String)
     */
    @WebMethod
    public void deployElement(String id) {

	T element = getElementById(id);

	mapOfModel.getPlugin().deployElement(element);
	element.markAsLocallyDeployed();
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.core.model.ws.IModelElementWebService#
     * deployElements
     * (at.ac.tuwien.vitalab.genesis2.core.model.ws.wrappers.ClosureWrapper)
     */
    @WebMethod
    public void deployElements(ClosureWrapper filter) {

	Closure closure = filter.unwrap();

	for (T element : mapOfModel.get(closure)) {
	    mapOfModel.getPlugin().deployElement(element);
	    element.markAsLocallyDeployed();
	}
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.core.model.ws.IModelElementWebService#
     * undeployElement(java.lang.String)
     */
    @WebMethod
    public void undeployElement(String id) {

	T element = getElementById(id);

	mapOfModel.getPlugin().undeployElement(element);
	element.markAsLocallyUndeployed();
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.core.model.ws.IModelElementWebService#
     * undeployElements
     * (at.ac.tuwien.vitalab.genesis2.core.model.ws.wrappers.ClosureWrapper)
     */
    @WebMethod
    public void undeployElements(ClosureWrapper filter) {

	Closure closure = filter.unwrap();

	for (T element : mapOfModel.get(closure)) {
	    mapOfModel.getPlugin().undeployElement(element);
	    element.markAsLocallyUndeployed();
	}
    }

}
