package at.ac.tuwien.vitalab.genesis2.core.model.ws.wrappers;

import groovy.lang.Closure;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import org.codehaus.groovy.runtime.InvokerHelper;

import at.ac.tuwien.vitalab.genesis2.core.model.AbstractModelElement;

/**
 * Base wrapper class for de-/serialization of objects into/from XML.
 * 
 * @author Lukasz Juszczyk
 * 
 * @param <T> Type to be wrapped in XML.
 * 
 */
public abstract class AbstractWrapper<T> {

    /**
     * Replacer for exchanging certain objects within the de/serialized types.
     * E.g. for replacing model elements with references to these.
     */
    protected transient ISerializationObjectReplacer replacer = null;

    protected static Map<Class<?>, Class<? extends AbstractWrapper<?>>> registered = new HashMap<Class<?>, Class<? extends AbstractWrapper<?>>>();

    // by default
    static {
	registered.put(KeyValue.class, KeyValueWrapper.class);
	registered.put(AbstractModelElement.class, ModelElementWrapper.class);
	registered.put(Collection.class, CollectionWrapper.class);
	registered.put(Map.class, MapWrapper.class);
	registered.put(Closure.class, ClosureWrapper.class);
    }

    /**
     * Register a wrapper class.
     * 
     * @param <T> Object type.
     * @param clazz Object class.
     * @param wrapper Wrapper class.
     * @return Old wrapper, if existed.
     */
    @SuppressWarnings("unchecked")
    public static <T> Class<? extends AbstractWrapper<T>> registerWrapper(Class<T> clazz,
	    Class<? extends AbstractWrapper<T>> wrapper) {
	return (Class<? extends AbstractWrapper<T>>) registered.put(clazz, wrapper);
    }

    /**
     * Deregister a wrapper class.
     * 
     * @param <T> Object type.
     * @param clazz Object class.
     * @return Old wrapper, if existed.
     */
    @SuppressWarnings("unchecked")
    public static <T> Class<? extends AbstractWrapper<T>> deregisterWrapper(Class<T> clazz) {
	return (Class<? extends AbstractWrapper<T>>) registered.remove(clazz);
    }

    /**
     * Constructor.
     */
    protected AbstractWrapper() {
	super();
    }

    /**
     * Constructor.
     * 
     * @param object Object to wrap.
     */
    protected AbstractWrapper(T object) {
	this(object, null);
    }

    /**
     * Constructor.
     * 
     * @param object Object to wrap.
     * @param rep Replacer.
     */
    protected AbstractWrapper(T object, ISerializationObjectReplacer rep) {
	replacer = rep;
	wrap(object);
    }

    /**
     * Set the replacer.
     * 
     * @param replacer Replacer instance.
     */
    public void setReplacer(ISerializationObjectReplacer replacer) {
	this.replacer = replacer;
    }

    /**
     * Deserialization.
     * 
     * @return Deserialized object.
     */
    public abstract T unwrap();

    /**
     * Serialization.
     * 
     * @param object Object to be serialized.
     */
    public abstract void wrap(T object);

    /**
     * Wraps an object.
     * 
     * @param object Object to be wrapped.
     * @param rep Replacer.
     * @return Wrapped object.
     */
    protected static Object wrapObject(Object object, ISerializationObjectReplacer rep) {

	Object result = object;

	if (object != null) {

	    Class<?> clazz = object.getClass();
	    Object[] constParams = new Object[] { object, rep };

	    for (Class<?> regClass : registered.keySet()) {
		if (regClass.isAssignableFrom(clazz)) {

		    Class<? extends AbstractWrapper<?>> wrapperClass = registered.get(regClass);
		    result = InvokerHelper.invokeConstructorOf(wrapperClass, constParams);
		    break;
		}
	    }
	}

	return result;
    }

    /**
     * Unwraps an object.
     * 
     * @param wrapped Wrapped object.
     * @return Unwrapped object.
     */
    protected static Object unwrapObject(Object wrapped) {

	Object unwrapped = wrapped;

	if (wrapped instanceof AbstractWrapper) {
	    unwrapped = ((AbstractWrapper<?>) wrapped).unwrap();
	}

	return unwrapped;
    }
}
