package at.ac.tuwien.vitalab.genesis2.core.model.ws.wrappers;

import groovyjarjarantlr.collections.AST;

import java.util.ArrayList;
import java.util.List;

import org.codehaus.groovy.antlr.AntlrParserPlugin;
import org.codehaus.groovy.antlr.LineColumn;
import org.codehaus.groovy.antlr.SourceBuffer;
import org.codehaus.groovy.ast.expr.ClosureExpression;
import org.codehaus.groovy.control.ParserPlugin;
import org.codehaus.groovy.control.ParserPluginFactory;
import org.codehaus.groovy.control.SourceUnit;
import org.codehaus.groovy.syntax.Reduction;

/**
 * Groovy Code Parser for extracting the source of closures.
 * 
 * @author Lukasz Juszczyk
 * 
 */
public class ClosureFetchingParser extends AntlrParserPlugin {

    private transient ClosureEntry main = new ClosureEntry();

    private transient ClosureEntry last = main;

    private transient SourceBuffer source = null;

    private transient List<String> sourceList = null;

    /*
     * (non-Javadoc)
     * 
     * @see org.codehaus.groovy.antlr.AntlrParserPlugin#closureExpression(
     * groovyjarjarantlr.collections.AST)
     */
    @Override
    public ClosureExpression closureExpression(AST node) {

	ClosureEntry pre = last;
	ClosureEntry entry = new ClosureEntry();

	last.nested.add(entry);
	last = entry;

	ClosureExpression expr = super.closureExpression(node);

	entry.closureSrc = source.getSnippet(new LineColumn(expr.getLineNumber(), expr.getColumnNumber()),
		new LineColumn(expr.getLastLineNumber(), expr.getLastColumnNumber()));

	last = pre;

	return expr;
    }

    /**
     * Returns a list of sorted closure sources.
     * 
     * @return List of source strings.
     */
    public synchronized List<String> getClosureSources() { // NOPMD

	if (sourceList == null) {
	    sourceList = new ArrayList<String>();
	    sourceList.add(null);
	    sourceList.addAll(main.createList());
	}

	return sourceList;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.codehaus.groovy.antlr.AntlrParserPlugin#outputAST(org.codehaus.groovy
     * .control.SourceUnit, org.codehaus.groovy.antlr.SourceBuffer)
     */
    @Override
    public Reduction outputAST(SourceUnit sourceUnit, SourceBuffer sourceBuffer) {

	this.source = sourceBuffer;

	return super.outputAST(sourceUnit, sourceBuffer);
    }

    private static class ClosureEntry {

	private transient String closureSrc;
	private final transient List<ClosureEntry> nested = new ArrayList<ClosureEntry>();

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
	    return closureSrc + nested;
	}

	/**
	 * Creates list of source blocks.
	 * 
	 * @return List of source blocks.
	 */
	private List<String> createList() {
	    List<String> result = new ArrayList<String>();

	    for (ClosureEntry ce : nested) {
		result.add(ce.closureSrc);
	    }

	    for (ClosureEntry ce : nested) {
		List<String> list = ce.createList();
		result.addAll(list);
	    }

	    return result;
	}
    }

    /**
     * Factory.
     * 
     * @author Lukasz Juszczyk
     * 
     */
    public static class ClosureFetchingParserFactory extends ParserPluginFactory {

	private transient ClosureFetchingParser parser = null;

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.codehaus.groovy.control.ParserPluginFactory#createParserPlugin()
	 */
	@Override
	public synchronized ParserPlugin createParserPlugin() { // NOPMD
	    if (parser == null) {
		parser = new ClosureFetchingParser();
	    }
	    return parser;
	}

	/**
	 * Returns a list of all closure sources.
	 * 
	 * @return List of sources.
	 */
	public List<String> getClosures() {
	    createParserPlugin();
	    return parser.getClosureSources();
	}

	/**
	 * Clear everything.
	 */
	public void reset() {
	    if (parser == null) {
		return;
	    }
	    parser.sourceList = null; // NOPMD
	    parser.source = null; // NOPMD
	    parser.last = parser.main = new ClosureEntry();
	}
    }
}
