package at.ac.tuwien.vitalab.genesis2.core.model.ws.wrappers;

import groovy.lang.Binding;
import groovy.lang.Closure;
import groovy.util.Expando;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import javax.xml.bind.annotation.XmlAccessType;
import javax.xml.bind.annotation.XmlAccessorType;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlTransient;
import javax.xml.bind.annotation.XmlType;

import lombok.Getter;
import lombok.Setter;

import org.codehaus.groovy.control.CompilationFailedException;

import at.ac.tuwien.vitalab.genesis2.control.InterceptingGroovyShell;
import at.ac.tuwien.vitalab.genesis2.utils.ClosureUtil;

/**
 * Closure wrapper for XML marshalling.
 * 
 * @author Lukasz Juszczyk
 * 
 */
@XmlAccessorType(XmlAccessType.NONE)
@XmlType(name = "Closure")
@XmlRootElement
public class ClosureWrapper extends AbstractWrapper<Closure> {

    private final static Map<String, List<String>> SOURCES = new HashMap<String, List<String>>();

    private final static Map<Closure, Map<String, Object>> BINDINGS = new HashMap<Closure, Map<String, Object>>();

    private final static Map<Closure, String> SOURCE_CACHE = new HashMap<Closure, String>();

    private final static Map<String, Closure> CLOSURE_CACHE = new HashMap<String, Closure>();

    @XmlTransient
    private Closure closure = null;

    /**
     * Source code block of the closre.
     */
    @XmlElement
    @Getter
    @Setter
    public String closureSource;

    /**
     * Variable binding.
     */
    @XmlElement
    @Getter
    @Setter
    private MapWrapper wrappedBinding;

    /**
     * Registers a list of closure sources for a given class name.
     * 
     * @param className Class name.
     * @param closureSources Closure source list.
     */
    public static void registerSources(String className, List<String> closureSources) {
	SOURCES.put(className, closureSources);
    }

    /**
     * Registers a variable binding for a given closure.
     * 
     * @param closure Closure.
     * @param binding Binding.
     */
    public static void registerBinding(Closure closure, Map<String, Object> binding) {
	BINDINGS.put(closure, binding);
    }

    /**
     * Constructor.
     */
    public ClosureWrapper() {
	super();
    }

    /**
     * Constructor.
     * 
     * @param clos Closure.
     */
    public ClosureWrapper(Closure clos) {
	this(clos, null);
    }

    /**
     * Constructor.
     * 
     * @param clos Closure.
     * @param rep Replacer.
     */
    public ClosureWrapper(Closure clos, ISerializationObjectReplacer rep) {
	super(clos, rep);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.core.model.ws.wrappers.AbstractWrapper#
     * wrap(java.lang.Object)
     */
    @Override
    public void wrap(Closure clos) {
	this.closure = clos;

	Map<?, ?> binding = BINDINGS.get(clos);

	if (binding != null) {
	    wrappedBinding = new MapWrapper(binding);
	}

	decompile();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.core.model.ws.wrappers.AWrapper#unwrap()
     */
    @Override
    public Closure unwrap() {
	compile();
	return closure;
    }

    /**
     * Compile a closure source into a closure instance.
     * 
     * @throws CompilationFailedException If the source cannot be compiled.
     */
    @SuppressWarnings("unchecked")
    private synchronized void compile() throws CompilationFailedException { // NOPMD

	if (closureSource == null) {
	    throw new IllegalArgumentException("No closure source found. Cannot compile.");
	}

	if (closure == null) {
	    closure = CLOSURE_CACHE.get(closureSource);
	}

	if (closure == null) {
	    InterceptingGroovyShell shell = InterceptingGroovyShell.newInstance(new Binding());
	    closure = (Closure) shell.evaluate("c=" + closureSource);

	    SOURCE_CACHE.put(closure, closureSource);
	    CLOSURE_CACHE.put(closureSource, closure);

	    if (wrappedBinding != null) {
		Map<String, Object> boundValues = (Map<String, Object>) wrappedBinding.unwrap();

		Expando delegate = new Expando();

		for (Entry<String, Object> entry : boundValues.entrySet()) {
		    delegate.setProperty(entry.getKey(), entry.getValue());
		}

		BINDINGS.put(this.unwrap(), boundValues);

		ClosureUtil.getInstance().attachDelegates(closure, true, delegate);
	    }
	}
    }

    /**
     * Retrieves the source code for a closure.
     */
    private void decompile() {

	if (closure == null) {
	    throw new IllegalArgumentException("No closure found. Cannot retrieve source.");
	}

	if (closureSource == null) {
	    closureSource = SOURCE_CACHE.get(closure);
	}

	if (closureSource == null) {

	    String notFoundErrorMsg = "Cannot decompile closure '" + closure + "'. Not defined within script?";

	    String owner = closure.getOwner().getClass().getName().replaceFirst("\\$.*", "");

	    List<String> srcList = SOURCES.get(owner);

	    if (srcList == null) {
		throw new IllegalArgumentException(notFoundErrorMsg);
	    }

	    String closureNrStr = closure.getClass().getName().replaceFirst(".*closure", "");

	    int closureNr = Integer.parseInt(closureNrStr);

	    if (closureNr > srcList.size()) {
		throw new IllegalArgumentException(notFoundErrorMsg);
	    }

	    closureSource = srcList.get(closureNr);

	    SOURCE_CACHE.put(closure, closureSource);
	    CLOSURE_CACHE.put(closureSource, closure);
	}
    }

}
