package at.ac.tuwien.vitalab.genesis2.plugin;

import groovy.lang.Closure;
import groovy.lang.GroovyObjectSupport;
import groovy.lang.MissingPropertyException;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.xml.ws.WebServiceException;

import org.apache.cxf.endpoint.Server;
import org.apache.cxf.jaxws.JaxWsProxyFactoryBean;
import org.apache.cxf.jaxws.JaxWsServerFactoryBean;
import org.apache.log4j.Logger;

import at.ac.tuwien.vitalab.genesis2.base.ExecutionEnvironment;
import at.ac.tuwien.vitalab.genesis2.config.Genesis2Configuration;
import at.ac.tuwien.vitalab.genesis2.core.model.AbstractModelElement;
import at.ac.tuwien.vitalab.genesis2.core.model.builder.AbstractModelBuilder;
import at.ac.tuwien.vitalab.genesis2.core.model.types.Host;
import at.ac.tuwien.vitalab.genesis2.core.model.ws.IModelElementWebService;
import at.ac.tuwien.vitalab.genesis2.core.model.ws.ModelElementWebService;
import at.ac.tuwien.vitalab.genesis2.errors.G2Exception;
import at.ac.tuwien.vitalab.genesis2.errors.G2LocalException;

/**
 * Abstract plugin class for Genesis2.
 * 
 * @author Lukasz Juszczyk
 * 
 * @param <T> Main type of element for this plugin.
 * 
 */
public abstract class AbstractGenesis2Plugin<T extends AbstractModelElement> extends GroovyObjectSupport {

    protected final Logger logger = Logger.getLogger(getClass()); // NOPMD

    // static registries

    private final static Map<String, AbstractGenesis2Plugin<? extends AbstractModelElement>> PLUGINS_BY_NAME = new LinkedHashMap<String, AbstractGenesis2Plugin<? extends AbstractModelElement>>();

    private final static Map<Class<? extends AbstractGenesis2Plugin<?>>, AbstractGenesis2Plugin<? extends AbstractModelElement>> PLUGINS_BY_CLASS = new LinkedHashMap<Class<? extends AbstractGenesis2Plugin<?>>, AbstractGenesis2Plugin<? extends AbstractModelElement>>();

    private final static Set<IPluginObserver> OBSERVERS = new HashSet<IPluginObserver>();

    private final static Map<Class<? extends AbstractModelElement>, Set<AbstractGenesis2Plugin<?>>> DECLARING_PLUGINS = new HashMap<Class<? extends AbstractModelElement>, Set<AbstractGenesis2Plugin<?>>>();

    private final static Map<AbstractGenesis2Plugin<?>, Set<Class<? extends AbstractModelElement>>> DECLARED_TYPES = new HashMap<AbstractGenesis2Plugin<?>, Set<Class<? extends AbstractModelElement>>>();

    // local variables

    private final Map<Class<? extends AbstractModelElement>, Server> modelWsEndpoints = new LinkedHashMap<Class<? extends AbstractModelElement>, Server>();

    private final Map<String, Object> cachedWsClients = new HashMap<String, Object>();

    // if this plugin offers some web service interface,
    // deploy it at this endpoint
    protected Server wsEndpoint = null;

    private final ExecutionEnvironment execEnv = ExecutionEnvironment.getGlobalInstance();

    /**
     * Returns the name of the plugin.
     * 
     * @return Name.
     */
    public abstract String getPluginName();

    /**
     * Registers a new plugin observer.
     * 
     * @param po Plugin observer instance.
     */
    public static synchronized void observe(IPluginObserver po) {

	if (OBSERVERS.add(po)) {
	    for (AbstractGenesis2Plugin<?> p : PLUGINS_BY_CLASS.values()) {
		po.onPluginStarted(p);
	    }
	}
    }

    /**
     * Deregisters a plugin observer.
     * 
     * @param po Plugin observer instance.
     */
    public static synchronized void stopObserving(IPluginObserver po) {
	OBSERVERS.remove(po);
    }

    /**
     * Starting procedure.
     */
    @SuppressWarnings("unchecked")
    public final void start() {

	try {

	    PLUGINS_BY_NAME.put(getPluginName(), this);
	    PLUGINS_BY_CLASS.put((Class<? extends AbstractGenesis2Plugin<?>>) getClass(), this); // NOPMD

	    deployWebService();
	    registerModelTypesAndWebServices();
	    registerAtExecutionEnvironment();
	    doStart();

	    for (IPluginObserver po : OBSERVERS) {
		po.onPluginStarted(this);
	    }

	} catch (Exception e) {

	    // clean up
	    try {
		stop();
	    } catch (Exception x) {
		// ignore
	    }

	    throw new G2LocalException("Cannot start plugin " + this.getPluginName() + ".", e);
	}
    }

    /**
     * Real method for starting the plugin code.
     * 
     * @throws Exception Exception.
     */
    protected abstract void doStart() throws Exception;

    /**
     * Stopping procedure.
     * 
     */
    public final void stop() {

	final String pluginName = getPluginName();

	PLUGINS_BY_NAME.remove(pluginName);
	PLUGINS_BY_CLASS.remove(getClass());

	for (IPluginObserver po : OBSERVERS) {
	    po.onPluginStopped(this);
	}

	try {
	    undeployWebService();
	} catch (Exception e) {
	    logger.warn("Cannot undeploy Web service of plugin '" + pluginName + "'.", e);
	}

	try {
	    doStop();
	} catch (Exception e) {
	    logger.warn("Cannot clean up plugin '" + pluginName + "'.", e);
	}

	try {
	    deregisterFromExecutionEnvironment();
	} catch (Exception e) {
	    logger.warn("Cannot deregister plugin '" + pluginName + "' from execution environment.", e);
	}

	try {
	    deregisterModelTypesAndWebServices();
	} catch (Exception e) {
	    logger.warn("Cannot undeploy Model Element Web services of failed plugin '" + pluginName + "'.", e);
	}
    }

    /**
     * Real method for stopping the plugin code.
     * 
     * @throws Exception Exception.
     */
    protected abstract void doStop() throws Exception;

    /**
     * Register an execution environment.
     */
    protected synchronized final void registerModelTypesAndWebServices() {

	Map<Class<? extends AbstractModelElement>, Class<? extends AbstractModelBuilder<? extends AbstractModelElement>>> typeBuilderMap = getTypesAndBuilders();

	if (typeBuilderMap == null || typeBuilderMap.isEmpty()) {
	    return;
	}

	Set<Class<? extends AbstractModelElement>> declaredTypes = typeBuilderMap.keySet();

	DECLARED_TYPES.put(this, declaredTypes);

	for (Class<? extends AbstractModelElement> typeClass : declaredTypes) {

	    Set<AbstractGenesis2Plugin<?>> declaredBy = DECLARING_PLUGINS.get(typeClass);

	    if (declaredBy == null) {
		declaredBy = new HashSet<AbstractGenesis2Plugin<?>>();
		DECLARING_PLUGINS.put(typeClass, declaredBy);
	    }

	    AbstractModelElement.declareElementType(typeClass, this);
	    AbstractModelBuilder.register(typeClass, typeBuilderMap.get(typeClass));

	    String name = typeClass.getSimpleName();

	    execEnv.setVariable(name.toLowerCase(), AbstractModelElement.getElementMap(typeClass));

	    logger.info("Registering model element '" + name + "' under alias [" + name.toLowerCase() + "]");

	    if (declaredBy.isEmpty()) {

		String url = getLocalModelWsURL(typeClass);

		try {

		    ModelElementWebService<?> implementor = new ModelElementWebService(
			    AbstractModelElement.getElementMap(typeClass));

		    // Endpoint e = Endpoint.publish(url, implementor);

		    JaxWsServerFactoryBean svrFactory = new JaxWsServerFactoryBean();
		    svrFactory.setAddress(url);
		    svrFactory.setServiceBean(implementor);
		    Server s = svrFactory.create();

		    modelWsEndpoints.put(typeClass, s);

		    logger.info("Model Element Web service deployed at '" + url + "'.");

		} catch (WebServiceException wse) {

		    logger.error("Cannot deploy model element Web service at '" + url + "': " + wse.getMessage());

		    throw wse;
		}
	    }

	    declaredBy.add(this);
	}
    }

    // deregister from execution environment
    protected synchronized final void deregisterModelTypesAndWebServices() {

	Set<Class<? extends AbstractModelElement>> declaredTypes = DECLARED_TYPES.get(this);

	if (declaredTypes == null) {
	    return;
	}

	DECLARED_TYPES.remove(this);

	for (Class<? extends AbstractModelElement> typeClass : declaredTypes) {

	    Set<AbstractGenesis2Plugin<?>> declaredBy = DECLARING_PLUGINS.get(typeClass);

	    declaredBy.remove(this);

	    if (declaredBy.isEmpty()) {

		Server wsEndpoint = modelWsEndpoints.remove(typeClass);

		wsEndpoint.stop();

		logger.info("Model Element Web service at '" + getLocalModelWsURL(typeClass) + "' stopped.");

		AbstractModelElement.undeclareElementType(typeClass);
		AbstractModelBuilder.deregister(typeClass);

		String name = typeClass.getSimpleName();

		execEnv.setVariable(name.toLowerCase(), null);

		logger.info("Deregistering model element '" + name + "' alias [" + name.toLowerCase() + "]");
	    }
	}
    }

    /**
     * Returns all types declared by this plugin.
     * 
     * @return Set of type
     */
    public final HashSet<Class<? extends AbstractModelElement>> getDeclaredTypes() {
	return new HashSet<Class<? extends AbstractModelElement>>(DECLARED_TYPES.get(this));
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Object#toString()
     */
    @Override
    public String toString() {
	return "Plugin[" + getPluginName() + "]";
    }

    /**
     * Returns the implementor class for this plugins's WS interface.
     * 
     * @return Implementing object or <code>null</code> if none has to be
     *         deployed.
     */
    public abstract Object getPluginWsImplementor();

    // if the plugin has a WS interface, deploy it
    private void deployWebService() throws WebServiceException {

	Object wsImpl = getPluginWsImplementor();

	if (wsImpl == null) {
	    return;
	}

	String url = getLocalPluginWsURL();

	try {

	    JaxWsServerFactoryBean svrFactory = new JaxWsServerFactoryBean();
	    svrFactory.setAddress(url);
	    svrFactory.setServiceBean(wsImpl);
	    wsEndpoint = svrFactory.create();

	    logger.info("Plugin Web service deployed at '" + url + "'.");

	} catch (WebServiceException e) {

	    logger.fatal("Cannot deploy plugin Web service at '" + url + "': " + e.getMessage());

	    wsEndpoint = null;

	    throw e;
	}
    }

    // undeploy WS interface
    private void undeployWebService() {
	if (wsEndpoint != null) {

	    wsEndpoint.stop();

	    logger.info("Plugin Web service at '" + getLocalPluginWsURL() + "' stopped.");

	    wsEndpoint = null;
	}
    }

    /**
     * Returns the Web service client.
     * 
     * @param host Host instance.
     * 
     * @return Client stubs or <code>null</code> if none is available.
     */
    public Object getPluginWsClient(Host host) {
	return getPluginWsClient(host.getHost(), host.getPort());
    }

    /**
     * Returns the Web service client.
     * 
     * @param host Host location.
     * @param port Port number.
     * 
     * @return Client stubs or <code>null</code> if none is available.
     */
    public Object getPluginWsClient(String host, int port) {

	String url = getPluginWsURL(host, port);

	Object cachedClient = cachedWsClients.get(url);

	if (cachedClient != null) {
	    return cachedClient;
	}

	Class<?> generatorClass = getPluginWsClientClass();
	if (generatorClass == null) {
	    return null;
	}

	Object result = generateWsClient(url, generatorClass);
	cachedWsClients.put(url, result);

	return result;
    }

    /**
     * Returns the Web service client for the model element web service.
     * 
     * @param host Host location.
     * @param port Port number.
     * @param elementClass Element class.
     * 
     * @return Client stubs or <code>null</code> if none is available.
     */
    public IModelElementWebService getModelWsClient(String host, int port,
	    Class<? extends AbstractModelElement> elementClass) {

	String url = getModelWsURL(host, port, elementClass);

	IModelElementWebService cachedClient = (IModelElementWebService) cachedWsClients.get(url);

	if (cachedClient != null) {
	    return cachedClient;
	}

	IModelElementWebService result = (IModelElementWebService) generateWsClient(url, IModelElementWebService.class);
	cachedWsClients.put(url, result);

	return result;

    }

    /*
     * Generates the Web service client.
     */
    private Object generateWsClient(String url, Class<?> generatorClass) {

	JaxWsProxyFactoryBean factory = new JaxWsProxyFactoryBean();

	factory.setServiceClass(generatorClass);
	factory.setAddress(url);

	return factory.create();
    }

    /**
     * Returns the class which generates clients stubs.
     * 
     * @return Class.
     */
    protected abstract Class<?> getPluginWsClientClass();

    /**
     * Returns the URL of the remote Web service. Can be overridden if it is not
     * the default URL.
     * 
     * @param host Host location.
     * @param port Port number.
     * 
     * @return URL string.
     */
    public String getPluginWsURL(String host, int port) {
	return "http://" + host + ":" + port + "/genesis2/plugin/" + getClass().getSimpleName();
    }

    /**
     * Returns the URL of the remote Web service. Can be overridden if it is not
     * the default URL.
     * 
     * @param host Host location.
     * @param port Port number.
     * @param elementClass Element class.
     * 
     * @return URL string.
     */
    public static String getModelWsURL(String host, int port, Class<? extends AbstractModelElement> elementClass) {
	return "http://" + host + ":" + port + "/genesis2/modelelement/" + elementClass.getSimpleName();
    }

    /**
     * Returns the URL of the local Web service.
     * 
     * @return URL string.
     */
    public String getLocalPluginWsURL() {

	String host = Genesis2Configuration.getLocalHost();
	int port = Genesis2Configuration.getLocalPort();

	return getPluginWsURL(host, port);
    }

    /**
     * Returns the URL of the local Element Model Web service.
     * 
     * @param elementClass Class of the model element.
     * 
     * @return URL string.
     */
    public static String getLocalModelWsURL(Class<? extends AbstractModelElement> elementClass) {

	String host = Genesis2Configuration.getLocalHost();
	int port = Genesis2Configuration.getLocalPort();

	return getModelWsURL(host, port, elementClass);
    }

    /**
     * Register this plugin at the global script execution environment.
     */
    protected void registerAtExecutionEnvironment() {

	List<String> aliases = getAliases();

	if (aliases != null && !aliases.isEmpty()) {

	    logger.info("Registering plugin '" + getPluginName() + "' under aliases " + aliases);

	    for (String alias : aliases) {
		execEnv.setVariable(alias, this);
	    }
	}

	Map<String, Closure> commands = getCommands();

	if (commands != null && !commands.isEmpty()) {

	    for (String name : commands.keySet()) {
		Closure c = commands.get(name);

		logger.info("Registering command '" + name + "' of plugin '" + getPluginName() + "'");

		execEnv.setVariable(name, c);
	    }
	}
    }

    /**
     * Unregister this plugin from the global script execution environment.
     */
    protected void deregisterFromExecutionEnvironment() {

	List<String> aliases = getAliases();

	if (aliases != null && !aliases.isEmpty()) {

	    logger.info("Deregistering plugin's '" + getPluginName() + "' aliases " + aliases);

	    for (String alias : aliases) {

		Object var = null;
		try {
		    var = execEnv.getVariable(alias);
		} catch (MissingPropertyException mpe) {
		    continue; // nothing to do
		}

		// check class in case that variable has been overwritten
		if (var != null && var.getClass().equals(getClass())) {
		    execEnv.setVariable(alias, null);
		}
	    }
	}

	Map<String, Closure> commands = getCommands();

	if (commands != null && !commands.isEmpty()) {
	    for (String name : commands.keySet()) {

		logger.info("Deregistering command '" + name + "' of plugin '" + getPluginName() + "'");

		execEnv.setVariable(name, null);
	    }
	}
    }

    /**
     * Declares relationships.
     * 
     * @return Array of declared model types.
     */
    public abstract Map<Class<? extends AbstractModelElement>, Class<? extends AbstractModelBuilder<? extends AbstractModelElement>>> getTypesAndBuilders();

    /**
     * Returns a list of aliases, defined by the plugin to be used.
     * 
     * @return List of alias names.
     */
    public abstract List<String> getAliases();

    /**
     * Returns a map of commands which should be registered at the execution
     * environment.
     * 
     * @return Map of command names and commands.
     */
    public abstract Map<String, Closure> getCommands();

    /**
     * Return a generated artifact.
     * 
     * @param <S> Type of artifact.
     * 
     * @param type Type of artifact.
     * @param modelelement Model element.
     * @param additionalIDs Additional information for querying the artifact.
     * 
     * @return Artifact object or <code>null</code> if nothing found.
     * 
     */
    public abstract <S> S getArtifact(Class<S> type, T modelelement, String... additionalIDs);

    /**
     * Deploys at backend.
     * 
     * @param e Element.
     * 
     * @throws IllegalArgumentException On faulty arguments.
     * @throws G2Exception On internal errors.
     * 
     */
    public abstract void deployElement(T e) throws IllegalArgumentException, G2Exception;

    /**
     * Undeployes from backend.
     * 
     * @param e Element
     * 
     * @throws IllegalArgumentException On faulty arguments.
     * @throws G2Exception On internal errors.
     */
    public abstract void undeployElement(T e) throws IllegalArgumentException, G2Exception;

    /**
     * Updates a deployed element.
     * 
     * @param e Element.
     * 
     * @throws IllegalArgumentException On faulty arguments.
     * @throws G2Exception On internal errors.
     */
    public final void redeployElement(T e) throws IllegalArgumentException, G2Exception {
	undeployElement(e);
	deployElement(e);
    }

    /**
     * Returns the set of currently running plugins.
     * 
     * @return Set of plugins.
     */
    public static Set<AbstractGenesis2Plugin<? extends AbstractModelElement>> getRunningPlugins() {
	return new HashSet<AbstractGenesis2Plugin<? extends AbstractModelElement>>(PLUGINS_BY_CLASS.values());
    }

    /**
     * Checks whether the plugin has been installed.
     * 
     * @param name Plugin name.
     * 
     * @return <code>true</code> if installed, otherwise <code>false</code>.
     */
    public static boolean hasPluginInstalled(String name) {
	return getPlugin(name) != null;
    }

    /**
     * Checks whether the plugin has been installed.
     * 
     * @param clazz Plugin class.
     * 
     * @return <code>true</code> if installed, otherwise <code>false</code>.
     */
    public static boolean hasPluginInstalled(
	    Class<? extends AbstractGenesis2Plugin<? extends AbstractModelElement>> clazz) {
	return PLUGINS_BY_CLASS.containsKey(clazz);
    }

    /**
     * Returns the plugin with that name.
     * 
     * @param name Plugin name.
     * 
     * @return Plugin instance or <code>null</code> if nothing got found.
     */
    public static AbstractGenesis2Plugin<? extends AbstractModelElement> getPlugin(String name) {
	return PLUGINS_BY_NAME.get(name);
    }

    /**
     * Returns the plugin of that class.
     * 
     * @param <S> Type of element.
     * 
     * @param clazz Plugin clazz.
     * 
     * @return Plugin instance or <code>null</code> if nothing got found.
     */
    public static <S extends AbstractGenesis2Plugin<? extends AbstractModelElement>> S getPlugin(Class<S> clazz) {
	return (S) PLUGINS_BY_CLASS.get(clazz); // safe
    }

    /**
     * Starts a plugin by its class name.
     * 
     * @param className Plugin class name.
     * @return Plugin instance.
     */
    @SuppressWarnings("unchecked")
    public static AbstractGenesis2Plugin<? extends AbstractModelElement> startPlugin(String className) {

	Class<AbstractGenesis2Plugin<AbstractModelElement>> pluginClass;
	try {
	    pluginClass = (Class<AbstractGenesis2Plugin<AbstractModelElement>>) Class.forName(className);
	} catch (ClassNotFoundException e1) {
	    throw new IllegalArgumentException("Unknown plugin class " + className + ".");
	}

	if (PLUGINS_BY_CLASS.containsKey(pluginClass)) {
	    return PLUGINS_BY_CLASS.get(pluginClass);
	}

	AbstractGenesis2Plugin<AbstractModelElement> pluginInstance;
	try {
	    pluginInstance = pluginClass.newInstance();
	} catch (Exception e) {
	    throw new G2LocalException("Cannot instantiate plugin " + pluginClass.getName() + ".", e);
	}

	pluginInstance.start();

	return pluginInstance;
    }

    /**
     * Call returning the CallProxy on which WS calls will be invoked.
     * 
     * @param c1 Filter closure.
     * 
     * @return CallProxy instance.
     */
    public final CallProxy callWs(Closure c1) {

	return new CallProxy(this, c1);
    }

    /**
     * Call returning the result of the invocations.
     * 
     * @param c1 Filter closure.
     * @param c2 Command closure.
     * 
     * @return Invocation result(s).
     */
    public final Object callWs(Closure c1, Closure c2) {

	CallProxy cp = new CallProxy(this, c1);
	return cp.invokeMethod("call", new Object[] { c2 });
    }

    /**
     * Call returning the result of the invocations.
     * 
     * @param host Host.
     * @param c Command closure.
     * 
     * @return CallProxy instance.
     */
    public final Object callWs(Host host, Closure c) {

	ArrayList<Host> hosts = new ArrayList<Host>();
	hosts.add(host);

	CallProxy cp = new CallProxy(this, hosts);
	return cp.invokeMethod("call", new Object[] { c });
    }

    /**
     * Call returning the CallProxy for the passed host.
     * 
     * @param host Host.
     * 
     * @return CallProxy instance.
     */
    public final CallProxy callWs(Host host) {

	ArrayList<Host> hosts = new ArrayList<Host>();
	hosts.add(host);

	return new CallProxy(this, hosts);
    }

    /**
     * Call returning the CallProxy for the passed hosts.
     * 
     * @param hosts Hosts.
     * 
     * @return CallProxy instance.
     */
    public final CallProxy callWs(Collection<Host> hosts) {

	return new CallProxy(this, hosts);
    }

    /**
     * Call returning the result of the invocations.
     * 
     * @param hosts Hosts.
     * @param c Command closure.
     * 
     * @return Invocation result(s).
     */
    public final Object callWs(Collection<Host> hosts, Closure c) {

	CallProxy cp = new CallProxy(this, hosts);
	return cp.invokeMethod("call", new Object[] { c });
    }

}
