package at.ac.tuwien.vitalab.genesis2.plugin.callinterceptor;

import groovy.lang.Binding;
import groovy.lang.Closure;
import groovy.util.Expando;

import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;

import org.apache.cxf.binding.soap.SoapMessage;
import org.apache.cxf.binding.soap.interceptor.AbstractSoapInterceptor;
import org.apache.cxf.interceptor.Fault;
import org.apache.cxf.interceptor.Interceptor;
import org.apache.cxf.interceptor.InterceptorProvider;
import org.apache.cxf.phase.Phase;

import at.ac.tuwien.vitalab.genesis2.core.model.AbstractModelElement;
import at.ac.tuwien.vitalab.genesis2.utils.ClosureUtil;

/**
 * Model for SOAP Call interceptors.
 * 
 * @author Lukasz Juszczyk
 * 
 */
@SuppressWarnings("PMD.UncommentedEmptyMethod")
public class CallInterceptor extends AbstractModelElement {

    // keywords
    /**
     * Ingoing body element.
     */
    public static final String IN = "in";

    /**
     * Ingoing fault.
     */
    public static final String INFAULT = "inFault";

    /**
     * Outgoing body element.
     */
    public static final String OUT = "out";

    /**
     * Outgoing fault.
     */
    public static final String OUTFAULT = "outFault";

    /**
     * Names of hooks.
     */
    public static final String[] HOOKNAMES = new String[] { IN, INFAULT, OUT, OUTFAULT };

    private static final ClosureUtil CLOSURE_UTIL = ClosureUtil.getInstance();

    /**
     * Constructor.
     */
    public CallInterceptor() {
	super();

	followUpProps = false;
    }

    /**
     * Constructor for the builder.
     * 
     * @param name Name of interceptor, will be ignored.
     */
    public CallInterceptor(String name) {
	super();
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.core.model.ICleanable#cleanup()
     */
    @Override
    public void cleanup() {

    }

    /**
     * Get the value of the phase declaration.
     * 
     * @param phaseName Phase name.
     * @return Value.
     */
    private String extractPhase(String phaseName) {

	try {
	    Field f = Phase.class.getDeclaredField(phaseName);
	    return (String) f.get(null);
	} catch (Exception e) {
	    // ignore
	}

	return phaseName;
    }

    /**
     * Binds the interceptor for an endpoint (service or client).
     * 
     * @param ip Endpoint.
     * @param parent Parent element for parameter lookups.
     * @param binding Additional Groovy binding for interceptor code.
     */
    public void attachToEndpoint(InterceptorProvider ip, AbstractModelElement parent, Binding binding) {

	Map<String, String> hooks = getHooks();

	String phase = null;

	if ((phase = extractPhase(hooks.get(IN))) != null) {
	    ip.getInInterceptors().add(new G2SoapInterceptor(this, parent, ip, IN, phase, binding));
	}

	if ((phase = extractPhase(hooks.get(OUT))) != null) {
	    ip.getOutInterceptors().add(new G2SoapInterceptor(this, parent, ip, OUT, phase, binding));
	}

	if ((phase = extractPhase(hooks.get(INFAULT))) != null) {
	    ip.getInFaultInterceptors().add(new G2SoapInterceptor(this, parent, ip, INFAULT, phase, binding));
	}

	if ((phase = extractPhase(hooks.get(OUTFAULT))) != null) {
	    ip.getOutFaultInterceptors().add(new G2SoapInterceptor(this, parent, ip, OUTFAULT, phase, binding));
	}
    }

    /**
     * Unbind the interceptor from endpoint (service or client).
     * 
     * @param ip Endpoint.
     */
    public void detachFromEndpoint(InterceptorProvider ip) {
	removeThisFromList(ip.getInInterceptors());
	removeThisFromList(ip.getOutInterceptors());
	removeThisFromList(ip.getInFaultInterceptors());
	removeThisFromList(ip.getOutFaultInterceptors());
    }

    @SuppressWarnings("rawtypes")
    private void removeThisFromList(List<Interceptor> interceptors) {

	ListIterator<Interceptor> iter = interceptors.listIterator();
	List<Interceptor> toDel = new ArrayList<Interceptor>();

	while (iter.hasNext()) {

	    Interceptor<?> i = iter.next();

	    if (i instanceof G2SoapInterceptor) {
		G2SoapInterceptor gi = (G2SoapInterceptor) i;

		if (gi.originalInterceptor == this) { // remove old one if
						      // matches
		    toDel.add(i);
		}
	    }
	}

	interceptors.removeAll(toDel);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.core.model.AModelElement#processProperty
     * (java.lang.String, java.lang.Object)
     */
    @Override
    protected Object processProperty(String propName, Object prop) {

	// we'll proces later
	return prop;
    }

    /**
     * Getter for interceptor code.
     * 
     * @return Code closure.
     */
    public Closure getCode() {
	return (Closure) getProperty("code");
    }

    /**
     * Setter for interceptor code.
     * 
     * @param code Code closure.
     */
    public void setCode(Closure code) {
	super.setProperty("code", code);
    }

    /**
     * Execute the closure in the context.
     * 
     * @param context Context.
     * @param closure Closure.
     */
    public void execute(InterceptorContext context, Closure closure) {
	if (closure != null) {
	    closure.call(context);
	}
    }

    /**
     * Returns a map of all hooks.
     * 
     * @return Hooks.
     */
    public Map<String, String> getHooks() {
	Map<String, String> result = (Map<String, String>) getProperty("hooks");
	if (result != null) {
	    return result;
	}
	return new HashMap<String, String>();
    }

    /**
     * Setter for hook map.
     * 
     * @param hooks Hooks,
     */
    public void setHooks(Map<String, String> hooks) {
	super.setProperty("hooks", hooks);
    }

    /**
     * Gets name of CXF processing phases AFTER that this interceptor should be
     * invoked.
     * 
     * @return List of phase names.
     */
    public List<String> getAfter() {
	List<String> result = (List<String>) getProperty("after");
	if (result != null) {
	    return result;
	}
	return new ArrayList<String>();
    }

    /**
     * Sets name of CXF processing phases AFTER that this interceptor should be
     * invoked.
     * 
     * @param afterSet List of phase names.
     */
    public void setAfter(List<?> afterSet) {
	setProperty("after", processClassSets(afterSet));
    }

    /**
     * Gets name of CXF processing phases BEFORE that this interceptor should be
     * invoked.
     * 
     * @return List of phase names.
     */
    public List<String> getBefore() {
	List<String> result = (List<String>) getProperty("before");
	if (result != null) {
	    return result;
	}
	return new ArrayList<String>();
    }

    /**
     * Sets name of CXF processing phases BEFORE that this interceptor should be
     * invoked.
     * 
     * @param beforeSet List of phase names.
     */
    public void setBefore(List<?> beforeSet) {
	setProperty("before", processClassSets(beforeSet));
    }

    private ArrayList<String> processClassSets(List<?> set) {

	ArrayList<String> realSet = new ArrayList<String>();

	if (set != null) {
	    for (Object o : set) {
		if (o == null) {
		    continue;
		}
		if (o instanceof Class) {
		    realSet.add(((Class<?>) o).getName());
		} else {
		    realSet.add(o.toString());
		}
	    }
	}

	return realSet;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.core.model.AModelElement#setProperty(java
     * .lang.String, java.lang.Object)
     */
    @Override
    public synchronized void setProperty(String property, Object newValue) {

	if ("after".equals(property)) {
	    setAfter((List<String>) newValue);
	} else if ("before".equals(property)) {
	    setBefore((List<String>) newValue);
	} else {
	    super.setProperty(property, newValue);
	}
    }

    /**
     * The actual interceptor.
     * 
     * @author Lukasz Juszczyk
     * 
     */
    private class G2SoapInterceptor extends AbstractSoapInterceptor {

	private CallInterceptor clonedInterceptor = null;
	private CallInterceptor originalInterceptor = null;
	private AbstractModelElement parent = null;

	private InterceptorProvider provider = null;
	private String hookName = null;
	private String phase = null;

	protected G2SoapInterceptor(CallInterceptor interceptor, AbstractModelElement parent,
		InterceptorProvider provider, String hookName, String phase, Binding variableBinding) {

	    super(interceptor.getId(), phase);

	    try {
		this.originalInterceptor = interceptor;
		this.clonedInterceptor = (CallInterceptor) interceptor.clone();
	    } catch (CloneNotSupportedException e) {
		// should never happen
		e.printStackTrace();
	    }

	    this.parent = parent;

	    this.provider = provider;
	    this.hookName = hookName;
	    this.phase = phase;

	    List<String> after = interceptor.getAfter();
	    if (after != null) {
		addAfter(after);
	    }

	    List<String> before = interceptor.getBefore();
	    if (before != null) {
		addBefore(before);
	    }
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.apache.cxf.interceptor.Interceptor#handleMessage(org.apache.cxf
	 * .message.Message)
	 */
	@Override
	public void handleMessage(SoapMessage msg) throws Fault {

	    try {
		InterceptorContext context = new InterceptorContext();

		context.setProvider(provider);
		context.setHookName(hookName);
		context.setPhaseName(phase);
		context.setSoapMsg(msg);

		// preprocess
		context = clonedInterceptor.preprocessContext(context);

		// process
		Closure code = getCode();

		if (code != null) {

		    code = (Closure) code.clone();

		    if (parent != null) { // for the sake of model reflection
			CLOSURE_UTIL.attachDelegates(code, this, parent);
		    } else {
			CLOSURE_UTIL.attachDelegates(code, this);
		    }

		    clonedInterceptor.execute(context, code);
		}

		// and postprocess
		clonedInterceptor.postprocessContext(context);

	    } catch (Exception e) {
		throw new Fault(e);
	    }

	}

	public CallInterceptor getInterceptor() {
	    return clonedInterceptor;
	}
    }

    /**
     * Preprocessor.
     * 
     * @param context Original context.
     * @return Processed context.
     */
    protected InterceptorContext preprocessContext(InterceptorContext context) {
	return context; // by default, do nothing. overridable.
    }

    /**
     * Postprocessor.
     * 
     * @param context Original context.
     * @return Processed context.
     */
    protected InterceptorContext postprocessContext(InterceptorContext context) {
	return context; // by default, do nothing. overridable.
    }

    @SuppressWarnings({ "PMD.UncommentedEmptyMethod", "javadoc" })
    public class InterceptorContext extends Expando {

	public InterceptorContext() {
	    super();
	}

	public InterceptorContext(InterceptorContext src) {
	    this.getProperties().putAll(src.getProperties());
	}

	public SoapMessage getSoapMsg() {
	    return (SoapMessage) getProperty("soapMsg");
	}

	public void setSoapMsg(SoapMessage soapMsg) {
	    super.setProperty("soapMsg", soapMsg);
	}

	public InterceptorProvider getProvider() {
	    return (InterceptorProvider) getProperty("provider");
	}

	public void setProvider(InterceptorProvider provider) {
	    super.setProperty("provider", provider);
	}

	public String getHookName() {
	    return (String) getProperty("hookName");
	}

	public void setHookName(String hookName) {
	    super.setProperty("hookName", hookName);
	}

	public String getPhaseName() {
	    return (String) getProperty("phaseName");
	}

	public void setPhaseName(String phaseName) {
	    super.setProperty("phaseName", phaseName);
	}

    }
}
