package at.ac.tuwien.vitalab.genesis2.plugin.clientgen;

import groovy.lang.Closure;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import at.ac.tuwien.vitalab.genesis2.core.model.AbstractModelElement;
import at.ac.tuwien.vitalab.genesis2.core.model.builder.AbstractModelBuilder;
import at.ac.tuwien.vitalab.genesis2.errors.G2Exception;
import at.ac.tuwien.vitalab.genesis2.plugin.AbstractStandaloneGenesis2Plugin;

/**
 * G2 plugin for generating client code as threads.
 * 
 * @author Lukasz Juszczyk
 * 
 */
public class ClientGeneratorPlugin extends AbstractStandaloneGenesis2Plugin<Client> {

    private final Map<Client, Thread> deployedClients = new HashMap<Client, Thread>();

    /**
     * Plugin name.
     */
    public static final String PLUGINNAME = "ClientGenerator";

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getPluginName()
     */
    @Override
    public String getPluginName() {
	return PLUGINNAME;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#deployElement(at
     * .ac.tuwien.vitalab.genesis2.core.model.AModelElement)
     */
    @Override
    public void deployElement(final Client client) throws IllegalArgumentException, G2Exception {

	// cleanup previously deployed instance
	undeployElement(client);

	final Closure closure = client.getCode();

	if (closure == null) {
	    return;
	}

	Thread t = new Thread() {
	    public void run() {
		try {
		    closure.call();
		} catch (Exception e) {
		    logger.warn("Client died.", e);
		}
	    }
	};

	deployedClients.put(client, t);

	t.setDaemon(true);
	t.start();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#undeployElement(
     * at.ac.tuwien.vitalab.genesis2.core.model.AModelElement)
     */
    @Override
    public void undeployElement(Client e) {

	Thread t = deployedClients.remove(e);

	if (t == null || !t.isAlive()) {
	    return;
	}

	try {
	    t.interrupt();
	} catch (Exception ex) {
	    // ignore
	}

    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getArtifact(java
     * .lang.Class, at.ac.tuwien.vitalab.genesis2.core.model.AModelElement,
     * java.lang.String[])
     */
    @SuppressWarnings("unchecked")
    @Override
    public <S> S getArtifact(Class<S> type, Client modelelement, String... additionalIDs) {

	if (type.equals(Thread.class)) {
	    try {
		return (S) deployedClients.get(modelelement);
	    } catch (NullPointerException npe) {
	    }
	}

	return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#doStart()
     */
    @Override
    protected void doStart() throws Exception {
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#doStop()
     */
    @Override
    protected void doStop() throws Exception {

	for (Client c : deployedClients.keySet()) {
	    undeployElement(c);
	}
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getAliases()
     */
    @Override
    public List<String> getAliases() {

	Vector<String> aliases = new Vector<String>();
	aliases.add("clientgen");
	return aliases;
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getCommands()
     */
    @Override
    public HashMap<String, Closure> getCommands() {
	return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getTypesAndBuilders
     * ()
     */
    @Override
    public Map<Class<? extends AbstractModelElement>, Class<? extends AbstractModelBuilder<? extends AbstractModelElement>>> getTypesAndBuilders() {

	Map<Class<? extends AbstractModelElement>, Class<? extends AbstractModelBuilder<? extends AbstractModelElement>>> result = new HashMap<Class<? extends AbstractModelElement>, Class<? extends AbstractModelBuilder<? extends AbstractModelElement>>>();

	result.put(Client.class, null);

	return result;
    }

}
