package at.ac.tuwien.vitalab.genesis2.plugin.hostlist;

import groovy.lang.Closure;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;

import at.ac.tuwien.vitalab.genesis2.Genesis2;
import at.ac.tuwien.vitalab.genesis2.core.model.AbstractModelElement;
import at.ac.tuwien.vitalab.genesis2.core.model.ElementMap;
import at.ac.tuwien.vitalab.genesis2.core.model.IModelObserver;
import at.ac.tuwien.vitalab.genesis2.core.model.builder.AbstractModelBuilder;
import at.ac.tuwien.vitalab.genesis2.core.model.types.Host;
import at.ac.tuwien.vitalab.genesis2.plugin.AbstractGenesis2Plugin;
import at.ac.tuwien.vitalab.genesis2.plugin.hostlist.ws.HostListWebService;
import at.ac.tuwien.vitalab.genesis2.plugin.hostlist.ws.IHostListWebService;

/**
 * 
 * Plugin providing up-to-date list of back-end service locations.
 * 
 * @author Lukasz Juszczyk
 * 
 */
@SuppressWarnings("PMD.UncommentedEmptyMethod")
public class HostListPlugin extends AbstractGenesis2Plugin<AbstractModelElement> {

    private HostObserver hostObserver = null;

    private final ElementMap<Host> hostMap = AbstractModelElement.getElementMap(Host.class);

    protected final Set<String> knownHosts = Collections.synchronizedSet(new HashSet<String>());

    /**
     * Plugin name.
     */
    public static final String PLUGINNAME = "HostListPlugin";

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getPluginName()
     */
    @Override
    public String getPluginName() {
	return PLUGINNAME;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#deployElement(at
     * .ac.tuwien.vitalab.genesis2.core.model.AModelElement)
     */
    @Override
    public void deployElement(AbstractModelElement e) {
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#doStart()
     */
    @Override
    protected void doStart() {
	if (Genesis2.isFrontend()) {
	    hostMap.observe(hostObserver = new HostObserver());
	}
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#doStop()
     */
    @Override
    protected void doStop() {
	if (Genesis2.isFrontend() && hostObserver != null) {
	    hostMap.stopObserving(hostObserver);
	}
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getTypesAndBuilders
     * ()
     */
    @Override
    public Map<Class<? extends AbstractModelElement>, Class<? extends AbstractModelBuilder<? extends AbstractModelElement>>> getTypesAndBuilders() {
	return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getAliases()
     */
    @Override
    public List<String> getAliases() {
	return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getCommands()
     */
    @Override
    public Map<String, Closure> getCommands() {
	return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getPluginWsClientClass
     * ()
     */
    @Override
    protected Class<?> getPluginWsClientClass() {
	return IHostListWebService.class;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getPluginWsImplementor
     * ()
     */
    @Override
    public Object getPluginWsImplementor() {
	return new HostListWebService(this, knownHosts);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#undeployElement(
     * at.ac.tuwien.vitalab.genesis2.core.model.AModelElement)
     */
    @Override
    public void undeployElement(AbstractModelElement e) {
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getArtifact(java
     * .lang.Class, at.ac.tuwien.vitalab.genesis2.core.model.AModelElement,
     * java.lang.String[])
     */
    @Override
    public <S> S getArtifact(Class<S> type, AbstractModelElement modelelement, String... additionalIDs) {
	return null;
    }

    /**
     * Extract the host string from a host:port tuple.
     * 
     * @param hostAndPort Tuple.
     * @return Host string.
     */
    public static String getHost(String hostAndPort) {
	StringTokenizer tok = new StringTokenizer(hostAndPort, ":");
	return tok.nextToken();
    }

    /**
     * Extraxt the port number from a host:port tuple.
     * 
     * @param hostAndPort Tuple.
     * @return Port number.
     */
    public static int getPort(String hostAndPort) {
	StringTokenizer tok = new StringTokenizer(hostAndPort, ":");
	tok.nextToken(); // host
	return Integer.parseInt(tok.nextToken());
    }

    /**
     * Returns the current list of host:port tuples.
     * 
     * @return Tuples.
     */
    public List<String> getList() {
	return new ArrayList<String>(knownHosts);
    }

    /**
     * Watches for host references and updates the already known ones.
     * 
     * @author Lukasz Juszczyk
     * 
     */
    private class HostObserver implements IModelObserver<Host> {

	/**
	 * Notify other hosts.
	 */
	private void update() {

	    List<Thread> threads = new ArrayList<Thread>();

	    for (final String location : knownHosts) {

		Thread t = new Thread() {

		    public void run() {
			String host = getHost(location);
			int port = getPort(location);

			try {
			    IHostListWebService ws = (IHostListWebService) getPluginWsClient(host, port);
			    ws.setHosts(new ArrayList<String>(knownHosts));
			} catch (Exception e) {
			    logger.error("Cannot update host list on host " + location);
			}
		    }
		};

		t.start();
		threads.add(t);
	    }

	    for (Thread th : threads) {
		try {
		    th.join();
		} catch (InterruptedException e) {
		}
	    }
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see at.ac.tuwien.vitalab.genesis2.core.model.IModelObserver#
	 * changedModelElement
	 * (at.ac.tuwien.vitalab.genesis2.core.model.AModelElement, int,
	 * java.lang.Object[])
	 */
	@Override
	public synchronized void onChangedModelElement(Host host, int how, Object... info) {

	    if (how != IModelObserver.CREATED && how != IModelObserver.REMOVED) {
		return;
	    }

	    if (host.isLocalhost()) {
		return;
	    }

	    int hash = knownHosts.hashCode();

	    switch (how) {
	    case IModelObserver.CREATED:
		knownHosts.add(host.getHost() + ":" + host.getPort());
		break;

	    case IModelObserver.REMOVED:
		knownHosts.remove(host.getHost() + ":" + host.getPort());
		break;
	    default:
	    }

	    // if something changed
	    if (hash != knownHosts.hashCode()) {
		update();
	    }
	}
    }
}
