package at.ac.tuwien.vitalab.genesis2.plugin.msgpert;

import groovy.lang.Closure;
import groovy.xml.DOMBuilder;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.apache.cxf.helpers.IOUtils;
import org.apache.cxf.io.CachedOutputStream;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

import at.ac.tuwien.vitalab.genesis2.errors.G2Exception;
import at.ac.tuwien.vitalab.genesis2.plugin.callinterceptor.CallInterceptor;

/**
 * Message perturbation/corruption interceptor.
 * 
 * @author Lukasz Juszczyk
 * 
 */
public class MsgPertInterceptor extends CallInterceptor {

    // which kind of data shall be corrupterd?
    // xml data?
    private static final int CORRTYPE_XML = 0;
    // unmarshalled java object?
    private static final int CORRTYPE_ARGS = 1;
    // or directly at the byte level?
    private static final int CORRTYPE_BYTES = 2;

    /**
     * Constructor.
     */
    public MsgPertInterceptor() {
	super();
    }

    /**
     * Constructor.
     * 
     * @param type Corruption type.
     */
    public MsgPertInterceptor(String type) {
	this();

	if (type.equalsIgnoreCase("args")) {
	    setType(CORRTYPE_ARGS);
	} else if (type.equalsIgnoreCase("xml") || type.equalsIgnoreCase("dom")) {
	    setType(CORRTYPE_XML);
	} else {
	    setType(CORRTYPE_BYTES);
	}
    }

    /**
     * Gets the corruption type.
     * 
     * @return Corruption type.
     */
    public int getType() {
	try {
	    return ((Integer) getProperty("intDataType")).intValue();
	} catch (Exception e) {
	    return CORRTYPE_XML;
	}
    }

    /**
     * Sets the corruption type.
     * 
     * @param type Corruption type.
     */
    public final void setType(int type) {
	setProperty("intDataType", Integer.valueOf(type));
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.plugin.callinterceptor.CallInterceptor#
     * getHooks()
     */
    @Override
    public final Map<String, String> getHooks() {
	HashMap<String, String> hooks = new HashMap<String, String>();

	String inPhase = null;
	String outPhase = null;

	if (getType() == CORRTYPE_ARGS) {
	    inPhase = "PRE_LOGICAL";
	    outPhase = "MARSHAL";

	} else { // BYTES AND XML
	    inPhase = "RECEIVE";
	    outPhase = "PRE_STREAM";

	}

	hooks.put("in", inPhase);
	hooks.put("inFault", inPhase);

	hooks.put("out", outPhase);
	hooks.put("outFault", outPhase);

	return hooks;
    }

    /**
     * Workaround for an output stream which writes back to the message and
     * executes the corruption routines.
     * 
     * @author Lukasz Juszczyk
     * 
     */
    private class CorruptingOutputStream extends OutputStream {

	private final ByteArrayOutputStream baos = new ByteArrayOutputStream();

	private OutputStream os = null;
	private MsgPertContext msgCtx = null;
	private MsgPertInterceptor mpInter = null;

	/**
	 * Constructor.
	 * 
	 * @param os Original output stream.
	 * @param msgCtx Context.
	 * @param mpInter Corrupting interceptor.
	 */
	public CorruptingOutputStream(OutputStream os, MsgPertContext msgCtx, MsgPertInterceptor mpInter) {
	    super();
	    this.os = os;
	    this.msgCtx = msgCtx;
	    this.mpInter = mpInter;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.io.OutputStream#write(int)
	 */
	@Override
	public void write(int b) throws IOException {
	    baos.write(b);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.io.OutputStream#close()
	 */
	@Override
	public void close() throws IOException {
	    baos.close();

	    String str = baos.toString();

	    if (getType() == CORRTYPE_XML) {
		try {
		    Document doc = DOMBuilder.parse(new StringReader(str));
		    msgCtx.setData("xml", doc.getDocumentElement());
		} catch (Exception e) {
		    throw new IOException(e);
		}
	    } else {
		msgCtx.setData("bytes", str);
	    }

	    mpInter.execute(msgCtx, null);

	    if (getType() == CORRTYPE_XML) {

		Node n = (Node) msgCtx.getData("xml");

		try {
		    str = docToString(n);
		    os.write(str.getBytes());
		} catch (Exception e) {
		    throw new IOException(e);
		}

	    } else {
		os.write(((String) msgCtx.getData("bytes")).getBytes());
	    }

	    os.flush();
	    os.close();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.io.OutputStream#flush()
	 */
	@Override
	public void flush() throws IOException {
	    os.flush();
	}

    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.plugin.callinterceptor.CallInterceptor#
     * postprocessContext
     * (at.ac.tuwien.vitalab.genesis2.plugin.callinterceptor.CallInterceptor
     * .InterceptorContext)
     */
    @Override
    protected InterceptorContext postprocessContext(InterceptorContext context) {

	MsgPertContext msgCtx = new MsgPertContext(context);

	if (getType() == CORRTYPE_ARGS) {
	    List<?> bodyObjs = (ArrayList<?>) msgCtx.getData("args");

	    if (bodyObjs != null) {

		Vector<Object> v = new Vector<Object>();
		v.addAll(bodyObjs);

		msgCtx.getSoapMsg().setContent(List.class, v);
	    }

	} else if (!msgCtx.getHookName().startsWith("out")) {

	    String str;

	    if (getType() == CORRTYPE_XML) {
		Node n = (Node) msgCtx.getData("xml");
		try {
		    str = docToString(n);
		} catch (TransformerException e) {
		    // should actually not happen...
		    throw new G2Exception("Cannot transform DOM tree into byte array.", e);
		}
	    } else {
		str = (String) msgCtx.getData("bytes");
	    }

	    msgCtx.getSoapMsg().setContent(InputStream.class, new ByteArrayInputStream(str.getBytes()));
	}

	return super.postprocessContext(msgCtx);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.plugin.callinterceptor.CallInterceptor#
     * preprocessContext
     * (at.ac.tuwien.vitalab.genesis2.plugin.callinterceptor.CallInterceptor
     * .InterceptorContext)
     */
    @Override
    protected InterceptorContext preprocessContext(InterceptorContext context) {

	MsgPertContext msgCtx = new MsgPertContext(context);

	if (getType() == CORRTYPE_ARGS) {

	    if (msgCtx.getSoapMsg().getContent(List.class) != null) {

		List<?> args = msgCtx.getSoapMsg().getContent(List.class);

		msgCtx.setData("args", new ArrayList<Object>(args));
	    }

	} else {

	    if (msgCtx.getHookName().startsWith("out")) { // response message

		OutputStream oldOutput = context.getSoapMsg().getContent(OutputStream.class);
		OutputStream os = new CorruptingOutputStream(oldOutput, msgCtx, this);

		context.getSoapMsg().setContent(OutputStream.class, os);

	    } else { // request message
		InputStream input = msgCtx.getSoapMsg().getContent(InputStream.class);

		CachedOutputStream output = new CachedOutputStream();

		String str;
		try {
		    IOUtils.copy(input, output);
		    input.close();
		    output.close();

		    str = output.getOut().toString();
		    msgCtx.getSoapMsg().setContent(InputStream.class, output.getInputStream());
		} catch (IOException e) {
		    throw new G2Exception("IO errors while copying streams.", e);
		}

		if (getType() == CORRTYPE_XML) {

		    Document doc;
		    try {
			doc = DOMBuilder.parse(new StringReader(str));
			msgCtx.setData("xml", doc.getDocumentElement());
		    } catch (Exception e) {
			throw new G2Exception("Cannot parse XML document of SOAP message.", e);
		    }

		} else {
		    msgCtx.setData("bytes", str);
		}
	    }
	}

	return super.preprocessContext(msgCtx);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.plugin.callinterceptor.CallInterceptor#
     * execute
     * (at.ac.tuwien.vitalab.genesis2.plugin.callinterceptor.CallInterceptor
     * .InterceptorContext, groovy.lang.Closure)
     */
    @Override
    public void execute(InterceptorContext context, Closure closure) {

	if ((getType() == CORRTYPE_XML || getType() == CORRTYPE_BYTES) && context.getHookName().startsWith("out")) {

	    Closure cl = (Closure) context.getProperty("closure");

	    if (cl == null) {
		context.setProperty("closure", closure);
	    } else if (cl != null) {
		run(context, cl);
		// super.execute(context, cl);
	    }
	} else {
	    run(context, closure);
	    // super.execute(context, closure);
	}
    }

    /**
     * Start the visitor run.
     * 
     * @param context Context.
     * @param closure Visitor closure.
     */
    private void run(InterceptorContext context, Closure closure) {

	Closure code = getCode();

	if (code == null) {
	    return;
	}

	MsgPertContext mpContext = (MsgPertContext) context;

	switch (getType()) {
	case CORRTYPE_XML:

	    Node n = (Node) mpContext.getData("xml");
	    MessagePerturbationPlugin.visitXML(code, n);

	    break;

	case CORRTYPE_ARGS:

	    List<Object> args = (ArrayList<Object>) mpContext.getData("args");
	    MessagePerturbationPlugin.visitArgs(code, args);

	    break;

	case CORRTYPE_BYTES:

	    String str = (String) mpContext.getData("bytes");
	    str = (String) code.call(str);
	    mpContext.setData("bytes", str);

	    break;
	}
    }

    /**
     * XML to String.
     * 
     * @param node XML.
     * @return String.
     * @throws TransformerException On errors.
     */
    protected static String docToString(Node node) throws TransformerException {

	TransformerFactory transFactory = TransformerFactory.newInstance();
	Transformer transformer = transFactory.newTransformer();
	StringWriter buffer = new StringWriter();
	transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
	transformer.transform(new DOMSource(node), new StreamResult(buffer));
	return buffer.toString();
    }

    /**
     * Interceptor context.
     * 
     * @author Lukasz Juszczyk
     * 
     */
    protected class MsgPertContext extends InterceptorContext {

	private HashMap<String, Object> body = new HashMap<String, Object>();

	public MsgPertContext(InterceptorContext src) {
	    super(src);

	    if (src.getProperty("data") == null) {
		setProperty("data", body);
	    } else {
		setProperty("data", body = (HashMap<String, Object>) src.getProperty("data"));
	    }
	}

	public void setData(String type, Object body) {
	    this.body.put(type, body);
	}

	public Object getData(String type) {
	    return body.get(type);
	}
    }
}
