package at.ac.tuwien.vitalab.genesis2.plugin.netemu;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import lombok.Getter;
import lombok.Setter;

import org.apache.log4j.Logger;

import at.ac.tuwien.vitalab.genesis2.core.model.AbstractModelElement;

/**
 * Model type for specifying communication/network faults. Will run only on
 * Linux with sudo rights for tc and ifconfig.
 * 
 * @author Lukasz Juszczyk
 * 
 */
public class CommFaults extends AbstractModelElement {

    /**
     * Real interface name.
     */
    @Getter
    @Setter
    private String interfaceName = null;

    /**
     * Fake/virtual one for traffic control.
     */
    @Getter
    @Setter
    private String fakeInterfaceName = null;

    /**
     * Which port should be hampered.
     */
    @Getter
    @Setter
    private int port = -1;

    protected Logger logger = Logger.getLogger(getClass()); // NOPMD

    private boolean initialized = false;

    /**
     * Consctructor.
     */
    public CommFaults() {
	super();

	setProperty("delay", new HashMap<String, Object>());
	setProperty("loss", new HashMap<String, Object>());
	setProperty("corrupt", new HashMap<String, Object>());
	setProperty("duplicate", new HashMap<String, Object>());
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.core.model.ICleanable#cleanup()
     */
    @Override
    public void cleanup() {

	if (!initialized) {
	    return;
	}

	for (String s : getCloseTcCommands()) {
	    try {
		runCommand(s);
	    } catch (Exception e) {
		logger.error("Cannot cleanup traffic control configuration.", e);
	    }
	}
    }

    /**
     * Initialization.
     * 
     * @throws InterruptedException If the commands get interrupted.
     * @throws IOException If the commands' output cannot be read.
     */
    protected synchronized void init() throws InterruptedException, IOException {

	if (interfaceName == null || port < 0) {
	    throw new IllegalStateException("Interface name and port must be specified.");
	}

	if (initialized) {
	    return;
	}

	for (String s : getInitTcCommands()) {
	    runCommand(s);
	}

	updateTc();

	initialized = true;
    }

    /**
     * Run a single command in the shell.
     * 
     * @param cmd Command.
     * @return Return value.
     * @throws InterruptedException If the command gets interrupted.
     * @throws IOException If the command's output cannot be read.
     */
    private int runCommand(String cmd) throws InterruptedException, IOException {
	Runtime runtime = Runtime.getRuntime();

	Process proc = runtime.exec(cmd);

	BufferedReader br = new BufferedReader(new InputStreamReader(proc.getErrorStream()));

	String el = null;
	while ((el = br.readLine()) != null) {
	    logger.error("Error output of command: " + el);
	}

	return proc.waitFor();
    }

    /**
     * Get all commands necessary to set up traffic control.
     * 
     * @return List of command strings.
     */
    private List<String> getInitTcCommands() {

	List<String> commands = new ArrayList<String>();

	commands.add("sudo tc qdisc add dev " + interfaceName + " handle 1: root htb");
	commands.add("sudo tc class add dev " + interfaceName + " parent 1: classid 1:1 htb rate 1000Mbps");
	commands.add("sudo tc qdisc add dev " + interfaceName + " parent 1:1 handle 10: netem");
	commands.add("sudo tc filter add dev " + interfaceName + " protocol ip prio 1 u32 match ip sport " + port
		+ " 0xffff flowid 1:1");
	commands.add("sudo tc qdisc add dev " + interfaceName + " ingress");
	commands.add("sudo tc filter add dev " + interfaceName + " parent ffff: protocol ip u32 match ip dport " + port
		+ " 0xffff flowid 1:1 action mirred egress redirect dev " + fakeInterfaceName);
	commands.add("sudo tc qdisc add dev " + fakeInterfaceName + " root netem");

	return commands;
    }

    /**
     * Get all commands necessary to clean up traffic control.
     * 
     * @return List of command strings.
     */
    private List<String> getCloseTcCommands() {

	List<String> commands = new ArrayList<String>();

	commands.add("sudo tc qdisc del dev " + interfaceName + " handle 1: root htb");
	commands.add("sudo tc qdisc del dev " + interfaceName + " ingress");
	commands.add("sudo tc qdisc del dev " + fakeInterfaceName + " root netem");

	return commands;
    }

    /**
     * Iterates through all local get*Command() methods and calls them.
     * 
     * @throws InterruptedException If the commands get interrupted.
     * @throws IOException If the commands' output cannot be read.
     */
    public void updateTc() throws InterruptedException, IOException {

	String command = "";

	Method[] methods = this.getClass().getDeclaredMethods();

	for (Method m : methods) {

	    if (m.getParameterTypes().length == 0 && m.getName().startsWith("get") && m.getName().endsWith("Command")
		    && m.getReturnType().equals(String.class)) {

		try {
		    String c = (String) m.invoke(this);

		    if (c != null) {
			command += " " + c.toString() + " ";
		    }
		} catch (Exception e) {
		    logger.error("Cannot call command.", e);
		}
	    }
	}

	runCommand("sudo tc qdisc change dev " + interfaceName + " parent 1:1 handle 10: netem " + command);
	runCommand("sudo tc qdisc change dev " + fakeInterfaceName + " root netem " + command);
    }

    /**
     * Command string for packet corruption emulation.
     * 
     * @return Command string.
     */
    protected String getCorruptionCommand() {

	Map<String, Object> delayMap = (Map<String, Object>) getProperty("corrupt");

	Number value = (Number) delayMap.get("value");

	if (value == null) {
	    return null;
	}

	return "corrupt " + Math.round(value.floatValue() * 1000) / 10.0 + "% ";
    }

    /**
     * Command string for packet duplication emulation.
     * 
     * @return Command string.
     */
    protected String getDuplicationCommand() {

	Map<String, Object> delayMap = (Map<String, Object>) getProperty("duplicate");

	Number value = (Number) delayMap.get("value");

	if (value == null) {
	    return null;
	}

	return "duplicate " + Math.round(value.floatValue() * 1000) / 10.0 + "% ";
    }

    /**
     * Command string for packet loss emulation.
     * 
     * @return Command string.
     */
    protected String getLossCommand() {

	Map<String, Object> delayMap = (Map<String, Object>) getProperty("loss");

	Number value = (Number) delayMap.get("value");

	if (value == null) {
	    return null;
	}

	Number correlation = (Number) delayMap.get("correlation");

	String command = "loss " + Math.round(value.floatValue() * 1000) / 10.0 + "% ";

	if (correlation != null) {
	    command += Math.round(correlation.floatValue() * 1000) / 10.0 + "% ";
	}

	return command;
    }

    /**
     * Command string for packet delay emulation.
     * 
     * @return Command string.
     */
    protected String getDelayCommand() {

	Map<String, Object> delayMap = (Map<String, Object>) getProperty("delay");

	Number value = (Number) delayMap.get("value");

	if (value == null) {
	    return null;
	}

	Number variation = (Number) delayMap.get("variation");
	Number correlation = (Number) delayMap.get("correlation");
	String distribution = (String) delayMap.get("distribution");

	String command = "delay " + value + "ms ";

	if (variation != null) {
	    command += variation + "ms ";

	    if (correlation != null) {
		command += Math.round(correlation.floatValue() * 1000) / 10.0 + "% ";
	    }
	}

	if (distribution != null) {
	    command += "distribution " + distribution + " ";
	}

	return command;
    }
}
