package at.ac.tuwien.vitalab.genesis2.plugin.netemu;

import groovy.lang.Closure;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import at.ac.tuwien.vitalab.genesis2.Genesis2;
import at.ac.tuwien.vitalab.genesis2.config.Genesis2Configuration;
import at.ac.tuwien.vitalab.genesis2.core.model.AbstractModelElement;
import at.ac.tuwien.vitalab.genesis2.core.model.IModelObserver;
import at.ac.tuwien.vitalab.genesis2.core.model.builder.AbstractModelBuilder;
import at.ac.tuwien.vitalab.genesis2.core.model.types.Host;
import at.ac.tuwien.vitalab.genesis2.errors.G2Exception;
import at.ac.tuwien.vitalab.genesis2.plugin.AbstractStandaloneGenesis2Plugin;

/**
 * G2 plugin for network fault emulation.
 * 
 * @author Lukasz Juszczyk
 * 
 */
public class NetEmuPlugin extends AbstractStandaloneGenesis2Plugin<CommFaults> {

    /**
     * Listener for updates on communication fault models.
     */
    private CommFaultListener listener = null;

    /**
     * Currently deployed model.
     */
    private CommFaults deployed = null;

    /**
     * Interface name.
     */
    private String ifName = null;

    /**
     * Fake interface name.
     */
    private String fakeIfName = null;

    /**
     * Constructor.
     */
    public NetEmuPlugin() {
	super();

	if (Genesis2.isBackend()) {

	    try {
		ifName = Genesis2Configuration.get("net.faultemu.if").toString();
		fakeIfName = Genesis2Configuration.get("net.faultemu.fakeIf").toString();
	    } catch (Exception e) {
		throw new G2Exception("Invalid/missing configuration entry 'net.faultemu.if' or 'net.faultemu.fakeIf'.");
	    }
	}
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#deployElement(at
     * .ac.tuwien.vitalab.genesis2.core.model.AModelElement)
     */
    @Override
    public synchronized void deployElement(CommFaults cf) {

	if (deployed != null) {
	    throw new G2Exception("A network fault model has already been deployed.");
	}

	deployed = cf;

	deployed.setInterfaceName(ifName);
	deployed.setFakeInterfaceName(fakeIfName);
	deployed.setPort(Host.getFakeLocalhostForDeployment().getPort());

	try {
	    deployed.init();

	} catch (InterruptedException e) {
	    // ignore
	} catch (IOException e) {
	    throw new G2Exception("Cannot update network fault emulation.", e);
	}
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#doStart()
     */
    @Override
    protected void doStart() {

	if (Genesis2.isBackend()) {

	    listener = new CommFaultListener();
	    AbstractModelElement.getElementMap(CommFaults.class).observe(listener);
	}
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#doStop()
     */
    @Override
    protected void doStop() {

	if (listener != null) {
	    AbstractModelElement.getElementMap(CommFaults.class).stopObserving(listener);
	}

    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getAliases()
     */
    @Override
    public List<String> getAliases() {
	List<String> aliases = new ArrayList<String>();
	aliases.add("netemu");
	return aliases;
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getCommands()
     */
    @Override
    public Map<String, Closure> getCommands() {
	return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getPluginName()
     */
    @Override
    public String getPluginName() {
	return "NetworkFaultEmulatorPlugin";
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getTypesAndBuilders
     * ()
     */
    @Override
    public Map<Class<? extends AbstractModelElement>, Class<? extends AbstractModelBuilder<? extends AbstractModelElement>>> getTypesAndBuilders() {
	Map<Class<? extends AbstractModelElement>, Class<? extends AbstractModelBuilder<? extends AbstractModelElement>>> result = new HashMap<Class<? extends AbstractModelElement>, Class<? extends AbstractModelBuilder<? extends AbstractModelElement>>>();

	result.put(CommFaults.class, null);

	return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#undeployElement(
     * at.ac.tuwien.vitalab.genesis2.core.model.AModelElement)
     */
    @Override
    public void undeployElement(CommFaults e) {

	if (deployed == null || !deployed.equals(e)) {
	    return;
	}

	deployed.cleanup();
	deployed.deregister();
    }

    /**
     * Listener for changes in network faults model, which executes this changes
     * immediately.
     * 
     * @author Lukasz Juszczyk
     * 
     */
    protected class CommFaultListener implements IModelObserver<CommFaults> {

	/*
	 * (non-Javadoc)
	 * 
	 * @see at.ac.tuwien.vitalab.genesis2.core.model.IModelObserver#
	 * onChangedModelElement
	 * (at.ac.tuwien.vitalab.genesis2.core.model.AModelElement, int,
	 * java.lang.Object[])
	 */
	@Override
	public void onChangedModelElement(CommFaults t, int how, Object... info) {

	    if (how == IModelObserver.UPDATED) {
		try {
		    if (t == deployed) {
			t.updateTc();
		    }
		} catch (Exception e) {
		    logger.warn("Cannot update traffic control configuration.", e);
		}
	    }
	}
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getArtifact(java
     * .lang.Class, at.ac.tuwien.vitalab.genesis2.core.model.AModelElement,
     * java.lang.String[])
     */
    @Override
    public <S> S getArtifact(Class<S> type, CommFaults modelelement, String... additionalIDs) {
	return null;
    }

}
