package at.ac.tuwien.vitalab.genesis2.plugin.propagator;

import groovy.lang.Closure;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import at.ac.tuwien.vitalab.genesis2.Genesis2;
import at.ac.tuwien.vitalab.genesis2.base.ExecutionEnvironment;
import at.ac.tuwien.vitalab.genesis2.core.model.AbstractModelElement;
import at.ac.tuwien.vitalab.genesis2.core.model.ElementMap;
import at.ac.tuwien.vitalab.genesis2.core.model.IModelObserver;
import at.ac.tuwien.vitalab.genesis2.core.model.builder.AbstractModelBuilder;
import at.ac.tuwien.vitalab.genesis2.core.model.types.Host;
import at.ac.tuwien.vitalab.genesis2.core.model.ws.wrappers.KeyValueWrapper;
import at.ac.tuwien.vitalab.genesis2.errors.G2Exception;
import at.ac.tuwien.vitalab.genesis2.plugin.AbstractGenesis2Plugin;
import at.ac.tuwien.vitalab.genesis2.plugin.propagator.ws.IObjectPropagatorWebService;
import at.ac.tuwien.vitalab.genesis2.plugin.propagator.ws.ObjectPropagatorWebService;

/**
 * G2 plugin for setting properties globally, which means propagating them to
 * every host.
 * 
 * @author Lukasz Juszczyk
 * 
 */
@SuppressWarnings("PMD.UncommentedEmptyMethod")
public class ObjectPropagatorPlugin extends AbstractGenesis2Plugin<AbstractModelElement> {

    /**
     * Plugin name.
     */
    public static final String PLUGINNAME = "ObjectPropagator";

    private static final ExecutionEnvironment GLOBAL_ENV = ExecutionEnvironment.getGlobalInstance();

    protected final Map<String, Object> objectMap = new HashMap<String, Object>();

    protected final Map<Host, Map<String, Integer>> hostObjectMap = Collections
	    .synchronizedMap(new HashMap<Host, Map<String, Integer>>());

    private HostObserver hostObserver = null;

    private final ElementMap<Host> hostMap = AbstractModelElement.getElementMap(Host.class);

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getPluginName()
     */
    @Override
    public String getPluginName() {
	return PLUGINNAME;
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#doStart()
     */
    @Override
    protected void doStart() {
	if (Genesis2.isFrontend()) {
	    hostMap.observe(hostObserver = new HostObserver());
	}
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#doStop()
     */
    @Override
    protected void doStop() {
	if (Genesis2.isFrontend() && hostObserver != null) {
	    hostMap.stopObserving(hostObserver);
	}
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getAliases()
     */
    @Override
    public List<String> getAliases() {
	List<String> aliases = new ArrayList<String>();
	aliases.add("propagator");
	aliases.add("prop");
	return aliases;
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getCommands()
     */
    @Override
    public Map<String, Closure> getCommands() {
	return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getPluginWsClientClass
     * ()
     */
    @Override
    protected Class<?> getPluginWsClientClass() {
	return IObjectPropagatorWebService.class;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getPluginWsImplementor
     * ()
     */
    @Override
    public Object getPluginWsImplementor() {
	return new ObjectPropagatorWebService();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getTypesAndBuilders
     * ()
     */
    @Override
    public Map<Class<? extends AbstractModelElement>, Class<? extends AbstractModelBuilder<? extends AbstractModelElement>>> getTypesAndBuilders() {
	// nothing to do
	return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#deployElement(at
     * .ac.tuwien.vitalab.genesis2.core.model.AModelElement)
     */
    @Override
    public void deployElement(AbstractModelElement e) {
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#undeployElement(
     * at.ac.tuwien.vitalab.genesis2.core.model.AModelElement)
     */
    @Override
    public void undeployElement(AbstractModelElement e) {
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getArtifact(java
     * .lang.Class, at.ac.tuwien.vitalab.genesis2.core.model.AModelElement,
     * java.lang.String[])
     */
    @Override
    public <S> S getArtifact(Class<S> type, AbstractModelElement modelelement, String... additionalIDs) {
	return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see groovy.lang.GroovyObjectSupport#getProperty(java.lang.String)
     */
    @Override
    public Object getProperty(String property) {
	return objectMap.get(property);
    }

    /*
     * (non-Javadoc)
     * 
     * @see groovy.lang.GroovyObjectSupport#setProperty(java.lang.String,
     * java.lang.Object)
     */
    @Override
    public void setProperty(String property, Object newValue) {

	synchronized (objectMap) {
	    objectMap.put(property, newValue);
	    GLOBAL_ENV.setProperty(property, newValue);

	    try {
		updateHosts();
	    } catch (Exception e) {
		logger.error("Cannot propagate objects/variables: " + e.getMessage());
	    }
	}
    }

    /**
     * Propagate.
     */
    private void updateHosts() {

	synchronized (objectMap) {
	    Set<Thread> threads = new HashSet<Thread>();
	    final Set<Exception> errors = new HashSet<Exception>();

	    for (final Host h : hostObjectMap.keySet()) {

		Thread t = new Thread() {
		    public void run() {
			try {
			    updateSingleHost(h);
			} catch (Exception e) {
			    errors.add(e);
			}
		    }
		};
		threads.add(t);
		t.start();
	    }

	    for (Thread t : threads) {
		try {
		    t.join();
		} catch (InterruptedException e) {
		    // ignore
		}
	    }

	    if (!errors.isEmpty()) {
		throw new G2Exception("Could not propagate to all hosts: " + errors.iterator().next().getMessage());
	    }
	}
    }

    /**
     * Propagate to a single host.
     * 
     * @param host Host.
     */
    private void updateSingleHost(Host host) {

	Map<String, Integer> formerHashMap = hostObjectMap.get(host);
	Map<String, Integer> newHashMap = new HashMap<String, Integer>();

	if (formerHashMap == null) {
	    formerHashMap = new HashMap<String, Integer>();
	}

	List<KeyValueWrapper> changed = new ArrayList<KeyValueWrapper>();

	for (String name : objectMap.keySet()) {
	    Object current = objectMap.get(name);
	    Integer formerHash = formerHashMap.get(name);

	    if (current == null) {
		if (formerHash != null) {
		    changed.add(new KeyValueWrapper(name, null));
		}

	    } else {
		Integer currentHash = current.hashCode();

		if (!currentHash.equals(formerHash)) {
		    changed.add(new KeyValueWrapper(name, current));
		}

		newHashMap.put(name, currentHash);
	    }
	}

	if (!changed.isEmpty()) {

	    IObjectPropagatorWebService wsc = (IObjectPropagatorWebService) getPluginWsClient(host);
	    wsc.set(changed);

	    hostObjectMap.put(host, newHashMap);
	}
    }

    private class HostObserver implements IModelObserver<Host> {

	/*
	 * (non-Javadoc)
	 * 
	 * @see at.ac.tuwien.vitalab.genesis2.core.model.IModelObserver#
	 * onChangedModelElement
	 * (at.ac.tuwien.vitalab.genesis2.core.model.AModelElement, int,
	 * java.lang.Object[])
	 */
	@Override
	public void onChangedModelElement(Host t, int how, Object... info) {

	    synchronized (objectMap) {

		switch (how) {
		case IModelObserver.CREATED:

		    if (!hostObjectMap.containsKey(t) && !t.isLocalhost()) {
			hostObjectMap.put(t, null);
		    }

		    break;
		case IModelObserver.REMOVED:

		    hostObjectMap.remove(t);

		    break;

		default:
		}
	    }
	}
    }
}
