package at.ac.tuwien.vitalab.genesis2.plugin.qosemu;

import groovy.lang.Closure;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Timer;
import java.util.TimerTask;
import java.util.UUID;

import org.apache.cxf.binding.soap.SoapMessage;
import org.apache.cxf.binding.soap.interceptor.AbstractSoapInterceptor;
import org.apache.cxf.endpoint.Endpoint;
import org.apache.cxf.endpoint.Server;
import org.apache.cxf.interceptor.Fault;
import org.apache.cxf.interceptor.Interceptor;
import org.apache.cxf.phase.Phase;

import at.ac.tuwien.vitalab.genesis2.Genesis2;
import at.ac.tuwien.vitalab.genesis2.core.model.AbstractModelElement;
import at.ac.tuwien.vitalab.genesis2.core.model.ElementMap;
import at.ac.tuwien.vitalab.genesis2.core.model.IModelObserver;
import at.ac.tuwien.vitalab.genesis2.core.model.builder.AbstractModelBuilder;
import at.ac.tuwien.vitalab.genesis2.errors.G2Exception;
import at.ac.tuwien.vitalab.genesis2.plugin.AbstractGenesis2Plugin;
import at.ac.tuwien.vitalab.genesis2.plugin.AbstractStandaloneGenesis2Plugin;
import at.ac.tuwien.vitalab.genesis2.plugin.wsgen.WebService;
import at.ac.tuwien.vitalab.genesis2.plugin.wsgen.WebServiceGeneratorPlugin;

/**
 * QOS emulator plugin for G2. It's a complex one...
 * 
 * @author Lukasz Juszczyk
 * 
 */
public class QOSEmulatorPlugin extends AbstractStandaloneGenesis2Plugin<QOSProps> {

    /**
     * For altering availability of Web services in intervals.
     */
    protected final Map<WebService, AvailabilityControlTask> availabilityTasks = new HashMap<WebService, AvailabilityControlTask>();

    /**
     * For altering availability of Web services in intervals.
     */
    protected Timer availabilityTimer;

    /**
     * Attaches the QOS interceptor to all found services.
     */
    private IModelObserver<WebService> serviceObserver = null;

    /**
     * Service map.
     */
    private final ElementMap<WebService> serviceMap = AbstractModelElement.getElementMap(WebService.class);

    /**
     * Web service generator plugin.
     */
    private final WebServiceGeneratorPlugin wsGen = AbstractGenesis2Plugin.getPlugin(WebServiceGeneratorPlugin.class);

    /**
     * Web service observer.
     * 
     * @author Lukasz Juszczyk
     * 
     */
    protected class ServiceObserver implements IModelObserver<WebService> {

	/*
	 * (non-Javadoc)
	 * 
	 * @see at.ac.tuwien.vitalab.genesis2.core.model.IModelObserver#
	 * changedModelElement
	 * (at.ac.tuwien.vitalab.genesis2.core.model.AModelElement, int,
	 * java.lang.Object[])
	 */
	@Override
	public void onChangedModelElement(WebService service, int how, Object... info) {

	    try {

		switch (how) { // NOPMD

		case IModelObserver.DEPLOYED:

		    // attach a QOS interceptor to web service

		    List<Interceptor> interceptors = wsGen.getArtifact(Endpoint.class, service).getOutInterceptors();

		    boolean found = false;

		    for (Interceptor<?> i : interceptors) {
			if (i instanceof QOSInterceptor) {
			    found = true;
			    break;
			}
		    }

		    if (!found) {
			interceptors.add(new QOSInterceptor(service));
		    }

		    // start availability timer

		    if (availabilityTimer != null) {

			AvailabilityControlTask act = availabilityTasks.get(service);

			if (act == null) {
			    act = new AvailabilityControlTask(service);
			    availabilityTimer.scheduleAtFixedRate(act, 0, 5000);
			}
		    }

		    break;

		case IModelObserver.UNDEPLOYED:

		    AvailabilityControlTask act = availabilityTasks.get(service);

		    if (act != null) {
			act.cancel();
		    }

		    break;
		}

	    } catch (Exception e) {
		// ignore
	    }

	}
    }

    /**
     * Interceptor for delaying Web service operations.
     * 
     * @author Lukasz Juszczyk
     * 
     */
    protected class QOSInterceptor extends AbstractSoapInterceptor {

	/**
	 * Intercepted Web service.
	 */
	private WebService ws;

	/**
	 * Map of remaining processing times per invocation ID.
	 */
	private Map<String, Integer> remainingTime = new HashMap<String, Integer>();

	/**
	 * Timestamp of last check.
	 */
	private Long lastCheckTS = -1L;

	private final Random rand = new java.util.Random();

	/**
	 * Constructor.
	 * 
	 * @param ws Intercepted Web service.
	 */
	public QOSInterceptor(WebService ws) {
	    super(Phase.SETUP);
	    this.ws = ws;
	}

	private synchronized String start() {

	    // each invocation get its own ID
	    String id = UUID.randomUUID().toString();

	    QOSProps qos = (QOSProps) ws.getProperty("qos");

	    // get current response time from QOS model
	    if (qos != null) {
		if (remainingTime.isEmpty()) {
		    lastCheckTS = System.currentTimeMillis();
		}

		remainingTime.put(id, qos.getResponseTime());
	    }

	    return id;
	}

	/**
	 * Returns the time by which the call should be delayed.
	 * 
	 * @param id Invocation ID.
	 * @return Time in msecs.
	 */
	private synchronized int getTimeToWait(String id) {

	    if (!remainingTime.containsKey(id)) {
		return 0;
	    }

	    long thisCheckTS = System.currentTimeMillis();
	    long tsDiff = thisCheckTS - lastCheckTS;
	    lastCheckTS = thisCheckTS;

	    QOSProps qos = (QOSProps) ws.getProperty("qos");

	    // complicated calculations:
	    // if there are more invocations than throughput allows
	    // the remaining processing time is growing according to scalability
	    // penalty
	    int concurrentCallNumber = remainingTime.size();
	    int throughput = qos != null ? qos.getThroughput() : 0;
	    double scalability = qos != null ? qos.getScalability() : 0.0;

	    if (concurrentCallNumber > throughput && throughput > 0) {
		double ratio = (double) throughput / (double) concurrentCallNumber;
		ratio = Math.pow(ratio, scalability);
		tsDiff *= ratio;
	    }

	    for (String tId : remainingTime.keySet()) {
		Integer time = remainingTime.get(tId);
		time -= (int) tsDiff;
		remainingTime.put(tId, time);
	    }

	    Integer time = remainingTime.get(id);

	    if (time <= 0) {
		remainingTime.remove(id);
	    }

	    return time;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.apache.cxf.interceptor.Interceptor#handleMessage(org.apache.cxf
	 * .message.Message)
	 */
	@Override
	public void handleMessage(SoapMessage message) throws Fault {

	    QOSProps qos = (QOSProps) ws.getProperty("qos");

	    if (qos == null) {
		return;
	    }

	    String id = start();

	    int delay = 0;

	    // delay until processed
	    while ((delay = getTimeToWait(id)) > 0) {
		try {
		    Thread.sleep(delay);
		} catch (InterruptedException e) {
		}
	    }

	    if (qos.getErrorRate() > rand.nextDouble()) {
		throw new Fault(new G2Exception("Generated Error."));
	    }
	}

    }

    /**
     * Task for altering a service's availability.
     * 
     * @author Lukasz Juszczyk
     * 
     */
    protected class AvailabilityControlTask extends TimerTask {

	/**
	 * Controlled Web service.
	 */
	private WebService ws;

	/**
	 * Constructor.
	 * 
	 * @param ws Web service.
	 */
	public AvailabilityControlTask(WebService ws) {
	    this.ws = ws;
	}

	private final Random rand = new java.util.Random();

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.util.TimerTask#run()
	 */
	@Override
	public void run() {

	    QOSProps qos = (QOSProps) ws.getProperty("qos");

	    Server s = wsGen.getArtifact(Server.class, ws);

	    if (s == null || qos == null) {
		return;
	    }

	    // randomize availability according to probability
	    if (qos.getAvailability() >= rand.nextDouble()) {
		s.start();
	    } else {
		s.stop();
	    }
	}
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#deployElement(at
     * .ac.tuwien.vitalab.genesis2.core.model.AModelElement)
     */
    @Override
    public void deployElement(QOSProps e) {
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#doStart()
     */
    @Override
    protected void doStart() {

	if (!Genesis2.isBackend()) {
	    return;
	}

	availabilityTimer = new Timer();

	serviceObserver = new ServiceObserver();
	serviceMap.observe(serviceObserver);
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#doStop()
     */
    @Override
    protected void doStop() {

	if (!Genesis2.isBackend()) {
	    return;
	}

	serviceMap.stopObserving(serviceObserver);

	if (availabilityTimer != null) {
	    availabilityTimer.cancel();
	    availabilityTimer = null;
	}
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getAliases()
     */
    @Override
    public List<String> getAliases() {
	return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getCommands()
     */
    @Override
    public Map<String, Closure> getCommands() {
	return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getPluginName()
     */
    @Override
    public String getPluginName() {
	return "QOSEmulator";
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getTypesAndBuilders
     * ()
     */
    @Override
    public Map<Class<? extends AbstractModelElement>, Class<? extends AbstractModelBuilder<? extends AbstractModelElement>>> getTypesAndBuilders() {
	HashMap<Class<? extends AbstractModelElement>, Class<? extends AbstractModelBuilder<? extends AbstractModelElement>>> result = new HashMap<Class<? extends AbstractModelElement>, Class<? extends AbstractModelBuilder<? extends AbstractModelElement>>>();

	result.put(QOSProps.class, null);

	return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#undeployElement(
     * at.ac.tuwien.vitalab.genesis2.core.model.AModelElement)
     */
    @Override
    public void undeployElement(QOSProps e) {
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.plugin.AGenesis2Plugin#getArtifact(java
     * .lang.Class, at.ac.tuwien.vitalab.genesis2.core.model.AModelElement,
     * java.lang.String[])
     */
    @Override
    public <S> S getArtifact(Class<S> type, QOSProps modelelement, String... additionalIDs) {
	return null;
    }

}
