package at.ac.tuwien.vitalab.genesis2.plugin.wsgen;

import groovy.lang.Closure;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.StringTokenizer;

import javax.jws.soap.SOAPBinding;

import at.ac.tuwien.vitalab.genesis2.core.model.AbstractModelElement;
import at.ac.tuwien.vitalab.genesis2.core.model.types.Host;
import at.ac.tuwien.vitalab.genesis2.plugin.AbstractGenesis2Plugin;

/**
 * Model type for SOAP Web services.
 * 
 * @author Lukasz Juszczyk
 * 
 */
public class WebService extends AbstractModelElement {

    /**
     * Constructor.
     */
    public WebService() {
	super();

	super.setProperty("operations", new HashSet<WsOperation>());
    }

    /**
     * Constructor.
     * 
     * @param name Web service name.
     */
    public WebService(String name) {
	this();

	if (name == null) {
	    throw new IllegalArgumentException("Name cannot be null.");
	}

	setName(name);
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.core.model.ICleanable#cleanup()
     */
    @Override
    public void cleanup() {
    }

    /**
     * Sets the binding (rpc or doc, literal or encoded, bare or wrapped) via a
     * string that is parsed.
     * 
     * @param binding Binding declaration.
     */
    public void setBinding(String binding) {

	StringTokenizer tokens = new StringTokenizer(binding, " \t,.#-+_/");

	while (tokens.hasMoreTokens()) {
	    String t = tokens.nextToken();

	    String foundStr = null;

	    // first check style
	    foundStr = findIn(t, SOAPBinding.Style.values());

	    if (foundStr != null) {
		setBindingStyle(foundStr);
		continue;
	    }

	    // then param style
	    foundStr = findIn(t, SOAPBinding.ParameterStyle.values());

	    if (foundStr != null) {
		setBindingParamStyle(foundStr);
		continue;
	    }

	    // and now the use
	    foundStr = findIn(t, SOAPBinding.Use.values());

	    if (foundStr != null) {
		setBindingUse(foundStr);
		continue;
	    }

	    throw new IllegalArgumentException("Unknown binding String '" + t + "'.");
	}
    }

    /*
     * Find a substring within a list of strings.
     */
    private String findIn(String str, Object[] list) {

	for (Object s : list) {
	    if (s.toString().toLowerCase().contains(str.toLowerCase())) {
		return s.toString();
	    }
	}

	return null;
    }

    /**
     * Gets the binding (rpc or doc).
     * 
     * @return Binding.
     */
    public String getBindingStyle() {

	String result = (String) getProperty("bindingStyle");

	if (result != null) {
	    return result;
	}

	return SOAPBinding.Style.DOCUMENT.toString();
    }

    /**
     * Sets the binding (rpc or doc).
     * 
     * @param bindingStyle Binding.
     */
    public void setBindingStyle(String bindingStyle) {
	super.setProperty("bindingStyle", SOAPBinding.Style.valueOf(bindingStyle).toString());
    }

    /**
     * Gets the parameter style (bare or wrapped).
     * 
     * @return Parameter style.
     */
    public String getBindingParamStyle() {
	String result = (String) getProperty("bindingParamStyle");

	if (result != null) {
	    return result;
	}

	return SOAPBinding.ParameterStyle.WRAPPED.toString();
    }

    /**
     * Sets the parameter style (bare or wrapped).
     * 
     * @param bindingParamStyle Parameter style.
     */
    public void setBindingParamStyle(String bindingParamStyle) {
	super.setProperty("bindingParamStyle", SOAPBinding.ParameterStyle.valueOf(bindingParamStyle).toString());
    }

    /**
     * Gets the binding use (literal or encoded).
     * 
     * @return Binding use.
     */
    public String getBindingUse() {
	String result = (String) getProperty("bindingUse");

	if (result != null) {
	    return result;
	}

	return SOAPBinding.Use.LITERAL.toString();
    }

    /**
     * Sets the binding use (literal or encoded).
     * 
     * @param bindingUse Binding use.
     */
    public void setBindingUse(String bindingUse) {
	super.setProperty("bindingUse", SOAPBinding.Use.valueOf(bindingUse).toString());
    }

    /**
     * Gets the closure which will be executed on deployment.
     * 
     * @return Closure.
     */
    public Closure getOnDeploy() {
	return (Closure) getProperty("onDeploy");
    }

    /**
     * Sets the closure which will be executed on deployment.
     * 
     * @param onDeploy Closure.
     */
    public void setOnDeploy(Closure onDeploy) {
	super.setProperty("onDeploy", onDeploy);
    }

    /**
     * Gets the closure which will be executed on undeployment.
     * 
     * @return Closure.
     */
    public Closure getOnUndeploy() {
	return (Closure) getProperty("onUndeploy");
    }

    /**
     * Sets the closure which will be executed on undeployment.
     * 
     * @param onDeploy Closure.
     */
    public void setOnUndeploy(Closure onUndeploy) {
	super.setProperty("onUndeploy", onUndeploy);
    }

    /**
     * Checks whether WS-Addressing is enabled.
     * 
     * @return <code>true</code> if enabled.
     */
    public Boolean isAddressingEnabled() {
	try {
	    return (Boolean) getProperty("addressingEnabled").equals(Boolean.TRUE);
	} catch (Exception e) {
	    return false;
	}
    }

    /**
     * Sets whether WS-Addressing is enabled.
     * 
     * @param addressingEnabled <code>true</code> if enabled.
     */
    public void setAddressingEnabled(Boolean addressingEnabled) {
	super.setProperty("addressingEnabled", addressingEnabled);
    }

    /**
     * Gets the service's name.
     * 
     * @return Name.
     */
    public String getName() {
	return (String) getProperty("name");
    }

    /**
     * Gets the service's namespace.
     * 
     * @return Namespace.
     */
    public String getNamespace() {

	String namespace = (String) getProperty("namespace");

	if (namespace != null) {
	    return namespace;
	}

	return "http://www.vitalab.tuwien.ac.at/Genesis2/generated/" + getName();
    }

    /**
     * Gets the service's port name.
     * 
     * @return Port name.
     */
    public String getPortName() {

	String portName = (String) getProperty("portName");

	if (portName != null) {
	    return portName;
	}

	return getName() + "Port";
    }

    /**
     * Sets the service's name.
     * 
     * @param name Name.
     */
    public final void setName(String name) {
	super.setProperty("name", name);
    }

    /**
     * Sets the service's namespace.
     * 
     * @param namespace Namespace.
     */
    public void setNamespace(String namespace) {
	super.setProperty("namespace", namespace);
    }

    /**
     * Sets the service's port name.
     * 
     * @param portName Port name.
     */
    public void setPortName(String portName) {
	super.setProperty("portName", portName);
    }

    /**
     * Gets all operations.
     * 
     * @return All operations.
     */
    public Collection<WsOperation> getOperations() {
	Collection<WsOperation> result = (Collection<WsOperation>) getProperty("operations");
	if (result != null) {
	    return new HashSet<WsOperation>(result);
	}
	return new ArrayList<WsOperation>();
    }

    /**
     * Gets a particular operation.
     * 
     * @param name Operation name.
     * @return Operation or <code>null</code> if none exists.
     */
    public WsOperation getOperation(String name) {
	for (WsOperation op : getOperations()) {
	    if (op.getName().equals(name)) {
		return op;
	    }
	}
	return null;
    }

    /**
     * Sets all operations.
     * 
     * @param operations Operations.
     */
    public void setOperations(Collection<WsOperation> operations) {
	super.setProperty("operations", operations);
    }

    /**
     * Adds an operation.
     * 
     * @param operation Operation.
     */
    public void addOperation(WsOperation operation) {
	Collection<WsOperation> ops = getOperations();
	deleteOperation(operation.getName());
	ops.add(operation);
	setOperations(ops);
    }

    /**
     * Deletes an operation.
     * 
     * @param operation Operation.
     */
    public void deleteOperation(WsOperation operation) {
	Collection<WsOperation> ops = getOperations();
	ops.remove(operation);
	setOperations(ops);
    }

    /**
     * Deletes an operation.
     * 
     * @param opName Operation name.
     */
    public void deleteOperation(String opName) {
	WsOperation op = getOperation(opName);
	if (op != null) {
	    deleteOperation(op);
	}
    }

    /**
     * Returns the URL string to the local element instance web service.
     * 
     * @return URL.
     */
    public String getLocalEndpointURL() {
	Host h = Host.getFakeLocalhostForDeployment();
	return getEndpointURL(h);
    }

    /**
     * Returns the URL string to the element instance web service.
     * 
     * @param host Host.
     * @return URL.
     */
    public String getEndpointURL(Host host) {
	return AbstractGenesis2Plugin.getModelWsURL(host.getHost(), host.getPort(), getClass()) + "/instance/" + getId();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.core.model.AModelElement#setProperty(java
     * .lang.String, java.lang.Object)
     */
    @Override
    public synchronized void setProperty(String property, Object newValue) {

	if ("operations".equals(property)) {
	    setOperations((Collection<WsOperation>) newValue);
	} else if ("binding".equals(property)) {
	    setBinding(newValue.toString());
	} else if ("bindingStyle".equals(property)) {
	    setBindingStyle(newValue.toString());
	} else if ("bindingParamStyle".equals(property)) {
	    setBindingParamStyle(newValue.toString());
	} else if ("bindingUse".equals(property)) {
	    setBindingUse(newValue.toString());
	} else if ("onDeploy".equals(property)) {
	    setOnDeploy((Closure) newValue);
	} else if ("onUndeploy".equals(property)) {
	    setOnUndeploy((Closure) newValue);
	} else {
	    super.setProperty(property, newValue);
	}
    }

}
