package at.ac.tuwien.vitalab.genesis2.plugin.wsgen;

import groovy.lang.Closure;

import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map.Entry;

import at.ac.tuwien.vitalab.genesis2.core.model.AbstractModelElement;
import at.ac.tuwien.vitalab.genesis2.core.model.ElementMap;
import at.ac.tuwien.vitalab.genesis2.core.model.types.DataType;

/**
 * Model type for Web service operations.
 * 
 * @author Lukasz Juszczyk
 * 
 */
public class WsOperation extends AbstractModelElement {

    private static final ElementMap<DataType> TYPEMAP = getElementMap(DataType.class);

    /**
     * Constructor.
     */
    public WsOperation() {
	super();
    }

    /**
     * Constructor.
     * 
     * @param name Operation name.
     */
    public WsOperation(String name) {
	super();

	if (name == null) {
	    throw new IllegalArgumentException("Name cannot be null.");
	}

	setName(name);
    }

    /**
     * Returns the callback operation that receives the result of the given
     * operation.
     * 
     * @param op Source operation.
     * @param behavior Callback behavior.
     * @return Callback operation.
     */
    public static WsOperation getCallbackOf(WsOperation op, Closure behavior) {

	WsOperation callback = new WsOperation(op.getName() + "Response");

	LinkedHashMap<String, DataType> opParams = op.getParamTypes();
	LinkedHashMap<String, DataType> callbackParams = new LinkedHashMap<String, DataType>();

	for (Entry<String, DataType> e : opParams.entrySet()) {

	    String name = e.getKey();
	    DataType par = e.getValue();

	    if (par.isOutHeader() || par.isInOutHeader()) {
		DataType cbPar = DataType.asHeader(par, "IN");

		callbackParams.put(name, cbPar);
	    }
	}

	if (op.getReturnType() != null) {
	    callbackParams.put("response", op.getReturnType());
	}

	callback.setParamTypes(callbackParams);
	callback.setBehavior(behavior);
	callback.setAction(op.getAction());

	return callback;
    }

    /**
     * Returns the callback operation that receives the faults of the given
     * operation.
     * 
     * @param op Source operation.
     * @param behavior Callback behavior.
     * @return Callback operation.
     */
    public static WsOperation faultCallbackOf(WsOperation op, Closure behavior) {
	WsOperation callback = new WsOperation(op.getName() + "Response");

	callback.setBehavior(behavior);
	callback.setAction(op.getAction());

	return callback;
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.core.model.ICleanable#cleanup()
     */
    @Override
    public void cleanup() {
    }

    /**
     * Whether this operation is a dummy one, which means that is returns
     * randomized dummy data.
     * 
     * @return <code>true</code> if dummy.
     */
    public Boolean isDummy() {
	try {
	    return (Boolean) super.getProperty("isDummy") == true;
	} catch (NullPointerException npe) {
	    return false;
	}
    }

    /**
     * Sets whether this operation is a dummy.
     * 
     * @param dummy <code>true</code> if dummy.
     */
    public void setDummy(Boolean dummy) {
	super.setProperty("isDummy", dummy);
    }

    /**
     * Gets the default SOAP action string for an operation.
     * 
     * @param operationName operation name.
     * @return Action.
     */
    public String getDefaultAction(String operationName) {
	return "http://www.vitalab.tuwien.ac.at/Genesis2/generated/" + operationName;
    }

    /**
     * Gets the SOAP action for this operation.
     * 
     * @return Action.
     */
    public String getAction() {

	String actionStr = (String) getProperty("action");

	if (actionStr != null) {
	    return actionStr;
	}

	return getDefaultAction(getName());
    }

    /**
     * Sets the SOAP action for this operation.
     * 
     * @param action Action.
     */
    public void setAction(String action) {
	super.setProperty("action", action);
    }

    /**
     * Getter for the operation name.
     * 
     * @return Name.
     */
    public String getName() {
	return (String) getProperty("name");
    }

    /**
     * Setter for the operation name.
     * 
     * @param name Name.
     */
    public final void setName(String name) {
	super.setProperty("name", name);
    }

    /**
     * Getter for the operation behavior.
     * 
     * @return Behavior closure.
     */
    public Closure getBehavior() {
	return (Closure) getProperty("behavior");
    }

    /**
     * Setter for the operation behavior.
     * 
     * @param behavior Behavior closure.
     */
    public void setBehavior(Closure behavior) {
	super.setProperty("behavior", behavior);
    }

    /**
     * Gets the return type.
     * 
     * @return Return type.
     */
    public DataType getReturnType() {

	DataType returnType = (DataType) getProperty("returnType");

	if (returnType == null) {
	    return TYPEMAP.create(void.class);
	}

	return returnType;
    }

    /**
     * Sets the return type.
     * 
     * @param returnType Return type.
     */
    public void setReturnType(DataType returnType) {
	super.setProperty("returnType", returnType);
    }

    /**
     * Sets the return type.
     * 
     * @param returnType Return type class.
     */
    public void setReturnType(Class<?> returnType) {
	super.setProperty("returnType", TYPEMAP.create(returnType));
    }

    /**
     * Gets all parameter types.
     * 
     * @return Parameter type map.
     */
    public LinkedHashMap<String, DataType> getParamTypes() {
	LinkedHashMap<String, DataType> types = (LinkedHashMap<String, DataType>) getProperty("paramTypes");

	if (types != null) {
	    return types;
	}

	return new LinkedHashMap<String, DataType>();
    }

    /**
     * Gets a specific parameter type.
     * 
     * @param paramName Parameter name.
     * @return Parameter type.
     */
    public DataType getParamType(String paramName) {
	return getParamTypes().get(paramName);
    }

    /**
     * Sets parameter types.
     * 
     * @param paramTypes Parameter type map.
     */
    public void setParamTypes(LinkedHashMap<String, ?> paramTypes) {

	LinkedHashMap<String, DataType> realMap = new LinkedHashMap<String, DataType>();

	Iterator<String> paramIter = paramTypes.keySet().iterator();

	while (paramIter.hasNext()) {
	    String name = paramIter.next();
	    Object typeVal = paramTypes.get(name);

	    if (typeVal instanceof Class) {
		realMap.put(name, TYPEMAP.create(typeVal));
	    } else {
		realMap.put(name, (DataType) typeVal);
	    }
	}

	super.setProperty("paramTypes", realMap);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * at.ac.tuwien.vitalab.genesis2.core.model.AModelElement#setProperty(java
     * .lang.String, java.lang.Object)
     */
    @Override
    public synchronized void setProperty(String property, Object newValue) {

	if ("returnType".equals(property)) {

	    if (newValue instanceof Class) {
		this.setReturnType((Class<?>) newValue);
	    } else {
		this.setReturnType((DataType) newValue);
	    }
	} else if ("paramTypes".equals(property)) {

	    setParamTypes((LinkedHashMap<String, ?>) newValue);

	} else {
	    super.setProperty(property, newValue);
	}
    }

}
