package at.ac.tuwien.vitalab.genesis2.plugin.wsreference;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;

import javax.xml.bind.annotation.XmlType;

import org.apache.cxf.binding.soap.model.SoapOperationInfo;
import org.apache.cxf.service.model.BindingInfo;
import org.apache.cxf.service.model.BindingOperationInfo;
import org.apache.cxf.service.model.MessagePartInfo;
import org.apache.cxf.service.model.ServiceInfo;
import org.codehaus.groovy.runtime.InvokerHelper;

import at.ac.tuwien.vitalab.genesis2.core.model.types.ADataTypeInstance;

/**
 * Web service request wrapper.
 * 
 * @author Lukasz Juszczyk
 * 
 */
public class WebServiceRequest {

    // arguments will be passed either as map
    private Map<String, Object> argumentMap = null;
    // or as list
    private List<Object> argumentList = null;

    private Iterator<Object> argumentIter = null;

    private boolean isMap = false;

    // final arguments passed to the jax-ws proxy
    private List<Object> callArguments = null;

    // model infos
    private Collection<MessagePartInfo> messageParts = null;
    private ServiceInfo serviceInfo = null;

    /**
     * Constructor.
     * 
     * @param serviceInfo Service.
     * @param messageParts Message parts.
     * @param args Argument map.
     */
    public WebServiceRequest(ServiceInfo serviceInfo, Collection<MessagePartInfo> messageParts, Map<String, ?> args) {

	super();

	this.serviceInfo = serviceInfo;
	this.messageParts = messageParts;

	this.argumentMap = new HashMap<String, Object>();
	if (args != null) {
	    this.argumentMap.putAll(args);
	}

	isMap = true;
    }

    /**
     * Constructor.
     * 
     * @param serviceInfo Service.
     * @param messageParts Message parts.
     * @param args Argument list.
     */
    public WebServiceRequest(ServiceInfo serviceInfo, Collection<MessagePartInfo> messageParts, List<?> args) {

	super();

	this.serviceInfo = serviceInfo;
	this.messageParts = messageParts;

	this.argumentList = new ArrayList<Object>();
	if (args != null) {
	    this.argumentList.addAll(args);
	}

	argumentIter = argumentList.iterator();
    }

    // get next param in list
    private Object nextParam() {
	try {
	    return argumentIter.next();
	} catch (NoSuchElementException nsee) {
	    throw new IllegalArgumentException("Not enough arguments.");
	}
    }

    // get param of name
    private Object nextParam(String name) {
	if (isMap) {
	    if (!argumentMap.containsKey(name)) {
		throw new IllegalArgumentException("Argument '" + name + "' has not been passed.");
	    }
	    return argumentMap.get(name);
	}
	return nextParam();
    }

    /**
     * Returns the list of argument and headers which shall be passed to the CXF
     * proxy.
     * 
     * @return Argument and header list.
     * 
     * @throws IllegalAccessException If the wrapper object cannot be created.
     * @throws InstantiationException If the wrapper object cannot be created.
     */
    public synchronized List<Object> getArgumentAndHeaderList() throws InstantiationException, IllegalAccessException {

	// already transformed? return the result
	if (callArguments != null) {
	    return callArguments;
	}

	List<Object> result = new ArrayList<Object>();

	if (messageParts.isEmpty()) {
	    return result;
	}

	boolean multipleParts = messageParts.size() > 1;

	int posInArgs = 0;

	boolean isRpcStyle = false;

	// find out whether this is RPC or DOC style
	for (BindingInfo bi : serviceInfo.getBindings()) {
	    for (BindingOperationInfo boi : bi.getOperations()) {
		isRpcStyle = boi.getExtensor(SoapOperationInfo.class).getStyle().toLowerCase().equals("rpc");
	    }
	}

	int cycle = -1;

	if (isRpcStyle) {

	    // in rpc, each argument has its messagepart

	    for (MessagePartInfo mpi : messageParts) {

		cycle++;

		Class<?> wrapperClass = mpi.getTypeClass();
		XmlType xmlTypeAno = wrapperClass.getAnnotation(XmlType.class);

		Object par = nextParam(mpi.getConcreteName().getLocalPart());

		if (xmlTypeAno != null) {
		    Object wrapper = wrapperClass.newInstance();
		    ADataTypeInstance.assignFromTo(par, wrapper);
		    result.add(wrapper);
		} else {
		    result.add(par);
		}
	    }
	} else {

	    // in doc, the arguments are merged into message parts

	    for (MessagePartInfo mpi : messageParts) {

		cycle++;

		Class<?> wrapperClass = mpi.getTypeClass();
		XmlType xmlTypeAno = wrapperClass.getAnnotation(XmlType.class);

		if (xmlTypeAno != null) {

		    Object wrapper = wrapperClass.newInstance();

		    if (cycle == 0) {
			String[] propOrder = xmlTypeAno.propOrder();

			for (String prop : propOrder) {

			    if (prop == null || prop.length() == 0) {
				result.add(null);
				continue;
			    }

			    Object np = nextParam(prop);
			    String name = "get" + prop.toLowerCase();

			    for (Method m : wrapper.getClass().getMethods()) {
				if (m.getName().toLowerCase().equals(name)) {

				    Class<?> retType = m.getReturnType();

				    if (retType.getAnnotation(XmlType.class) != null) {
					Object no = m.getReturnType().newInstance();
					ADataTypeInstance.assignFromTo(np, no);
					np = no;
				    }

				    if (multipleParts || posInArgs > 0) {
					InvokerHelper.setProperty(wrapper, prop, np);
				    } else {
					result.add(np);
				    }
				}
			    }

			    if (multipleParts || posInArgs > 0) {
				result.add(wrapper);
			    }
			}
		    } else {
			Object np = nextParam(mpi.getConcreteName().getLocalPart());
			ADataTypeInstance.assignFromTo(np, wrapper);
			result.add(wrapper);
		    }
		} else {
		    result.add(nextParam(mpi.getConcreteName().getLocalPart()));
		}
		posInArgs++;
	    }
	}

	return callArguments = result;
    }
}
