package at.ac.tuwien.vitalab.genesis2.plugin.wsreference;

import groovy.lang.GroovyObjectSupport;
import groovy.lang.GroovyRuntimeException;

import java.io.File;
import java.io.FileWriter;
import java.io.StringWriter;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.wsdl.Definition;
import javax.xml.bind.annotation.XmlType;
import javax.xml.namespace.QName;

import org.apache.cxf.endpoint.Client;
import org.apache.cxf.endpoint.ClientCallback;
import org.apache.cxf.jaxws.endpoint.dynamic.JaxWsDynamicClientFactory;
import org.apache.cxf.service.model.MessagePartInfo;
import org.apache.cxf.service.model.OperationInfo;
import org.apache.cxf.service.model.ServiceInfo;
import org.apache.cxf.ws.addressing.AddressingPropertiesImpl;
import org.apache.cxf.ws.addressing.AttributedURIType;
import org.apache.cxf.ws.addressing.EndpointReferenceType;
import org.apache.cxf.ws.addressing.JAXWSAConstants;
import org.apache.cxf.ws.addressing.WSAddressingFeature;
import org.codehaus.groovy.runtime.InvokerHelper;

import at.ac.tuwien.vitalab.genesis2.base.ExecutionEnvironment;
import at.ac.tuwien.vitalab.genesis2.core.model.AbstractModelElement;
import at.ac.tuwien.vitalab.genesis2.plugin.AbstractGenesis2Plugin;
import at.ac.tuwien.vitalab.genesis2.plugin.callinterceptor.CallInterceptor;

import com.ibm.wsdl.xml.WSDLReaderImpl;
import com.ibm.wsdl.xml.WSDLWriterImpl;

/**
 * Web service reference model element.
 * 
 * @author Lukasz Juszczyk
 * 
 */
public class WsReference extends AbstractModelElement {

    private static final Map<String, HashMap<Collection<CallInterceptor>, Client>> CACHE_BY_URL = new HashMap<String, HashMap<Collection<CallInterceptor>, Client>>();

    private static final Map<String, Integer> CHECKSUM_BY_URL = new HashMap<String, Integer>();

    private HashMap<Collection<CallInterceptor>, Client> cache = null;

    private File wsdlFile = null;

    /**
     * Constructor.
     */
    public WsReference() {
	super();
    }

    /**
     * Constructor for external services.
     * 
     * @param wsdlUrl URL to WSDL.
     * @param endpointUrl URL to Endpoint.
     */
    public WsReference(String wsdlUrl, String endpointUrl) {
	this();

	if (wsdlUrl == null || endpointUrl == null) {
	    throw new IllegalArgumentException("Argument must be not null.");
	}

	setProperty("wsdlURL", wsdlUrl);
	setProperty("endpointURL", endpointUrl);
    }

    private void createCache() {
	synchronized (CACHE_BY_URL) {

	    String wsdlUrl = (String) getProperty("wsdlURL");

	    cache = CACHE_BY_URL.get(wsdlUrl);

	    if (cache == null) {
		cache = new HashMap<Collection<CallInterceptor>, Client>();
		CACHE_BY_URL.put(wsdlUrl, cache);
	    }
	}
    }

    /*
     * (non-Javadoc)
     * 
     * @see at.ac.tuwien.vitalab.genesis2.core.model.ICleanable#cleanup()
     */
    @Override
    public void cleanup() {
	if (wsdlFile != null) {
	    wsdlFile.delete();
	}
    }

    /**
     * Returns the URL to the WSDL file.
     * 
     * @return URL String.
     */
    public String getWsdlUrl() {
	return (String) getProperty("wsdlURL");
    }

    /**
     * Returns the URL to the Endpoint.
     * 
     * @return URL String.
     */
    public String getEndpointUrl() {
	return (String) getProperty("endpointURL");
    }

    /*
     * (non-Javadoc)
     * 
     * @see groovy.util.Expando#invokeMethod(java.lang.String, java.lang.Object)
     */
    @Override
    public Object invokeMethod(String name, Object args) {

	try {
	    return call(name, InvokerHelper.asList(args), null, null, null);
	} catch (Exception e) {
	    if (e instanceof GroovyRuntimeException) {
		throw (GroovyRuntimeException) e;
	    }
	    throw new GroovyRuntimeException(e.getMessage(), e);
	}
    }

    /**
     * Returns an intermediate wrapper that allows call interceptors.
     * 
     * @param interceptors Interceptors.
     * @return Wrapped reference.
     */
    public IntermediateWrapper withInterceptors(CallInterceptor... interceptors) {
	return withInterceptors(InvokerHelper.asList(interceptors));
    }

    /**
     * Returns an intermediate wrapper that allows call interceptors.
     * 
     * @param interceptors Interceptors.
     * @return Wrapped reference.
     */
    public IntermediateWrapper withInterceptors(Collection<CallInterceptor> interceptors) {
	IntermediateWrapper wrap = new IntermediateWrapper(this);
	wrap.withInterceptors(interceptors);
	return wrap;
    }

    /**
     * Returns an intermediate wrapper that allows to redirect replies.
     * 
     * @param replyTo Destination.
     * @return Wrapped reference.
     */
    public IntermediateWrapper replyTo(WsReference replyTo) {
	IntermediateWrapper wrap = new IntermediateWrapper(this);
	wrap.replyTo(replyTo);
	return wrap;
    }

    /**
     * Returns an intermediate wrapper that allows to redirect fault replies.
     * 
     * @param faultTo Destination.
     * @return Wrapped reference.
     */
    public IntermediateWrapper faultTo(WsReference faultTo) {
	IntermediateWrapper wrap = new IntermediateWrapper(this);
	wrap.faultTo(faultTo);
	return wrap;
    }

    /**
     * Web service reference wrapper.
     * 
     * @author Lukasz Juszczyk
     * 
     */
    private class IntermediateWrapper extends GroovyObjectSupport {

	private WsReference wsref = null;
	private WsReference replyTo = null;
	private WsReference faultTo = null;
	private Collection<CallInterceptor> interceptors = null;

	/**
	 * Constructor.
	 * 
	 * @param wsref Original web service reference.
	 */
	public IntermediateWrapper(WsReference wsref) {
	    this.wsref = wsref;
	}

	/**
	 * Returns an intermediate wrapper that allows to redirect replies.
	 * 
	 * @param replyTo Destination.
	 * @return Wrapped reference.
	 */
	public IntermediateWrapper replyTo(WsReference replyTo) {
	    this.replyTo = replyTo;
	    return this;
	}

	/**
	 * Returns an intermediate wrapper that allows to redirect fault
	 * replies.
	 * 
	 * @param faultTo Destination.
	 * @return Wrapped reference.
	 */
	public IntermediateWrapper faultTo(WsReference faultTo) {
	    this.faultTo = faultTo;
	    return this;
	}

	/**
	 * Returns an intermediate wrapper that allows call interceptors.
	 * 
	 * @param interceptors Interceptors.
	 * @return Wrapped reference.
	 */
	public IntermediateWrapper withInterceptors(Collection<CallInterceptor> interceptors) {
	    this.interceptors = interceptors;
	    return this;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see groovy.lang.GroovyObjectSupport#invokeMethod(java.lang.String,
	 * java.lang.Object)
	 */
	@Override
	public HashMap<String, Object> invokeMethod(String name, Object args) {

	    try {
		return wsref.call(name, InvokerHelper.asList(args), interceptors, replyTo, faultTo);
	    } catch (Exception e) {
		throw new GroovyRuntimeException(e.getMessage(), e);
	    }
	}
    }

    /**
     * Performs the actual call to the Web service.
     * 
     * @param operationName Name of the called operation.
     * @param args Call arguments.
     * @param interceptors Used interceptors.
     * @param replyTo ReplyTo destination.
     * @param faultTo FaultTo destination.
     * @return Response map. The key is the name of the outgoing header or
     *         "response" for the SOAP body content.
     * @throws Exception On all the errors which might happen.
     */
    protected ResponseMap call(String operationName, List<Object> args, Collection<CallInterceptor> interceptors,
	    WsReference replyTo, WsReference faultTo) throws Exception {

	// it's a complex method...

	Client client = null;

	WSDLReaderImpl wsdlReader = new WSDLReaderImpl();

	Definition wsdlDoc = wsdlReader.readWSDL(getWsdlUrl());

	StringWriter sw = new StringWriter();
	WSDLWriterImpl wsdlWriter = new WSDLWriterImpl();
	wsdlWriter.writeWSDL(wsdlDoc, sw);
	String wsdlString = sw.toString();

	createCache();

	synchronized (CHECKSUM_BY_URL) {

	    // try to get from cache

	    Integer checksum = wsdlString.hashCode();
	    Integer oldChecksum = CHECKSUM_BY_URL.get(getWsdlUrl());

	    if (checksum.equals(oldChecksum)) {
		client = cache.get(interceptors);
	    }

	    CHECKSUM_BY_URL.put(getWsdlUrl(), checksum);
	}

	if (client == null) {

	    // if not in cache, generate a new one and put into cache

	    WebServiceReferencePlugin refPlugin = AbstractGenesis2Plugin.getPlugin(WebServiceReferencePlugin.class);

	    wsdlFile = new File(refPlugin.repoDir, "ws" + getWsdlUrl().hashCode() + ".wsdl");

	    FileWriter fileWriter = new FileWriter(wsdlFile);
	    fileWriter.write(wsdlString);
	    fileWriter.close();

	    JaxWsDynamicClientFactory dcf = JaxWsDynamicClientFactory.newInstance();

	    client = dcf.createClient(wsdlFile.getAbsolutePath());

	    if (cache != null) {
		cache.put(interceptors, client);
	    }
	}

	// create common execution environment for data exchange among
	// interceptors
	if (interceptors != null) {
	    ExecutionEnvironment env = new ExecutionEnvironment();

	    for (CallInterceptor ci : interceptors) {
		ci.attachToEndpoint(client, null, env);
	    }
	}

	// retrieve model info of called operation
	Collection<OperationInfo> opInfos = client.getEndpoint().getEndpointInfo().getService().getInterface()
		.getOperations();
	OperationInfo opInfo = null;

	for (OperationInfo o : opInfos) {
	    if (o.getName().getLocalPart().equals(operationName)) {
		opInfo = o;
	    }
	}

	if (opInfo == null) {
	    throw new IllegalArgumentException("Operation '" + operationName + "' not found.");
	}

	// service & input message model info
	Collection<MessagePartInfo> mpInfos = opInfo.getInput().getMessageParts();
	ServiceInfo srvInfo = opInfo.getInterface().getService();

	// transform arguments into final format
	WebServiceRequest request = null;

	List<?> argList = InvokerHelper.asList(args);

	// are arguments a list or a map?
	if (argList.size() == 1) {
	    Object firstArg = argList.get(0);
	    if (firstArg instanceof Map) {
		request = new WebServiceRequest(srvInfo, mpInfos, (Map) firstArg);
	    }
	}

	if (request == null) {
	    request = new WebServiceRequest(srvInfo, mpInfos, argList);
	}

	if (replyTo != null || faultTo != null) {

	    WSAddressingFeature wsaf = new WSAddressingFeature();

	    wsaf.setAddressingRequired(true);
	    wsaf.initialize(client, null);

	    AddressingPropertiesImpl maps = new AddressingPropertiesImpl();

	    if (replyTo != null) {
		AttributedURIType aut = new AttributedURIType();
		EndpointReferenceType ert = new EndpointReferenceType();
		aut.setValue(replyTo.getEndpointUrl());
		ert.setAddress(aut);
		maps.setReplyTo(ert);
	    }

	    if (faultTo != null) {
		AttributedURIType aut = new AttributedURIType();
		EndpointReferenceType ert = new EndpointReferenceType();
		aut.setValue(faultTo.getEndpointUrl());

		ert.setAddress(aut);
		maps.setFaultTo(ert);
	    }

	    client.getRequestContext().put(JAXWSAConstants.CLIENT_ADDRESSING_PROPERTIES, maps);
	}

	// invoke the Web service
	Object[] responses = null;

	// response is always a map containing direct response and output
	// headers
	ResponseMap response = new ResponseMap();

	if (replyTo == null) {
	    responses = client.invoke(operationName, request.getArgumentAndHeaderList().toArray());

	    int i = 0;
	    if (responses != null && responses.length > 0) {
		for (MessagePartInfo mpi : opInfo.getOutput().getMessageParts()) {

		    QName name = mpi.getConcreteName();

		    Object r = responses[i++];

		    Class<?> rc = r.getClass();
		    XmlType xmlTypeAno = rc.getAnnotation(XmlType.class);

		    if (xmlTypeAno != null) {
			for (String prop : xmlTypeAno.propOrder()) {
			    if (prop.length() == 0) {
				continue;
			    }

			    if (prop.startsWith("_")) {
				prop = prop.substring(1);
			    }

			    response.put(prop, InvokerHelper.getProperty(r, prop));
			}
		    } else {
			String opResponseName = operationName + "Response";
			opResponseName = opResponseName.toLowerCase();

			if (name.getLocalPart().toLowerCase().equals(opResponseName)) {
			    response.put("response", r);
			}

			response.put(name.getLocalPart(), r);
		    }
		}
	    }

	} else {
	    client.invoke(new ClientCallback(), operationName, request.getArgumentAndHeaderList().toArray());
	}

	return response;
    }

    protected class ResponseMap extends HashMap<String, Object> {

	private static final long serialVersionUID = 4728750098412357986L;

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.util.HashMap#get(java.lang.Object)
	 */
	@Override
	public Object get(Object key) {
	    // if only one element, direct each request to it
	    if (size() == 1) {
		return super.get(keySet().iterator().next());
	    }
	    return super.get(key);
	}
    }
}
