package at.ac.tuwien.vitalab.genesis2.utils;

import groovy.lang.Closure;

import java.util.Map;

import at.ac.tuwien.vitalab.genesis2.core.model.AbstractModelElement;
import at.ac.tuwien.vitalab.genesis2.core.model.ClosureDelegateDispatcher;

/**
 * @author Lukasz Juszczyk
 * 
 */
public final class ClosureUtil {

    /**
     * Static instance.
     */
    private static final ClosureUtil INSTANCE = new ClosureUtil();

    /**
     * Constructor.
     */
    private ClosureUtil() {
	// nothing
    }

    /**
     * Returns a singleton instance.
     * 
     * @return Instance.
     */
    public static ClosureUtil getInstance() {
	return INSTANCE;
    }

    /**
     * Assigns a primary dispatcher, to whic all requests are directed first.
     * 
     * @param closure Closure.
     * @param dispObject Primary dispatcher object.
     * @return Intermediary dispatcher.
     */
    public ClosureDelegateDispatcher setPrimaryDispatcher(Closure closure, Object dispObject) {

	ClosureDelegateDispatcher dispatcher = getDispatcher(closure);

	dispatcher.setPrimaryDispatcher(dispObject);

	return dispatcher;
    }

    /**
     * Register constant values.
     * 
     * @param closure Closure.
     * @param consts Map of constants.
     * @return Intermediary dispatcher.
     */
    public ClosureDelegateDispatcher registerConstants(Closure closure, Map<String, Object> consts) {

	ClosureDelegateDispatcher dispatcher = getDispatcher(closure);

	for (String name : consts.keySet()) {
	    dispatcher.getConstants().put(name, consts.get(name));
	}

	return dispatcher;
    }

    /**
     * Register constant value.
     * 
     * @param closure Closure.
     * @param name Constant name.
     * @param value Constant value.
     * @return Intermediary dispatcher.
     */
    public ClosureDelegateDispatcher registerConstant(Closure closure, String name, Object value) {
	ClosureDelegateDispatcher dispatcher = getDispatcher(closure);

	dispatcher.getConstants().put(name, value);

	return dispatcher;
    }

    /**
     * Factory method returning a dispatched version of the closure or
     * 
     * @param closure
     * @return
     */
    private ClosureDelegateDispatcher getDispatcher(Closure closure) {
	closure.setResolveStrategy(Closure.DELEGATE_FIRST);

	Object origDelegate = closure.getDelegate();

	ClosureDelegateDispatcher dispatcher;

	if (origDelegate instanceof ClosureDelegateDispatcher) {
	    dispatcher = (ClosureDelegateDispatcher) origDelegate;
	} else {
	    dispatcher = new ClosureDelegateDispatcher();

	    dispatcher.getVarBindingInterceptors().add(origDelegate);

	    closure.setDelegate(dispatcher);
	}

	return dispatcher;
    }

    /**
     * Attaches delegate Objects for variable access.
     * 
     * @param closure Closure.
     * @param delegates Delegates.
     * @return Intermediary dispatcher.
     */
    public ClosureDelegateDispatcher attachDelegates(Closure closure, Object... delegates) {

	ClosureDelegateDispatcher dispatcher = getDispatcher(closure);

	for (Object del : delegates) {

	    if (dispatcher.getVarBindingInterceptors().contains(del)) {
		continue;
	    }

	    dispatcher.getVarBindingInterceptors().add(del);

	    if ((del instanceof AbstractModelElement)) {
		dispatcher.getReflectionMap().put(del.getClass().getSimpleName().toLowerCase(),
			(AbstractModelElement) del);
	    }
	}

	return dispatcher;
    }

    /**
     * Assings a value to a local variable withing a closure dispatcher.
     * 
     * @param closure Closure.
     * @param name Variable name.
     * @param value Value.
     * @return Instance of the closure dispatcher.
     */
    public ClosureDelegateDispatcher setVariable(Closure closure, String name, Object value) {

	ClosureDelegateDispatcher dispatcher = getDispatcher(closure);

	dispatcher.setVariable(name, value);

	return dispatcher;
    }

}
