package at.ac.tuwien.vitalab.genesis2.utils;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import javax.tools.JavaCompiler;
import javax.tools.ToolProvider;

import org.apache.log4j.Logger;

import com.sun.tools.internal.ws.wscompile.WsimportTool;
import com.sun.tools.internal.xjc.Driver;

/**
 * Compilation helper.
 * 
 * @author Lukasz Juszczyk
 * 
 */
public final class CompilationUtils {

    private static final Logger LOGGER = Logger.getLogger(CompilationUtils.class);

    /**
     * Static instance.
     */
    private static final CompilationUtils INSTANCE = new CompilationUtils();

    /**
     * Constructor.
     */
    private CompilationUtils() {
	// nothing
    }

    /**
     * Returns a singleton instance.
     * 
     * @return Instance.
     */
    public static CompilationUtils getInstance() {
	return INSTANCE;
    }

    /**
     * Call Java XML Compiler.
     * 
     * @param packageName Package name.
     * @param destDir Destination dir for files.
     * @param file Source XSD file.
     * @throws IOException If IO errors occur.
     */
    public void xjc(String packageName, File destDir, File file) throws IOException {

	long timestamp = System.currentTimeMillis();

	LOGGER.debug(timestamp + ": Starting XJC.");

	ByteArrayOutputStream os = new ByteArrayOutputStream();

	int ret = -1;

	try {
	    PrintStream status = new PrintStream(os);
	    PrintStream error = new PrintStream(os);

	    ret = Driver.run(new String[] { "-quiet", "-extension", "-p", packageName, "-d", destDir.getAbsolutePath(),
		    file.getAbsolutePath() }, status, error);

	} catch (Throwable t) {
	    throw new IOException(timestamp + ": XJC error: " + t.getMessage());
	}

	BufferedReader sr = new BufferedReader(new StringReader(os.toString()));

	String line = null;
	while ((line = sr.readLine()) != null) {
	    LOGGER.info(timestamp + ": XJC output: " + line);
	}

	if (ret != 0) {
	    throw new IOException("Error: cannot generate message types.");
	}

	LOGGER.debug(timestamp + ": Finished XJC.");
    }

    /**
     * Call Java Code Compiler.
     * 
     * @param classPath Classpath.
     * @param sourcePath Source path.
     * @param destPath Destination path.
     * @param javaFiles Paths to java files.
     * @throws IOException If IO errors occur.
     */
    public void javac(File classPath, File sourcePath, File destPath, File... javaFiles) throws IOException {

	List<File> paths = new ArrayList<File>();

	if (classPath != null) {
	    paths.add(classPath);
	}

	javac(paths, sourcePath, destPath, javaFiles);
    }

    /**
     * Call Java Code Compiler.
     * 
     * @param classPaths List of classpaths.
     * @param sourcePath Source path.
     * @param destPath Destination path.
     * @param javaFiles Paths to java files.
     * @throws IOException If IO errors occur.
     */
    public void javac(Collection<File> classPaths, File sourcePath, File destPath, File... javaFiles)
	    throws IOException {

	long timestamp = System.currentTimeMillis();

	LOGGER.debug(timestamp + ": Starting JAVAC.");

	// first get all arguments into a vector

	List<String> args = new ArrayList<String>();

	String classPathStrings = null;

	if (classPaths != null) {
	    for (File cpf : classPaths) {
		if (classPathStrings == null) {
		    classPathStrings = cpf.getAbsolutePath();
		} else {
		    classPathStrings += File.pathSeparator + cpf.getAbsolutePath();
		}
	    }
	}

	if (classPathStrings != null) {
	    args.add("-cp");
	    args.add(classPathStrings);
	}

	if (sourcePath != null) {
	    args.add("-sourcepath");
	    args.add(sourcePath.getAbsolutePath());
	}

	if (destPath != null) {
	    args.add("-d");
	    args.add(destPath.getAbsolutePath());
	}

	for (File f : javaFiles) {
	    args.add(f.getAbsolutePath());
	}

	// call the compiler

	ByteArrayOutputStream os = new ByteArrayOutputStream();

	JavaCompiler compiler = ToolProvider.getSystemJavaCompiler();

	if (compiler == null) {
	    throw new IOException("JDK cannot find java compiler class.");
	}

	int code = compiler.run(null, os, os, args.toArray(new String[] {}));

	BufferedReader sr = new BufferedReader(new StringReader(os.toString()));

	String line = null;
	while ((line = sr.readLine()) != null) {
	    LOGGER.info(timestamp + ": JAVAC output: " + line);
	}

	if (code != 0) {
	    throw new IOException("Error: cannot compile Java source.");
	}

	LOGGER.debug(timestamp + ": Finished JAVAC.");
    }

    /**
     * Runs wsinfo to generate the web service client classes.
     * 
     * @param url URL of the Web service.
     * @param packageName Package name of client.
     * @param destDir Destination directory where to put the artifacts.
     * @throws IOException If IO errors occur.
     */
    public void wsimport(String url, String packageName, File destDir) throws IOException {

	long timestamp = System.currentTimeMillis();

	LOGGER.debug(timestamp + ": Starting WSIMPORT.");

	ByteArrayOutputStream os = new ByteArrayOutputStream();

	WsimportTool wsimport = new WsimportTool(os);

	boolean returnVal = wsimport.run(new String[] { "-d", destDir.getAbsolutePath(), "-s",
		destDir.getAbsolutePath(), "-p", packageName, url });

	if (!returnVal) {

	    BufferedReader sr = new BufferedReader(new StringReader(os.toString()));

	    String line = null;
	    while ((line = sr.readLine()) != null) {
		LOGGER.info(timestamp + ": WSIMPORT output: " + line);
	    }

	    throw new IOException("Error: cannot generate WS client stubs.");
	}

	LOGGER.debug(timestamp + ": Finished WSIMPORT.");
    }
}
