package at.ac.tuwien.vitalab.genesis2.utils;

import groovy.lang.Closure;

import java.util.Collection;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import at.ac.tuwien.vitalab.genesis2.core.model.AbstractModelElement;
import at.ac.tuwien.vitalab.genesis2.core.model.types.Host;
import at.ac.tuwien.vitalab.genesis2.core.model.ws.wrappers.ModelElementWrapper;

/**
 * Utility and convenience methods.
 * 
 * @author Lukasz Juszczyk
 * 
 */
public final class ModelElementUtil {

    /**
     * Static instance.
     */
    private static final ModelElementUtil INSTANCE = new ModelElementUtil();

    /**
     * Constructor.
     */
    private ModelElementUtil() {
	// nothing
    }

    /**
     * Returns a singleton instance.
     * 
     * @return Instance.
     */
    public static ModelElementUtil getInstance() {
	return INSTANCE;
    }

    /**
     * Checks the passed object for referenced model elements.
     * 
     * @param obj Object to check.
     * @return Set of model element references.
     */
    public Set<AbstractModelElement> listReferences(Object obj) {
	return listReferences(obj, false);
    }

    /**
     * Checks the passed object for referenced model elements.
     * 
     * @param obj Object to check.
     * @return Set of model element references, also including their nested
     *         references.
     */
    public Set<AbstractModelElement> listAllReferences(Object obj) {
	return listReferences(obj, true);
    }

    /**
     * Checks the passed object for referenced model elements.
     * 
     * @param obj Object to check.
     * @param recursive Whether to call recursively into model elements.
     * 
     * @return Set of model element references.
     */
    public Set<AbstractModelElement> listReferences(Object obj, boolean recursive) {

	HashSet<AbstractModelElement> result = new HashSet<AbstractModelElement>();

	if (obj != null) {

	    // basically we have 3 types of children: elements, collections, and
	    // maps

	    if (obj instanceof Map) {

		Map<?, ?> m = (Map<?, ?>) obj;

		for (Object k : m.keySet()) {
		    result.addAll(listReferences(k));
		    result.addAll(listReferences(m.get(k)));
		}

	    } else if (obj instanceof Collection) {

		Collection<?> c = (Collection<?>) obj;

		for (Object ci : c) {
		    result.addAll(listReferences(ci));
		}

	    } else if (obj instanceof AbstractModelElement) {

		result.add((AbstractModelElement) obj);

		if (recursive) {
		    result.addAll(listReferences(((AbstractModelElement) obj).getProperties()));
		}
	    }
	}

	return result;
    }

    /**
     * Returns a set of wrapped elements, for those who have not been created
     * yet at the given host.
     * 
     * @param host Host.
     * @param elements Elements.
     * @return Set of wrappers.
     */
    public Set<ModelElementWrapper> wrapFreshElements(Host host, Collection<AbstractModelElement> elements) {
	Set<ModelElementWrapper> result = new HashSet<ModelElementWrapper>();

	for (AbstractModelElement el : elements) {
	    Set<AbstractModelElement> refSet = new HashSet<AbstractModelElement>(el.getAllChildren());
	    refSet.add(el);

	    for (AbstractModelElement tr : refSet) {
		if (!tr.getCreatedAt().contains(host)) {
		    result.add(new ModelElementWrapper(tr));
		}
	    }

	}

	return result;
    }

    /**
     * Returns a clone of the passed object in which all referenced have been
     * attached the given model elements as delegates.
     * 
     * @param obj Object
     * @param elements Elements.
     * @return Cloned object.
     * 
     * @throws IllegalAccessException If collections or maps cannot be
     *             instantiated.
     * @throws InstantiationException If collections or maps cannot be
     *             instantiated.
     */
    @SuppressWarnings("unchecked")
    public Object cloneAndEnhanceClosures(Object obj, AbstractModelElement... elements) throws InstantiationException,
	    IllegalAccessException {

	Object result = obj;

	if (obj != null) {

	    if (obj instanceof Closure) {

		Closure clone = (Closure) ((Closure) obj).clone();
		ClosureUtil.getInstance().attachDelegates(clone, elements);

		result = clone;

	    } else if (obj instanceof Collection) {

		Collection<Object> list = (Collection<Object>) obj;
		Collection<Object> copy = (Collection<Object>) list.getClass().newInstance();

		copy.addAll(list);

		list.clear();

		for (Object lo : copy) {
		    list.add(cloneAndEnhanceClosures(lo, elements));
		}

		result = list;

	    } else if (obj instanceof Map) {

		Map<Object, Object> map = (Map<Object, Object>) obj;
		Map<Object, Object> copy = (Map<Object, Object>) map.getClass().newInstance();

		copy.putAll(map);

		map.clear();

		for (Entry<Object, Object> entry : copy.entrySet()) {
		    Object key = cloneAndEnhanceClosures(entry.getKey(), elements);
		    Object value = cloneAndEnhanceClosures(entry.getValue(), elements);

		    map.put(key, value);
		}

		result = map;
	    }
	}

	return result;
    }

}
