package at.ac.tuwien.vitalab.genesis2.utils;

import java.io.File;
import java.util.Random;
import java.util.UUID;

/**
 * Assigns repository directories to modules.
 * 
 * @author Lukasz Juszczyk
 * 
 */
public final class Repository {

    /**
     * Static instance.
     */
    private static final Repository INSTANCE = new Repository();

    /**
     * Constructor.
     */
    private Repository() {
	// nothing
    }

    /**
     * Returns a singleton instance.
     * 
     * @return Instance.
     */
    public static Repository getInstance() {
	return INSTANCE;
    }

    private Boolean initialized = false;

    // base dir
    private File repoDir = null;

    // initializes the base dir and makes sure it is writeable
    private void init() {

	synchronized (initialized) {

	    if (initialized) {
		return;
	    }

	    repoDir = new File("repository");
	    int i = 0;

	    while (true) {

		if (!repoDir.exists()) {
		    if (repoDir.mkdirs()) {
			break;
		    }
		}

		if (repoDir.isDirectory() && repoDir.canWrite()) {
		    break;
		}

		repoDir = new File("repository" + (i++));
	    }

	    initialized = true;
	}
    }

    /**
     * Create a new repository directory.
     * 
     * @param prefix Prefix string.
     * @return Directory.
     */
    public synchronized File newRepoDir(String prefix) {

	init();

	int i = -1;

	while (true) {

	    String name = null;

	    if (prefix == null) {
		name = "tmp_" + Math.abs(new Random().nextInt());
	    } else {
		if (i++ >= 0) {
		    name = prefix + "_" + (UUID.randomUUID().toString());
		} else {
		    name = prefix;
		}
	    }

	    File d = new File(repoDir, name);

	    if (d.mkdir()) {
		return d;
	    }
	}
    }

    /**
     * Create a new repository directory.
     * 
     * @return Directory.
     */
    public synchronized File newRepoDir() {

	return newRepoDir(null);
    }

    /**
     * Deletes recursively a directory.
     * 
     * @param dir Directory.
     * @return Whether it could be deleted.
     */
    public boolean deleteRepoDir(File dir) {

	if (dir == null || !dir.exists()) {
	    return true;
	}

	boolean result = true;
	if (dir.isDirectory()) {
	    for (File df : dir.listFiles()) {
		if (!deleteRepoDir(df)) {
		    result = false;
		}
	    }
	}
	if (!dir.delete()) {
	    result = false;
	}
	return result;
    }

    /**
     * Empties recursively a directory.
     * 
     * @param dir Directory.
     * @return Whether it could be emptied.
     */
    public boolean emptyRepoDir(File dir) {

	if (dir == null || !dir.exists()) {
	    return false;
	}

	boolean result = true;
	if (dir.isDirectory()) {
	    for (File df : dir.listFiles()) {
		if (!deleteRepoDir(df)) {
		    result = false;
		}
	    }
	}
	return result;
    }
}
